/*
 *	TCX Compiler Standard Definitions
 *
 *	THIS FREE SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 *	WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *	SPECIFICALLY DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR OTHER
 *	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *	TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 *	OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *	THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 *	USE OF THIS SOFTWARE OR ANY DERIVATIVE WORK, EVEN IF ADVISED OF THE
 *	POSSIBILITY OF SUCH DAMAGE.
 *
 *	This program is hereby placed in the public domain.  In
 *	contrast to other claims of "public domain", this means no
 *	copyright is claimed and you may do anything you like with TCX,
 *	including selling it!  As a gesture of courtesy, please retain
 *	the authorship information in the source code and documentation.
 *
 *	Contact:	Kirk Bailey (bailey@ieee.org)
 *			P.O. Box 1702
 *			Corvallis, OR 97339 (USA)
 *
 *	03/17/87
 */

#define	KIT_VERSION	FALSE		/* TRUE for T400-only version	*/

#ifdef LATTICE
#undef	max
#undef	min
#define	far
#define	itoa(n, str, base)	(sprintf(str, "%d", n), str)
#define	ltoa(n, str, base)	(sprintf(str, "%ld", n), str)
#endif		/* LATTICE */

#define TRUE		1
#define FALSE		0

#define	PUBLIC				/* Global to everyone		 */
#define	PRIVATE		static		/* Private to this file		 */
/*#define	register	*/	/* Delete registers for CodeView */
#define	VOID		void		/* No return value		 */

/*
 *	Variants of instruction address types:
 *		ABS	Symbols used are not self-relative;
 *		NONE	No address used;
 *		REL	Symbols used are forced to be self-relative.
 */

#define	ABS	0
#define	NONE	1
#define	REL	2

#include "tcxenv.h"
#include "tcxtypes.h"

typedef	INT32		LABEL;		/* Label description		*/
typedef	INT32		TEMP;		/* Temporary location description */

/*
 *	Define various lengths of character arrays
 */

#define	MAXLABEL	25		/* Max length of a label string	*/
#define	NAMEMAX		255		/* Max # of chars in a symbol name */
#define LINESIZE	2048		/* Size of input line array	*/
#define LINEMAX		(LINESIZE-1)	/* Leave room for a NUL character */
#define	MAXDBCH		10		/* Max # of #'s per line on .DBs */

/*
 *	Define output file for error messages.
 *
 *	Later - make this a command line parameter.
 */

#if 	((H_OPSYS == UNIX) || (H_OPSYS == BSD))
#define	STDERR	stderr
#else
#define	STDERR	stdout
#endif

/* Later - VMS BADEXIT should probably be a warning-level error */

#if H_OPSYS == VMS
#include "ssdef.h"			/* VMS exit codes are non-standard */
#define GOODEXIT	SS$_NORMAL
#define BADEXIT		SS$_NORMAL

#else
#define GOODEXIT	0
#define BADEXIT		1
#endif

/*
 *	Define target machine type ("Cpu_type").
 */

#define	TALL		0x8000		/* Gen. code for "any" 32 bit Tputer */
#define	T212		0x4000		/* Select T212/T222 CPU type	     */
#define	T2EX		0x2000		/* Select T225 CPU type		     */
#define	T414		0x1000		/* Select T414 CPU type		     */
#define	T4EX		0x0800		/* Select T400/T425 CPU type	     */
#define	T800		0x0400		/* Select T800 CPU type		     */
#define	T8EX		0x0200		/* Select T801 CPU type		     */

/*
 *	Define which procesor types support the "dup" and native FPU
 *	instructions, etc.
 */
#define	HAS_ANY		(T212 | T2EX | TALL | T414 | T4EX | T800 | T8EX)
#define	HAS_DUP		(T2EX | T4EX | T800 | T8EX)
#define	CPU_DUP		(Cpu_type & HAS_DUP)
#define	HAS_FPU		(T800 | T8EX)
#define	CPU_FPU		(Cpu_type & HAS_FPU)
#define	CPU_NO_DUP	(! CPU_DUP)
#define	CPU_NO_FPU	(! CPU_FPU)

/*
 *	Define floating point mode types.
 */

#define	FLOAT_PREANSI	0		/* Old style (force float to double */
#define	FLOAT_ANSI	1		/* Normal ANSI mode (default) */
#define	FLOAT_SINGLE	2		/* All FP done in single precision */

/*
 *	Define memory segment parameters (for compiled program).
 */

#define	SEG_NONE	-1		/* Flag to say no segment specified */

#define	SEG_CODE	0		/* Executable    readonly code	*/
#define	SEG_UNINIT	1		/* Uninitialized writable data	*/
#define	SEG_INIT	2		/* Initialized   writable data	*/

/*
 *	Define codes embedded within diversion -- used to switch streams
 *	and dvrt_out() time.
 */
#define	CODE_STREAM	-1		/* Switch to code/data space	*/
#define	LIT_STREAM	-2		/* Switch to literal space	*/
#define	LITLIT_STREAM	-3		/* Switch to literal-literal sp	*/

/*	Define the switch/loop queue		*/

struct	swq
	{
	struct	swq	*swq_prev;	/* Previous queue entry		*/
	LABEL		loop;
	LABEL		exit;
	};

/*
 *	Switch statement data structures.  There is a "switem" for every case
 *	label in a switch statement.  All of the "switem"s for a particular
 *	switch statement are linked together.
 */

struct	switem
	{
	struct	switem	*sw_next;	/* Next case label		*/
	INT32		sw_value;	/* Value of case label		*/
	LABEL		sw_label;	/* Label for this chunk of code */
	};

/*
 *	There is one "swheader" for every switch statement.  This is a
 *	structure for the convienence of the routines that mess with some of
 *	the sub-portions.  Actually, to the extent possible, it is probably
 *	reasonable to take functionality out of this structure.
 */

struct	swheader
	{
	struct	swheader *sh_prev;	/* Previous switch header block */
	struct	op_node	*sh_operand;	/* Operand block for switch stmt */
	struct	swq	sh_swqueue;	/* Labels for switch		*/
	LABEL		sh_default;	/* Default label		*/
	struct	switem	*sh_caselist;	/* Case entries			*/
	};

/*
 *	Text Diversion structure (for code output).  Logically, a "diversion"
 *	is a sequence of text that can be copied, moved and increased.
 */

#define	DVRTBLKSIZE	128		/* # data bytes/diversion buffer*/
#define DVRT_ALLOC	50		/* # of struct dvrt's to alloc	*/

struct	dvrt
	{
	struct	dvrt	*first;		/* Head of diver list or queue	*/
	struct	dvrt	*next;		/* Next newer diver list item	*/
	struct	dvrt	*prev;		/* Next older diver list item	*/
	int		count;		/* # of bytes in this element	*/
	int		status;		/* code/literal/litlit status	*/
	union				/* NOTE: This must be the last	*/
		{			/*	 field in the union	*/
		char		buf[2];	/* buffer if Inter_file == FALSE*/
		unsigned	blk_num;/* blk_num = fseek()/DISKBLKSIZE*/
		} text;
	};

/*
 *	Instruction information structure used for assembly language output.
 */

struct	ins_info
	{
	char		*name;		/* Instruction name		*/
	int		info;		/* Info about instruction:	*/
/*
 *	The bottom bits of this field hold the "average" instruction
 *	execution time in machine cycles.  The  rest of the field holds bits
 *	which define the characteristics of the instruction for the use of
 *	the peephole optimizer.  See "tcx7.c" for more information.
 */
#define	I_PEEPABLE	0x8000		/* Inst. starts peephole pattern*/
#define	I_WORDTIMING	0x4000		/* Timing depends on word size	*/
#define	I_FPOP1ROLE	0x2000		/* FP inst. for "fpop1" role	*/
#define	I_FPOP2ROLE	0x1000		/* FP inst. for "fpop2" role	*/
#define	I_OP3ROLE	0x0800		/* Inst. for "op3" role		*/
#define	I_BITMASK (I_PEEPABLE|I_WORDTIMING|I_FPOP1ROLE|I_FPOP2ROLE|I_OP3ROLE)
	};

/*
 *	Partially processed argument descriptor (used by "tcx6?.c").
 */

struct	arg_info
	{
	struct	dvrt	*code;		/* Code diversion handle	*/
	unsigned int	depth;		/* Depth used to evaluate	*/
	struct	op_node	*expr;		/* Expression ("op_node") handle */
	struct	op_node	*node;		/* Node in OP_PARAM list handle	*/
	INT32		size;		/* Size of parameter		*/
	};

/*
 *	Function #'s for internal functions (intrinsics).
 */

#define	BI_ajw			(0x10|T212|T2EX|TALL|T414|T4EX|T800|T8EX)
#define	BI_eval			(0x11|T212|T2EX|TALL|T414|T4EX|T800|T8EX)
#define	BI_fabs			(0x12|T800|T8EX)
#define	BI_fabsf		(0x13|T800|T8EX)
#define	BI_sqrt			(0x14|T800|T8EX)
#define	BI_sqrtf		(0x15|T800|T8EX)
#define	BI_memcpy		(0x16|T212|T2EX|TALL|T414|T4EX|T800|T8EX)
#define	BI_ChanIn		(0x17|T212|T2EX|TALL|T414|T4EX|T800|T8EX)
#define	BI_ChanInChar		(0x18|T212|T2EX|TALL|T414|T4EX|T800|T8EX)
#define	BI_ChanInInt		(0x19|T212|T2EX|TALL|T414|T4EX|T800|T8EX)
#define	BI_ChanOut		(0x20|T212|T2EX|TALL|T414|T4EX|T800|T8EX)
#define	BI_ChanOutChar		(0x21|T212|T2EX|TALL|T414|T4EX|T800|T8EX)
#define	BI_ChanOutInt		(0x22|T212|T2EX|TALL|T414|T4EX|T800|T8EX)
#define	BI_ChanReset		(0x23|T212|T2EX|TALL|T414|T4EX|T800|T8EX)
#define	BI_ProcAfter		(0x24|T212|T2EX|TALL|T414|T4EX|T800|T8EX)
#define	BI_ProcGetPriority	(0x25|T212|T2EX|TALL|T414|T4EX|T800|T8EX)
#define	BI_ProcReschedule	(0x26|T212|T2EX|TALL|T414|T4EX|T800|T8EX)
#define	BI_ProcStop		(0x27|T212|T2EX|TALL|T414|T4EX|T800|T8EX)
#define	BI_ProcWait		(0x28|T212|T2EX|TALL|T414|T4EX|T800|T8EX)
#define	BI_SetTime		(0x29|T212|T2EX|TALL|T414|T4EX|T800|T8EX)
#define	BI_Time			(0x30|T212|T2EX|TALL|T414|T4EX|T800|T8EX)
#define	BI_SetLoPriQ		(0x31|T212|T2EX|TALL|T414|T4EX|T800|T8EX)
#define	BI_SetHiPriQ		(0x32|T212|T2EX|TALL|T414|T4EX|T800|T8EX)
#define	BI_GetLoPriQ		(0x33|T212|T2EX|TALL|T414|T4EX|T800|T8EX)
#define	BI_GetHiPriQ		(0x34|T212|T2EX|TALL|T414|T4EX|T800|T8EX)
#define	BI_BitCnt		(0x35|T2EX|T4EX|T800|T8EX)
#define	BI_BitRevNBits		(0x36|T2EX|T4EX|T800|T8EX)
#define	BI_BitRevWord		(0x37|T2EX|T4EX|T800|T8EX)
#define	BI_Move2D		(0x38|T4EX|T800|T8EX)
#define	BI_Move2DNonZero	(0x39|T4EX|T800|T8EX)
#define	BI_Move2DZero		(0x40|T4EX|T800|T8EX)
#define	BI_PHalt		(0x41|T212|T2EX|TALL|T414|T4EX|T800|T8EX)
#define	BI_PRun			(0x42|T212|T2EX|TALL|T414|T4EX|T800|T8EX)

/*
 *	The New Expanded Super Duper Symbol Table Stuff!!!
 *
 *	The member list is required to do structure initialization.
 */

/*
 *	Combined symbol/type table element.
 */

union	value
	{
	INT32		integral;	/* Holds integral constants	*/
	UINT32		uintegral;	/* For unsigned values		*/
	INT32		offset;		/* Offset to auto variable	*/
	INT32		single;		/* Holds floating constants	*/
	LABEL		label;		/* Holds literal string label	*/
	double		fvalue;		/* Holds a floating point val	*/
	INT32		funcnum;	/* Holds internal function #	*/
	struct
		{
		char	bitoffset;	/* Offset within word		*/
		char	fieldwidth;	/* Width of a bitfield		*/
		} bitfield;
	};

#define	SYM_ALLOC	100		/* Number to allocate in a chunk */
#define	MAX_SYM_BUCKETS	4095		/* Size of symbol bucket table	*/

struct	id_node
	{
	INT32		type;		/* What type of a node it is	*/
	char		*name;		/* Name field			*/
	union
		{
		struct	id_node	*next_ptr;	/* Next bucket element	*/
		struct	id_node	**head_ptr;	/* Bucket list head	*/
		} bucket;
	struct	id_node	*level_ptr;	/* Links the nodes in a given scope */
	union
		{
		struct	id_node *base_type;	/* Points to base type node */
		struct	id_node *member_head;	/* Start of struct memb list */
		} type_ptr;
	int		name_space;	/* Name space for members	*/
	union
		{
		struct	id_node	*next_member;	/* Next member list ptr	*/
		int	storage_class;	/* Storage class of a variable	*/
		} info;
	int		level;		/* Nesting level for this symbol */
/*
 *	The value field holds the size in bytes for primitive types
 *	and the size of an element (in bytes) for arrays.
 */
	union	value	val;		/* Value of node		*/
	union
		{
		INT32	offset;		/* "address" of symbol on stack */
		LABEL	surrogate;	/* Surrogate label name for statics */
		INT32	tempsize;	/* Temp space size for struct/union */
		} address;
	struct	dvrt	*diversion;	/* Diverted code if needed	*/
	struct	prototype *proto_list;	/* Proto-type list		*/
	};

#ifdef LINT_ARGS
#ifndef	MAKE_HEADERS
#include "externs.h"
#endif		/* MAKE_HEADERS */
#endif		/* LINT_ARGS */

/*
 *	Definitions for the op_node structure for building expressions.
 *
 *	An <expression> is a tree of operand nodes (op_node) that define
 *	the subexpressions and operators comprising an expression.
 */
#define	OP_ALLOC	50		/* Allocate at a time		*/

struct	op_node
	{
	int	operator;		/* Code for operator of this node */
#define	OP_TERM		0x0000		/* Defines a terminal node with	*/
					/* "right" pointing to id_node	*/

/*	Note:	In the case of "op=" functions, the commutable flag	*/
/*		refers to the commutability of the "op" rather than	*/
/*		assignment operator.					*/
#define	OP_COMM		0x8000		/* Operator is commutative	*/
#define	OP_AOP		0x4000		/* Some form of assignment op	*/

#define	OP_MEMBER	(0x0001)		/* "." member of	*/
						/* Note: x->y becomes	*/
						/* (*x).y		*/
#define	OP_OFFSET	(0x0002)		/* Like a member, but	*/
						/* only an offset	*/
#define	OP_INDIRECT	(0x0003)		/* * indirect 		*/
#define	OP_ADDRESS	(0x0004)		/* & address of primary	*/
#define	OP_UMIN		(0x0005)		/* Unary minus		*/
#define	OP_UPLUS	(0x0006)		/* Unary plus		*/
#define	OP_LNOT		(0x0007)		/* ! logical NOT	*/
#define	OP_BNOT		(0x0008)		/* ~ Bitwise NOT	*/
#define	OP_POSTADD	(0x0009)		/* ++ Post Increment	*/
#define	OP_POSTSUB	(0x000A)		/* -- Post Decrement	*/
/*
 * The left and right nodes on INCREMENT and DECREMENT operators are used
 * in the following manner:
 *
 *	If the left node pointer is non NULL then the operator is POSTFIX
 *	else it is PREFIX (and the pointer had better be non-NULL!!)
 */
#define	OP_MUL		(0x000B|OP_COMM)	/* * Multiply		*/
#define	OP_DIV		(0x000C)		/* / Divide		*/
#define	OP_MOD		(0x000D)		/* % Modulus		*/
#define	OP_ADD		(0x000E|OP_COMM)	/* + Addition		*/
#define	OP_SUB		(0x000F)		/* - Subtraction	*/
#define	OP_RSHIFT	(0x0010)		/* >> Shift right	*/
#define	OP_LSHIFT	(0x0011)		/* << Shift left	*/
#define	OP_AMUL		(OP_MUL|OP_AOP)		/* *= Multiply		*/
#define	OP_ADIV		(OP_DIV|OP_AOP)		/* /= Divide		*/
#define	OP_AMOD		(OP_MOD|OP_AOP)		/* %= Modulus		*/
#define	OP_AADD		(OP_ADD|OP_AOP)		/* += Addition		*/
#define	OP_ASUB		(OP_SUB|OP_AOP)		/* -= Subtraction	*/
#define	OP_ARSHIFT	(OP_RSHIFT|OP_AOP)	/* >>= Shift right	*/
#define	OP_ALSHIFT	(OP_LSHIFT|OP_AOP)	/* <<= Shift left	*/
#define	OP_LT		(0x0012)		/* < Less than		*/
#define	OP_GT		(0x0013)		/* > Greater than	*/
#define	OP_LE		(0x0014)		/* <= Less than or equal*/
#define	OP_GE		(0x0015)		/* >= Greater or equal	*/
#define	OP_EQ		(0x0016|OP_COMM)	/* == Equal		*/
#define	OP_NE		(0x0017|OP_COMM)	/* != Not equal		*/
#define	OP_BAND		(0x0018|OP_COMM)	/* & Bitwise and	*/
#define	OP_BXOR		(0x0019|OP_COMM)	/* ^ Bitwise xor	*/
#define	OP_BOR		(0x001A|OP_COMM)	/* | Bitwise or		*/
#define	OP_ABAND	(OP_BAND|OP_AOP)	/* &= Bitwise and	*/
#define	OP_ABXOR	(OP_BXOR|OP_AOP)	/* ^= Bitwise xor	*/
#define	OP_ABOR		(OP_BOR|OP_AOP)		/* |= Bitwise or	*/
#define	OP_LAND		(0x001B)		/* && Logical and	*/
#define	OP_LOR		(0x001C)		/* || Logical or	*/
#define	OP_TERN		(0x001D)		/* ? Ternary lead-in 	*/
#define	OP_TSEL		(0x001E)		/* : Ternary select	*/
#define	OP_SEQ		(0x001F)		/* , Sequence operator	*/
#define	OP_CALL		(0x0020)		/* Function call.	*/
						/* "right" is function	*/
						/* "left" is params	*/
#define	OP_ICALL	(0x0021)		/* Internal func. call	*/
						/* "right" is func #	*/
						/* "left" is params	*/
#define	OP_PARAM	(0x0022)		/* Parameter chain	*/
#define	OP_CONVERT	(0x0023)		/* Cnvrt right operand	*/
						/* to type of this node	*/
#define	OP_ASSIGN	(0x0024|OP_AOP)		/* = Assignment		*/
#define	OP_BOOL		(0x0025)		/* Make canonical BOOL	*/

	int		idepth;		/* Depth of integer stack	*/
	int		fdepth;		/* Depth of floating stack	*/

	struct
		{
		union
			{
			struct	id_node	*terminal;
			struct	op_node	*non_terminal;
			} ptr;
		} left,right;

	struct	id_node	*node_type;	/* Type of this node	*/
	BOOL		lvalue;		/* TRUE if node is lvalue */
	};

/*
 *	Prototype -- Function prototype structure
 *
 *	Hanging off the node representing a function is a list of parameter
 *	types.  The list is made up of these nodes.
 *
 *	There are several special cases to this list.
 *		type = NULL, name = NULL	->	"..." parameter
 *		type = Voidtype, name = NULL	->	No parameters allowed
 *		type = <type>, name = NULL	->	(int, char, etc.)
 *		type = NULL, name = <name>	->	(a, b, etc.)
 *
 *	If the pointer to this list in the parent id_node is NULL, then
 *	it represents an "old style" (pre-ANSI) form of function declaration
 *	e.g. "extern func()".
 */
struct prototype
	{
	struct	prototype *next;	/* Next in prototype list	*/
	struct	id_node	*type;		/* Type of parameter		*/
	char		*name;		/* The (optional) name given.	*/
    /* The ANSI form of function prototype declaration can optionally	*/
    /* specify a name.  It is (of course) required if a function body	*/
    /* declaration.  This pointer to name can be NUKED after the	*/
    /* function body has been processed.				*/
	};

/*
 *	Scoping information structure.  Keeps track of information about a
 *	current scope such as the symbol table bucket rip up chain, whether
 *	the scope needs to be mentioned in the debugging information, etc.
 */
struct	scope_info
	{
	struct	id_node		*scope_chain;	/* Symbol table rip up chain */
	BOOL			debug_needs;	/* TRUE/needed in debug info */
	LABEL			start_label;	/* Scope starting label */
	};

/*
 *	Shortform data structure access macro definitions.
 */
#define	ISSIGNED(NODE)		((NODE)->node_type->type & SIGNED_NODE)

#define	ISUNSIGNED(NODE)	((NODE)->node_type->type & UNSIGNED_NODE)

#define	ISTERMINAL(NODE)	((NODE)->operator == 0)

#define	ISNON_TERMINAL(NODE)	(!ISTERMINAL(NODE))

#define	ISCONSTANT(NODE)		\
	(ISTERMINAL(NODE)	&&	\
	((NODE)->right.ptr.terminal->info.storage_class == SC_CONSTANT))

#define	ISRIGHTCONSTANT(NODE) (ISCONSTANT((NODE)->right.ptr.non_terminal))

#define	RIGHTCONSTANT(NODE)	\
	((NODE)->right.ptr.non_terminal->right.ptr.terminal->val.integral)

#define	RIGHTFCONSTANT(NODE)	\
	((NODE)->right.ptr.non_terminal->right.ptr.terminal->val.fvalue)
	
#define	ISLEFTCONSTANT(NODE) (ISCONSTANT((NODE)->left.ptr.non_terminal))

#define	LEFTCONSTANT(NODE)	\
	((NODE)->left.ptr.non_terminal->right.ptr.terminal->val.integral)

#define	LEFTFCONSTANT(NODE)	\
	((NODE)->left.ptr.non_terminal->right.ptr.terminal->val.fvalue)
	
#define	ISINTEGRAL(NODE)	((NODE)->node_type->type & INTEGRAL_NODE)

#define	ISPTR_TO_SYM(NODE)				\
	(ISNON_TERMINAL(NODE)	&&			\
	((NODE)->operator == OP_ADDRESS)	&&	\
	(ISTERMINAL((NODE)->right.ptr.non_terminal)))

#define	EXPRTYPE(NODE)		((NODE)->node_type->type)

#define	RIGHT(NODE)		((NODE)->right.ptr.non_terminal)
#define	RIGHTTYPE(NODE)		(EXPRTYPE(RIGHT(NODE)))

#define	LEFT(NODE)		((NODE)->left.ptr.non_terminal)
#define	LEFTTYPE(NODE)		(EXPRTYPE(LEFT(NODE)))

/*
 *	The expression code generation uses a flag (need_value)
 *	that contains some bit fields:
 *
 *		EVAL_ADDR	Evaluate target to an address only.
 *				If the target is an Lvalue, then
 *				DO NOT emit the indirection to load
 *				it's contents.  If the target is an
 *				Rvalue, you can crash and burn, if
 *				needed, as this is an error.
 *
 *		! EVAL_ADDR	Evaluate target to a value.  If the
 *				target is an Lvalue, then do whatever
 *				indirection is needed to load it's
 *				contents.  If the target is an Rvalue,
 *				then stop, as the value is now present.
 *
 *		NEED_VALUE	If NEED_VALUE is not set, then evaluation
 *				only proceeds to the extent necessary to
 *				compute side-effects.  It is required that
 *				the stack (A,B,C) be left at the same
 *				level whenever a NOT NEED_VALUE expression
 *				is processed.  The correct depth-usage
 *				must also be returned to the next higher
 *				level in any case.
 *
 *				If it is set, the expression is evaluated
 *				as one would expect.
 *
 *		RELAX_BOOL	If set, relaxes strict conformance to boolean
 *				values (I.e. 0 and 1)  If not set, then only
 *				values of 0 and 1 are returned by boolean
 *				operators (!= == <= < >= > && ||)
 */

#define	NO_VALUE		0x0000
#define	NEED_VALUE		0x0001
#define	EVAL_ADDR		0x0002
#define	RELAX_BOOL		0x0004

/*
 * Define the format of the "depth" value returned by all expr routines.
 *
 * This value combines four items:
 *
 *	Integer stack depth usage
 *	Integer temps used
 *	Floating stack depth usage
 *	Floating temps used.
 */
#define	MAX_TEMPS	32		/* Maximum # of temps in an expr */

#define	SH_INT_STACK	0
#define	SH_INT_TEMPS	3
#define	SH_FLT_STACK	8
#define	SH_FLT_TEMPS	11

#define	INT_STACK	(((unsigned int) 0x0007) << SH_INT_STACK)
#define	INT_TEMPS	(((unsigned int) 0x001F) << SH_INT_TEMPS)
#define	FLT_STACK	(((unsigned int) 0x0007) << SH_FLT_STACK)
#define	FLT_TEMPS	(((unsigned int) 0x001F) << SH_FLT_TEMPS)

#define	MAX_INT_STACK	((unsigned int) (3 << SH_INT_STACK))
#define	MAX_INT_TEMPS	((unsigned int) (INT_TEMPS - (1 << SH_INT_TEMPS)))
#define	MAX_FLT_STACK	((unsigned int) (3 << SH_FLT_STACK))
#define	MAX_FLT_TEMPS	((unsigned int) (FLT_TEMPS - (1 << SH_FLT_TEMPS)))

#define	INT_DEEP(X)	((X) << SH_INT_STACK)
#define	FLT_DEEP(X)	((X) << SH_FLT_STACK)

#define	UNDEF_DEEP	0xFFFF		/* Undefined depth */

/* End of TCX.H */
