/*
 *	TCX Compiler (2 of 11)
 *
 *	THIS FREE SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 *	WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *	SPECIFICALLY DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR OTHER
 *	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *	TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 *	OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *	THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 *	USE OF THIS SOFTWARE OR ANY DERIVATIVE WORK, EVEN IF ADVISED OF THE
 *	POSSIBILITY OF SUCH DAMAGE.
 *
 *	This program is hereby placed in the public domain.  In
 *	contrast to other claims of "public domain", this means no
 *	copyright is claimed and you may do anything you like with TCX,
 *	including selling it!  As a gesture of courtesy, please retain
 *	the authorship information in the source code and documentation.
 *
 *	Contact:	Kirk Bailey (bailey@ieee.org)
 *			P.O. Box 1702
 *			Corvallis, OR 97339 (USA)
 *
 *	03/17/87
 */

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<ctype.h>
#include	"tcx.h"
#include	"tcxdecl.h"
#include	"tcxext.h"

/*
 *	Local function declarations.
 */
PRIVATE	struct	prototype	*decl_comp();
PRIVATE	struct	prototype	*def_comp();
PRIVATE	VOID			flush_pnames();
PRIVATE	struct	prototype	*type_proto();
PRIVATE	VOID			bind_args();
PRIVATE	VOID			count_args();
PRIVATE	VOID			decl_args();
PRIVATE	VOID			declarg();
PRIVATE	struct	id_node		*make_arg();
PRIVATE	VOID			prologue();
PRIVATE	VOID			epilogue();
PRIVATE	VOID			alloc_stack();

/*
 *	This section processes function definitions and declarations.
 */

PRIVATE INT32	arg_locptr = 0;	/* Location pointer for assigned arg offsets */

/*
 *	Look for a function definition/declaration.
 */

#ifdef	MSC_OPT
#pragma	optimize("e",off)		/* Disable global reg optimizing */
#endif	/* MSC_OPT */

PUBLIC	struct	id_node	*
procfunc(funcname,type,sc)
	char			*funcname;
	struct	id_node		*type;
	int			sc;
	{
	struct	prototype	*args;
	struct	dvrt		*debug_sym;
	int			fscold;		/* "old" func storage class */
	struct	dvrt		*func;		/* The entire function	*/
	struct	scope_info	old_scope;
	BOOL			old_sinfo;
	struct	dvrt		*pro_dvrt;	/* Prologue of function	*/
	int			scp;		/* Storage class of param */
	struct	id_node		*sym;
	BOOL			was_decl;	/* TRUE if was a declaration */

	if((sc == SC_EXTERN) || (sc == SC_NONE))
		sc = SC_GLOBAL;		/* Make SC be SC_GLOBAL or SC_ST_GLB */
	else if(sc == SC_ST_EXT)
		sc = SC_ST_GLB;

	fscold = SC_NONE;		/* Assume no previous storage class */
	if((sym = add_loc(funcname,type,sc,NS_VAR,(INT32) FUNC_NODE)) == NULL)
		{
		sym = lookup_node(funcname,NS_VAR);	/* Find the symbol */
		if(! same_type(type,sym->type_ptr.base_type,
			/* ptr_eq_array = */ FALSE,/* void_ptr_ok = */ FALSE))
			{
			errname("Extern type redefinition",funcname);
			}
		fscold = sym->info.storage_class;
		}

	was_decl = FALSE;
	args = type->proto_list;	/* New parameter list */
	type->proto_list = NULL;	/* Don't want to free it twice... */
	if((args == NULL) && (! nextc('{')))
		was_decl = TRUE;	/* Old style declaration */
	else if((args != NULL) && (args->type != NULL) && (! nextc('{')))
		{
		was_decl = TRUE;
		sym->type_ptr.base_type->proto_list = decl_comp(args,
			sym->type_ptr.base_type->proto_list);
		/* New style declaration, remove names */
		flush_pnames(sym->type_ptr.base_type->proto_list);
		}
/*
 *	First we handle the case where we just got a declaration.  Note that
 *	this logic violates ANSI since we don't force the first "extern" of a
 *	function to mean external linkage.  We use the older convention that
 *	the "extern" is basically a vehicle to hang function type information
 *	on.  When ANSI is more accepted this should be modified (when every
 *	large program in existance has ALREADY been rewritten...).
 */
	if(was_decl)			/* Declaration of some type */
		{
		if(sc == SC_GLOBAL)	/* Equivalent to "extern" */
			sc = SC_EXTERN;
		else if(sc == SC_ST_GLB)/* Equivalent to "static" */
			sc = SC_ST_EXT;
		else
			fatal("Compiler error: procfunc #1");
		if(fscold != SC_NONE)	/* If previous def/decl */
			{
			if(sc == SC_ST_EXT)	/* "static" declaration */
				{
				if(fscold == SC_ST_GLB)
					sc = fscold;
				else if(fscold == SC_GLOBAL)
					{
			errname("Function already defined as non 'static'",
						funcname);
					sc = fscold;
					}
				else if((fscold != SC_ST_EXT) &&
					(fscold != SC_EXTERN))
					{
					fatal("Compiler error: procfunc #2");
					}
				}
			else
				sc = fscold;	/* Allow extern before/after */
			}
		sym->info.storage_class = sc;
		return (sym);		/* Indicate was declaration */
		}
/*
 *	Now handle storage class processing for a function definition.  Note
 *	that the above comment about non-ANSI "extern" processing applies here
 *	also.
 */
	if(fscold != SC_NONE)		/* Previously defined/declared? */
		{
		if((fscold == SC_ST_GLB) || (fscold == SC_ST_GLB))
			multidef(funcname);
		else if((fscold == SC_ST_EXT) && (sc == SC_GLOBAL))
			{
			errname("Function already declared as 'static'",
				funcname);
			}
		sym->info.storage_class = sc;
		}

	mark_scope_used();		/* Ensure function debug info output */
	if(Verbose)			/* Should we announce the function? */
		fprintf(STDERR, "=== %s()\n", funcname);

	if(Casestack != NULL)		/* This better not happen	*/
		fatal("Compiler error: profunc #3");

	if(Swqueue != NULL)		/* This either			*/
		fatal("Compiler error: profunc #4");

	Autospace = 0;			/* No autos for this func yet	*/
	Structspace = 0;		/* No struct return space yet	*/
	Tempspace = 0;			/* No temps for this func yet	*/

	Static_dvrt = NULL;		/* No static diversions yet	*/
	Func_type = sym->type_ptr.base_type->type_ptr.base_type;
#ifdef	DEBUG
		{
		extern	long	Dvrttotal;
		extern	int	Dvrtcount;

		Dvrttotal = Dvrtcount = 0;	/* Zero diversion usage nums */
		}
#endif	/* DEBUG */

	enter_scope(&old_scope);	/* Save the scoping info	*/
	bind_args(args);		/* Bind arguments in scope	*/
	if((args == NULL) || (args->type == NULL))	/* Old style def */
		{
		while(isdecl(&scp,&type,Inttype))
			{		/* Look for arg declarations	*/
			declarg(scp,type);
			}
		args = type_proto(args);/* Add types to proto list	*/
		/* New/old style decl, old style def - test match */
		args = def_comp(args,sym->type_ptr.base_type->proto_list,
			/*delete*/ FALSE);
		free_plist(args);	/* Old def doesn't create proto */
		args = NULL;
		}
	else
		{
		/* New style definition -- test match with declaration */
		sym->type_ptr.base_type->proto_list = def_comp(args,
			sym->type_ptr.base_type->proto_list,/*delete*/ TRUE);
		}
	flush_pnames(sym->type_ptr.base_type->proto_list);	/* No names */
	arg_locptr = Sizeof_int;	/* Define base offset for arguments */
	scan_ripup(count_args);		/* Count space used by args	*/

	dvrt_stack();			/* Divert code into a sep diversion */
	mark_scope_start();
	mark_scope_used();
	glblabel(funcname,sc);		/* Print function name		*/
	needpunc('{');			/* Function must begin with {	*/
	pro_dvrt = dvrt_unstack();	/* Get the start of the function */

	dvrt_stack();			/* Save debugging code offsets	*/
	Newfile = FALSE;		/* Diversion stuff turned it on	*/
	scan_ripup(decl_args);		/* Declare arguments		*/
	debug_sym = dvrt_unstack();	/* Debugging symbol offsets	*/

	if(Static_link)
		{
		if((sym = lookup_node(SLBNAME,NS_VAR)) == NULL)
			fatal("Compiler error: procfunc5");
		Slboff = word(sym->address.offset);	/* Static lnk offset */
		}

	dvrt_stack();
	Ncmp++;				/* New level of bracketing	*/
	block((INT32) 0);
	epilogue();			/* Ending code for function	*/
	exit_scope(&old_scope);		/* Terminate the scoping level	*/
	scan_literals(emit_literals);	/* Emit literal defs used in func */
	func = dvrt_unstack();		/* Save the function		*/

	dvrt_append(Static_dvrt);	/* Get statics			*/
	Static_dvrt = NULL;		/* Empty it			*/
	segment(SEG_CODE);
	dvrt_append(pro_dvrt);		/* Get the start of the function */

	old_sinfo = Source_info;
	Source_info = FALSE;		/* Turn off #line for a bit	*/
	alloc_stack();			/* Allocate stack space		*/
	dvrt_append(debug_sym);		/* Write debugging code offsets	*/
	prologue();			/* Beginning code for function	*/
	Source_info = old_sinfo;	/* Restore old status		*/

	dvrt_append(func);		/* Write the body of the function */
	dvrt_out(dvrt_get());		/* Output code to this point	*/
	Func_type = NULL;		/* No longer in function	*/

#ifdef	DEBUG
	if(Debug)
		{
		extern	int	Alloc_dvrt;
		extern	int	Alloc_oplist;
		extern	int	Alloc_sym;
		extern	int	Dvrtcount;
		extern	long	Dvrttotal;
		extern	int	Nfree_dvrt;
		extern	int	Nfree_oplist;
		extern	int	Nfree_sym;

		fprintf(STDERR,
			"Average diversion usage: %d bytes in %d blocks\n",
				(int) (Dvrttotal / Dvrtcount),Dvrtcount);
		fprintf(STDERR,
			"Op_nodes: %d/%d   Id_nodes: %d/%d   Dvrts: %d/%d\n",
				Nfree_oplist,Alloc_oplist,
					Nfree_sym,Alloc_sym,
						Nfree_dvrt,Alloc_dvrt);
		}
#endif	/* DEBUG */
	return (NULL);
	}

#ifdef	MSC_OPT
#pragma	optimize("e",on)		/* Enable global reg optimizing */
#endif	/* MSC_OPT */

/*
 *	Compare two prototype lists and delete the non-prototype form if
 *	present.
 */
PRIVATE	struct	prototype	*
decl_comp(new,old)
	struct	prototype	*new;
	struct	prototype	*old;
	{
	int			n;
	register struct	prototype	*nptr;
	register struct	prototype	*optr;
	char			pbuf[10];
	char			*pname;

	if(new == NULL)
		return (old);
	else if(old == NULL)
		return (new);

	for(n = 1, nptr = new, optr = old; (optr != NULL) || (nptr != NULL);
		nptr = nptr->next, optr = optr->next, ++n)
		{
		if(((nptr == NULL) || (optr == NULL)) ||
			(! same_type(nptr->type,optr->type,
				/* ptr_eq_array = */ TRUE,
					/* void_ptr_ok = */ FALSE)))
			{
			if(nptr != NULL)
				{
				if(nptr->name != NULL)
/* Use name of new param */		pname = nptr->name;
				else
					{
/* Else number of parameter */		sprintf(pbuf,"#%d",n);
					pname = pbuf;
					}
	errname("Parameter incompatible with earlier declaration",pname);
				}
			else
				{
				sprintf(pbuf,"#%d",n);
				errname("Parameter missing",pbuf);
				}
			break;
			}
		}
	free_plist(old);
	return (new);
	}

/*
 *	Compare a prototype list and a prototype list from a definition and
 *	do error checking.  Throw away the non-definition list.
 */
PRIVATE struct	prototype	*
def_comp(new,old,delete)
	struct	prototype	*new;
	struct	prototype	*old;
	BOOL			delete;
	{
	int			n;
	register struct prototype	*nptr;
	register struct prototype	*optr;
	char			pbuf[10];
	char			*pname;

	if(old != NULL)
		{
		for(n = 1, nptr = new, optr = old;
			(optr != NULL) || (nptr != NULL);
				nptr = nptr->next, optr = optr->next, ++n)
			{
			if(((nptr == NULL) || (optr == NULL)) ||
				(! same_type(nptr->type,optr->type,
					/* ptr_eq_array = */ TRUE,
						/* void_ptr_ok = */ FALSE)))
				{
				if(nptr != NULL)
					{
					if(nptr->name != NULL)
/* Use name of new param */			pname = nptr->name;
					else
						{
/* Else number of parameter */			sprintf(pbuf,"#%d",n);
						pname = pbuf;
						}
	errname("Parameter incompatible with earlier declaration",pname);
					}
				else
					{
					sprintf(pbuf,"#%d",n);
					errname("Parameter missing",pbuf);
					}
				break;
				}
			}
		if(delete)
			free_plist(old);
		}
	return (new);
	}

/*
 *	Remove parameter names from prototype list and return storage.
 */
PRIVATE	VOID
flush_pnames(list)
	struct	prototype	*list;
	{
	for(; list != NULL; list = list->next)
		{
		if(list->name != NULL)
			{
			free((char *) list->name);
			list->name = NULL;
			}
		}
	}

/*
 *	Add types to prototype list generated from old style argument list.
 */
PRIVATE	struct	prototype	*
type_proto(list)
	struct	prototype	*list;
	{
	struct	prototype	*proto;
	struct	id_node		*sym;

	if(list == NULL)
		{
/*
 *	Convert old style definitions with no arguments to new style.
 */
		list = get_proto();
		list->next = NULL;
		list->name = NULL;
		list->type = Voidtype;
		}
	else
		{
/*
 *	Add later determined types to prototype argument list.
 */
		for(proto = list; proto != NULL; proto = proto->next)
			{
			if(proto->name != NULL)
				{
				if((sym = lookup_node(proto->name,NS_VAR)) !=
					NULL)
					{
					proto->type = sym->type_ptr.base_type;
					}
				else
					fatal("Compiler error: type_proto");
				}
			else
				proto->type = Inttype;
			}
		}
	return (list);
	}

/*
 *	Process all args for a function
 */
PUBLIC	struct	prototype	*
procarg()
	{
	BOOL			ellipse;
	struct	prototype	*head;
	BOOL			isproto;
	VOID			junk();
	char			name[NAMEMAX + 1];
	struct	prototype	*new;
	int			sc;
	struct	id_node		*stack;
	struct	prototype	*tail;
	struct	id_node		*type;

	isproto = FALSE;		/* Assume !using prototypes to start */
	head = tail = NULL;		/* Empty prototype list */
	ellipse = FALSE;

	while ((! matchc(')')) && (! endst()) && (! ellipse))
		{
		name[0] = '\0';		/* Wipe out name */
/*
 *	Get storage class and type information from ANSI prototype list.
 */
		if(match("..."))
			{
			/* Ellipse parameter -- must be last */
			if(! nextc(')'))
				error("'...' must be last argument");
			ellipse = TRUE;
			type = NULL;
			sc = SC_ARG;	/* Default for lint */
			}
		else
			{
			(VOID) isdecl(&sc,&type,Inttype);
			if(type != NULL)
				{
				name[0] = '\0';
				stack = NULL;
				dodecl(&stack,name,FALSE);
				type = loadtype(&stack,type,SC_ARG);
				}
			else if((type == NULL) && !symname(name))
				error("identifier expected");
			}
/*
 *	Allows "register a, b" to look like legal non-ansi form.
 */
		if(! ellipse)
			{
			if(type != NULL)	/* ANSI proto form? */
				{
				if(type->type & (LABEL_NODE | FUNC_NODE))
					{
					error("Argument cannot be that type");
					type = Inttype;
					}
/*
 *	Something may be amiss here: ANSI says that prototype checking of
 *	function arguments must check array bound info.  Must ensure that
 *	this isn't being lost here.  I'm just not sure.  1-8-88 GO
 */
				if(type->type & ARRAY_NODE)
					{
					/* Chg array-of-x to ptr-to-x */
					type = ptrto(type->type_ptr.base_type);
					}

				if((!isproto) && (tail != NULL))
					{
				error("Missing previous parameter prototype");
					for(new = head; new != NULL;
						new = new->next)
						{
						if(new->type == NULL)
/* Dummies for err recovery */				new->type = Inttype;
						}
					}
				isproto = TRUE;
				if((sc == SC_REG) || (sc == SC_ARG))
/* SC_REG not currently used */		sc = SC_ARG;
				else if(sc != SC_NONE)
					{
					scerr();
					sc = SC_ARG;
					}
/*
 *	For machines other than the Transputer you may need to include a field
 *	in the prototype list to hold the storage class computed above (and
 *	below).  For now we avoid the extra field and ignore the storage class
 *	(it is implicitly assumed to be SC_ARG by "bind_args" and "declarg").
 */
				}
			else
				{
				if(isproto)
					{
					error("Missing parameter prototype");
					type = Inttype;
					}
				sc = SC_ARG;	/* Default storage class */
				}
			}

		new = get_proto();

		if(tail == NULL)	/* First on list? */
			{
			head = tail = new;
			if((type != NULL) && (type->type & VOID_NODE) &&
				(name[0] != '\0'))
				{
			error("void must be used without parameter name");
				}
			}
		else
			{
			tail->next = new;	/* Link in at end */
			tail = new;
			if((type != NULL) && (type->type & VOID_NODE))
				{
		error("void (sans name) must be first and only argument");
				}
			}

		new->next = NULL;	/* Initialize link pointer */
		new->type = type;	/* Initialize parameter type */
/*
 *	Ok, now we see if we have a parameter name.
 */
		if(name[0] != '\0')
			{
			if((new->name = (char *) malloc((unsigned int)
				strlen(name) + 1)) == NULL)
				{
				mem_err();
				}
			strcpy(new->name,name);
			}
		else
			new->name = NULL;

		if((!nextc(')')) && (!matchc(',')))
			{
			error("Expected comma");
			junk();		/* Try to find somewhere to restart */
			}
		}

	return (head);
	}

/*
 *	Get prototype list element.
 */
PUBLIC	struct prototype	*
get_proto()
	{
	struct	prototype	*new;

	if((new = (struct prototype *)
		malloc((unsigned int) sizeof(struct prototype))) == NULL)
		{
		mem_err();
		}
	return (new);
	}

PRIVATE	VOID
bind_args(list)
	struct	prototype	*list;
	{
	register struct	prototype	*new;
	struct	id_node		*sym;

	if(Func_type->type & SU_NODE)	/* Function returns struct/union */
		{			/* 1st param points to result mem */
		if((sym = add_loc(RSLINK,Inttype,SC_ARG,NS_VAR,
			(INT_NODE | SIGNED_NODE))) != NULL)
			{
			sym->val.integral = Inttype->val.integral;
			}
		else
			fatal("Compiler error: bind_args1");
		}
	if(Static_link)			/* Using static link data model */
		{
		if((sym = add_loc(SLBNAME,Inttype,SC_ARG,NS_VAR,
			(INT_NODE | SIGNED_NODE))) != NULL)
			{
			sym->val.integral = Inttype->val.integral;
			}
		else
			fatal("Compiler error: bind_args2");
		}
	new = list;
	if((new != NULL) && (new->next == NULL) && (new->type == Voidtype))
		return;			/* zip(void) {...} */
	else
		{
		for(; new != NULL; new = new->next)
			{
			if(new->name == NULL)
				{
/*
 *	See if we have an ellipse before calling it an error...
 */
				if(! ((new->next == NULL) &&
					(new->type == NULL)))
					{
					error("Missing formal parameter name");
					}
				}
			else
				{
				if(new->type == NULL)
					{
					if((sym = add_loc(new->name,Inttype,
						SC_ARG,NS_VAR,(INT_NODE |
						SIGNED_NODE |
						TENTATIVE_NODE))) != NULL)
						{
						sym->val.integral = Inttype->
							val.integral;
						}
					}
				else
					{
					if((sym = add_loc(new->name,new->type,
						SC_ARG,NS_VAR,new->type->
							type)) != NULL)
						{
						sym->val.integral = new->type->
							val.integral;
						}
					}
				if(sym == NULL)
					multidef(new->name);
				}
			}
		}
	}

/*
 *	Helper function for counting space used by argument variables.
 */
PRIVATE VOID
count_args(sym)
	struct	id_node		*sym;
	{
	if(sym->info.storage_class == SC_ARG)
		{
		arg_locptr +=  ((sym->val.offset + Sizeof_int - 1) /
			Sizeof_int) * Sizeof_int;
		}
	}

/*
 *	Helper function for declaring the offsets for argument variables.
 */
PRIVATE VOID
decl_args(sym)
	struct	id_node		*sym;
	{
	if((sym->type & TYPE_MASK) == 0)
		sym->type |= INT_NODE;	/* Undefined args become ints */

	arg_locptr = decl_help(sym,SC_ARG,arg_locptr);
	}

/*
 *	Add arguments to local symbol table.
 */
PRIVATE	VOID
declarg(sc,type)
	int			sc;
	struct	id_node		*type;
	{
	BOOL			func_def;
	char			name[NAMEMAX + 1];

/*
 *	Test validity of storage class.
 */
	if((sc == SC_REG) || (sc == SC_ARG))
		sc = SC_ARG;		/* SC_REG not currently used */
	else if(sc == SC_NONE)
		sc = SC_ARG;
	else
		{
		scerr();
		sc = SC_ARG;
		}
/*
 *	If not a null declaration, process the symbol name.
 */
	if(!nextc(';'))
		{
		do
			{
			(VOID) declvar(name,type,sc,make_arg,&func_def);
			} while (matchc(','));	/* End DO{} */
		}
	ns();
	}

/*
 *	Define an argument -- help for for declarg
 */

#ifdef	__TURBOC__
#pragma	warn	-par
#endif	/* __TURBOC__ */

PRIVATE	struct	id_node	*
make_arg(name,t,sc,info)
	char			*name;
	struct	id_node		*t;
	int			sc;
	INT32			info;
	{
	register struct	id_node	*sym;

	if(t->type & (LABEL_NODE | FUNC_NODE))
		errname("Argument can't be that type",name);
	else
		{
		if(t->type & ARRAY_NODE)
			{
			/* Chg array-of-x to ptr-to-x */
			t = ptrto(t->type_ptr.base_type);
			}
		else if(t->type & FP_NODE)
			t = Doubletype;
/*
 *	Make floating point arguments which aren't in the scope of a prototype
 *	be re-typed to "double".
 */
		if(((sym = lookup_node(name,NS_VAR)) == NULL) ||
			((sym->info.storage_class != SC_ARG) &&
				(sym->info.storage_class != SC_REG)))
			{
			errname("Not in argument list",name);
			}
		else if((sym->type & TENTATIVE_NODE) == 0)
			multidef(sym->name);
		else
			{
/*
 *	Fill in symbol table info - Now remove all other type bits and install
 *	the new type field.
 */
			sym->type = (sym->type & LAST_NODE) |
				(t->type & ~LAST_NODE);
			sym->val.integral = t->val.integral;
			sym->type_ptr.base_type = t;
			sym->info.storage_class = sc;
			}
		}
	return (NULL);			/* Not needed except for lint... */
	}

#ifdef	__TURBOC__
#pragma	warn	.par
#endif	/* __TURBOC__ */

/*
 *	Do function entry code.
 */
PRIVATE	VOID
prologue()
	{
	if(Tracein)			/* Do function entry trace fn	*/
		{
		instruction(TOP_LDC,ABS,WORKSPACE,NULL,(INT32) 0,NULL);
		instruction(TOP_CALL,REL,Triname,NULL,(INT32) 0,NULL);
		}
	}

/*
 *	Do function exit code.
 */
PRIVATE	VOID
epilogue()
	{
	if(Traceout)			/* Do function exit trace fn	*/
		instruction(TOP_CALL,REL,Troname,NULL,(INT32) 0,NULL);

	del_label();			/* Remove labels from the sym tab */
	instruction(TOP_RETF,ABS,WORKSPACE,NULL,(INT32) 0,NULL);
/*
 *	We need to make sure that there is at least eight bytes starting from
 *	the final "ret" instruction prior to running into any data or code
 *	following this function.  This is necessary in order to permit a worst
 *	case sized breakpoint at the function return point.  Since the "ret"
 *	instruction is 2 bytes long we must add another 6 bytes of padding.
 */
	if(Bloated)
		abs_oper(TOP_DS,(INT32) 6);
	}

/*
 *	Do what is necessary to allocate room on the stack for all of the
 *	local symbols.
 */
PRIVATE VOID
alloc_stack()
	{
	INT32			total;

	assign(WORKSPACE,NULL,FALSE,(total = (word(Autospace) +
		word(Structspace + (Sizeof_int - 1)) + Tempspace)));
	assign(TEMPBASE,NULL,FALSE,
		(INT32) word(Structspace + (Sizeof_int - 1)));
	assign(STRUCTBASE,NULL,FALSE,(INT32) 0);

	if(total != 0)			/* Allocate space */
		abs_oper(TOP_AJW,-total);
	}

/* End of TCX2.C */
