/*
 *	TCX Compiler (3 of 11)
 *
 *	THIS FREE SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 *	WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *	SPECIFICALLY DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR OTHER
 *	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *	TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 *	OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *	THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 *	USE OF THIS SOFTWARE OR ANY DERIVATIVE WORK, EVEN IF ADVISED OF THE
 *	POSSIBILITY OF SUCH DAMAGE.
 *
 *	This program is hereby placed in the public domain.  In
 *	contrast to other claims of "public domain", this means no
 *	copyright is claimed and you may do anything you like with TCX,
 *	including selling it!  As a gesture of courtesy, please retain
 *	the authorship information in the source code and documentation.
 *
 *	Contact:	Kirk Bailey (bailey@ieee.org)
 *			P.O. Box 1702
 *			Corvallis, OR 97339 (USA)
 *
 *	05/7/87
 *
 *	TCX general statement processor
 *
 *	Note:  The statement routine causes the Lattice C compiler to get
 *	a stack overflow.  If your compiler blows up, this may be the routine.
 *
 *	Note that a function body and a compound statement are very much alike.
 *	The major differences are handled in the "prologue" and "epilogue"
 *	routines elsewhere.
 */

#include	<stdio.h>
#include	<ctype.h>
#include	"tcx.h"
#include	"tcxdecl.h"
#include	"tcxext.h"

/*
 *	Local function declarations.
 */
PRIVATE	VOID			statement();
PRIVATE	BOOL			chklabel();
PRIVATE	INT32			declarations();
PRIVATE	VOID			decl_auto();
PRIVATE	VOID			decl_inits();
PRIVATE	VOID			decl_noninits();
PRIVATE	VOID			decl_regs();
PRIVATE	VOID			dobreak();
PRIVATE	VOID			docase();
PRIVATE	VOID			docontinue();
PRIVATE	VOID			dodefault();
PRIVATE	VOID			dodowhile();
PRIVATE	VOID			doexpr();
PRIVATE	VOID			dofor();
PRIVATE	VOID			dogoto();
PRIVATE	VOID			doif();
PRIVATE	VOID			doreturn();
PRIVATE	VOID			doswitch();
PRIVATE	VOID			dowhile();
PRIVATE	struct	op_node		*bool_expr();
PRIVATE	struct	op_node		*test_expression();

/*
 *	Process compound statements.
 */
PUBLIC	VOID
comp_block(locptr)
	INT32			locptr;	/* Stack offset for local vars */
	{
	struct	scope_info	old_scope;	/* Scope info on entry */

	Ncmp++;				/* Entered a new bracket level */
	enter_scope(&old_scope);	/* Save old scope */
	mark_scope_start();
	block(locptr);
	exit_scope(&old_scope);		/* Remove local symbols, etc. */
	}

/*
 *	Parse a block.
 */
PUBLIC	VOID
block(locptr)
	INT32			locptr;	/* Stack offset for local vars */
	{
	locptr = declarations(locptr);	/* Do block level declarations */

	while((! matchc('}')) && (! Eof))
		{
		statement(locptr);
		blanks();
		}
	if(! Eof)
		Ncmp--;
	}

/*
 *	Process a single C statement.
 */
PRIVATE	VOID
statement(locptr)
	INT32			locptr;
	{
	lineline(Lineno);
/*
 *	Note that labels may appear where other statement type are allowed but
 *	they do not alter the flow of control.  Thus the following "while"
 *	loop (this could also be done by having "chklabel" call "statement",
 *	but why waste the stack space...)
 */
	while(chklabel())
		;

	if((ch() == 0) && Eof)
		return;

	if(matchc('{'))
		comp_block(locptr);
	else if(amatch("break"))
		dobreak();
	else if(amatch("case"))
		docase();
	else if(amatch("continue"))
		docontinue();
	else if(amatch("default"))
		dodefault();
	else if(amatch("do"))
		dodowhile(locptr);
	else if(amatch("else"))
		error("else not matched with if");
	else if(amatch("for"))
		dofor(locptr);
	else if(amatch("goto"))
		dogoto();
	else if(amatch("if"))
		doif(locptr);
	else if(amatch("return"))
		doreturn();
	else if(amatch("switch"))
		doswitch(locptr);
	else if(amatch("while"))
		dowhile(locptr);
	else if(matchc(';'))
		;
	else
		doexpr();
	}

/*
 *	Check for a LABEL: on a line.
 */
PRIVATE	BOOL
chklabel()
	{
	char			labname[NAMEMAX+1];
	register char		*p;
	struct	id_node		*symbol;

	if(((p = chksym(labname)) == NULL) || chks("default"))
		return (FALSE);

	while(isspace(*p))
		++p;			/* Skip leading spaces */
/*
 *	Note that this allows white space, but not newlines.  Nasty.
 *		DCR		25 May 87
 */
	if(*p != ':')
		return (FALSE);

	if(chkeyword())
		keyerr(labname);	/* Invalid keyword usage */

	Lptr = p + 1;			/* Advance past the label */

	if((symbol = add_label(labname,getlabel(),TRUE)) == NULL)
		multidef(labname);
	padlabel(symbol->val.label);

	return (TRUE);
	}

/*
 *	Do local variable declarations for both function level stuff
 *	and block level variables.
 */

PRIVATE	INT32	help_locptr = 0;/* Helper variable used only by "declarations"
				 * and "decl_help".			*/

PRIVATE	INT32
declarations(locptr)
	INT32			locptr;
	{
	int			sc;
	struct	id_node		*type;

	Initlist = NULL;	/* Nothing in initialization list */

	while(isdecl(&sc,&type,Inttype))
		{
/*
 *	Process local declarations.  If storage class not set, use default
 */
		declloc(((sc == SC_NONE) ? SC_AUTO : sc),type);
		}
/*
 *	Now we go through all of the local symbols we just declared, and
 *	assign offsets to them starting with locptr (in terms of bytes
 *	rounded to words).
 */
	help_locptr = locptr;

	scan_ripup(decl_inits);		/* Initializers	*/
	scan_ripup(decl_noninits);	/* Non-initialized aggregates */
	scan_ripup(decl_auto);		/* All other non-register autos	*/
	scan_ripup(decl_regs);		/* Registers */

	if(-help_locptr > Autospace)	/* Remember the high water mark */
		Autospace = -help_locptr;
/*
 *	Output initialization expressions.
 */
	expr_out(Initlist,FALSE);

	return (help_locptr);
	}

/*
 *	Assign offset for AUTO variables.
 */
PRIVATE	VOID
decl_auto(sym)
	struct	id_node		*sym;
	{
	help_locptr = decl_help(sym,SC_AUTO,help_locptr);
	}

/*
 *	Assign offset to initialized aggregates.
 */
PRIVATE	VOID
decl_inits(sym)
	struct	id_node		*sym;
	{
	help_locptr = decl_aggreg(sym,/* init = */ TRUE,help_locptr);
	}

/*
 *	Assign offset for non-initialized aggregates.
 */
PRIVATE	VOID
decl_noninits(sym)
	struct	id_node		*sym;
	{
	help_locptr = decl_aggreg(sym,/* init = */ FALSE,help_locptr);
	}

/*
 *	Assign offset for Register variables.
 */
PRIVATE	VOID
decl_regs(sym)
	struct	id_node		*sym;
	{
	help_locptr = decl_help(sym,SC_REG,help_locptr);
	}

/*
 *	This routine is called once for every variable in the local scope.
 *	We must assign offsets to all variables which are of AUTO storage
 *	class and don't have an offset already applied to them.  The current
 *	offset is passed and the updated value is returned.
 */
PUBLIC	INT32
decl_help(sym,sc,offset)
	struct	id_node		*sym;
	int			sc;	/* Storage class to address	*/
	INT32			offset;	/* Offset to use		*/
	{
	INT32			need;	/* Number of words needed	*/

	if((! (sym->type & OFFSET_ASSIGNED)) &&
		(sym->info.storage_class == sc))
		{
		sym->type |= OFFSET_ASSIGNED;
		need = ((sym->val.integral + Sizeof_int - 1) / Sizeof_int) *
			Sizeof_int;
		sym->address.offset = offset - need;
		mark_scope_used();
		}
	else
		need = 0;

	return (offset - need);
	}

/*
 *	This routine is called once for every variable in the local scope.
 *	We must assign offsets to all variables which are of AUTO storage
 *	class and don't have an offset already applied to them.  The current
 *	offset is passed and the updated value is returned. 
 *
 *	The "init" parameter tells whether to assign storage to initialized
 *	aggregates or to any aggregate.  An aggregate is any item that is
 *	is an ARRAY_NODE, STRUCT_NODE or UNION_NODE.
 */
PUBLIC	INT32
decl_aggreg(sym,init,offset)
	struct	id_node		*sym;
	BOOL			init;	/* TRUE/looking for init. vars	*/
	INT32			offset;	/* Offset to use		*/
	{
	INT32			need;	/* Number of words needed	*/

	if((! (sym->type & OFFSET_ASSIGNED))			&&
			((sym->info.storage_class == SC_AUTO) ||
	(sym->info.storage_class == SC_REG))			&&
	(sym->type & (ARRAY_NODE | STRUCT_NODE | UNION_NODE))	&&
	(! init))
		{
		sym->type |= OFFSET_ASSIGNED;
/*
 *	The access to "val.integral" should probably be an access to the
 *	type pointer information.  In general we need to examine where the
 *	size information is getting stored (type versus value nodes).  Part of
 *	TCX uses one, part uses the other.  KAB & DCR 4/10/88.
 */
		need = ((sym->val.integral + Sizeof_int - 1) / Sizeof_int) *
			Sizeof_int;

		sym->address.offset = offset - need;
		mark_scope_used();
		}
	else
		need = 0;

	return (offset - need);
	}

/*
 *	Process the break statement.
 */
PRIVATE	VOID
dobreak()
	{
	char			lblbuf[NAMEMAX + 1];

	if((Swqueue != NULL) && (Swqueue->exit != (LABEL) 0))
		{
		instruction(TOP_J,REL,strlabel(Swqueue->exit,lblbuf),NULL,
			(INT32) 0, NULL);
		}
	else
		error("No active switch or loop");

	ns();
	}

/*
 *	Process case labels.
 */
PRIVATE	VOID
docase()
	{
	struct	op_node		*case_expression;
	LABEL			label;

	case_expression = expression();
	needpunc(':');

	if(! ISCONSTANT(case_expression))
		error("Case label is not a constant");
	else if(! ISINTEGRAL(case_expression))
		error("Case label is not integral");
	else
		{
/*
 *	Case label is a constant integral.
 *
 *	ANSI says case expression is (cast) to type of switch expression.  We
 *	don't need to do anything here for Transputer, but think about this
 *	for other processors.  7-Jul-87 GO
 */
		addcase(case_expression->right.ptr.terminal->val.integral,
			label = getlabel());
		padlabel(label);
		}
	free_op_node(case_expression);
	}

/*
 *	Process continue statement.
 */
PRIVATE	VOID
docontinue()
	{
	char			lblbuf[NAMEMAX + 1];

	if((Swqueue != NULL) && (Swqueue->loop != (LABEL) 0))
		{
		instruction(TOP_J,REL,strlabel(Swqueue->loop,lblbuf),NULL,
			(INT32) 0, NULL);
		}
	else
		error("No active loop statements");

	ns();
	}

/*
 *	Process default statement
 */
PRIVATE	VOID
dodefault()
	{
	needpunc(':');

	if(Casestack != NULL)
		{
		if(Casestack->sh_default != (LABEL) 0)
			error("Only one default label permitted");
		else
			padlabel(Casestack->sh_default = getlabel());
		}
	else
		error("No active switch statement");
	}

/*
 *	Process the do..while statement.
 */
PRIVATE	VOID
dodowhile(locptr)
	INT32			locptr;
	{
	char			lblbuf[NAMEMAX + 1];
	LABEL			loop_top;
	int			temp_line;
	struct	dvrt		*while_body;
	struct	op_node		*while_expression;
	struct	swq		while_queue;	/* Element for sw/loop queue */

	loop_top = getlabel();
	while_queue.loop = getlabel();
	while_queue.exit = getlabel();	/* Add labels to queue element */

	addswq(&while_queue);		/* Add this level to while queue */

	dvrt_stack();
	statement(locptr);		/* Capture body of while body */
	while_body = dvrt_unstack();

	temp_line = Lineno;

	if(! match("while"))
		error("Missing while in dowhile");

	while_expression = test_expression();

	padlabel(loop_top);
	dvrt_append(while_body);
	padlabel(while_queue.loop);

	lineline(temp_line);
/*
 *	Check for constant "dowhile" expression.
 */
	if(! ISCONSTANT(while_expression))
		{
		expr_out(while_expression,NEED_VALUE | RELAX_BOOL);
		instruction(TOP_CJ,REL,strlabel(while_queue.exit,lblbuf),NULL,
			(INT32) 0,NULL);
	
		instruction(TOP_J,REL,strlabel(loop_top,lblbuf),NULL,
			(INT32) 0, NULL);
		}
	else
		{
		if(while_expression->right.ptr.terminal->val.integral != 0L)
			{
			instruction(TOP_J,REL,strlabel(loop_top,lblbuf),NULL,
				(INT32) 0,NULL);
			}
		free_op_node(while_expression);
		}
	padlabel(while_queue.exit);

	delswq();

	ns();
	}

/*
 *	If it wasn't a keyword, then we just have an expression (e.g. a
 *	function call or assignment).
 */
PRIVATE	VOID
doexpr()
	{
	struct	op_node 	*expr;	/* Hold the value */

	expr = expression();		/* Get the expression */
	lineline(Lineno);
	expr_out(expr,NO_VALUE);	/* Write the code out */
	ns();
	}

/*
 *	Process the for statement.
 *
 *	Remember to handle the special case where the test expression is
 *	blank.
 *
 *	Generate code that looks like:
 *
 *		initializer
 *	top:
 *		test expression
 *		cj	@done
 *		. . . body . . .
 *	inclabel:
 *		increment expression
 *		j	@top
 *	done:
 *
 *	Note that on a machine where "descheduling" is not needed the
 *	above form can be improved on using in the following fashion:
 *
 *		initializer
 *		j	@testexpression
 *	top:
 *		. . . body . . .
 *	inclabel:
 *		increment expression
 *	testexpression:
 *		test expression
 *		negate
 *		cj	@not_done top
 *	done:
 */
PRIVATE	VOID
dofor(locptr)
	INT32			locptr;
	{
	struct	dvrt		*body;
	struct	swq		for_queue;	/* element for sw/loop queue */
	struct	op_node		*increment;
	struct	op_node		*initializer;
	char			lblbuf[NAMEMAX + 1];
	struct	op_node		*test;
	LABEL			toplab;		/* Start of the body */

	initializer = test = increment = NULL;	/* No expressions yet */

	needpunc('(');
	if(! matchc(';'))		/* If there is an initializer get it */
		{
		initializer = expression();
		needpunc(';');
		}
	if(! matchc(';'))		/* If there is a test get it */
		{
		test = bool_expr();
		needpunc(';');
		}
	if(! matchc(')'))		/* If there is an increment get it */
		{
		increment = expression();
		needpunc(')');
		}

	lineline(Lineno);

	for_queue.loop = getlabel();	/* Get the various labels */
	for_queue.exit = getlabel();
	toplab = getlabel();
	addswq(&for_queue);		/* Add this level to switch queue */

	dvrt_stack();			/* Get the body of the statement */
	statement(locptr);
	body = dvrt_unstack();

	expr_out(initializer,NO_VALUE);	/* Initializer */
/*
 *	Test for a constant valued "test" expression.
 */
	if((test != NULL) && ISCONSTANT(test))
		{
		if(test->right.ptr.terminal->val.integral == 0L)
			{
			dvrt_free(body);
			free_op_node(test);
			free_op_node(increment);
			delswq();
			return;
			}
		else			/* Unconditional true */
			{
			free_op_node(test);
			test = NULL;
			}
		}
/*
 *	Now do the general case (including the unconditional true cases).
 */
	padlabel(toplab);
	if(test != NULL)
		{
		expr_out(test,NEED_VALUE | RELAX_BOOL);
		instruction(TOP_CJ,REL,strlabel(for_queue.exit,lblbuf),NULL,
			(INT32) 0, NULL);
		}
	dvrt_append(body);
	padlabel(for_queue.loop);
	expr_out(increment,NO_VALUE);
	instruction(TOP_J,REL,strlabel(toplab,lblbuf),NULL,(INT32) 0,NULL);
	padlabel(for_queue.exit);

	delswq();
	}

/*
 *	Process the goto statement.
 *
 *	Slight problem:  A goto to a label that is never defined is not
 *	caught as an error.		DCR	20 May 87
 */
PRIVATE	VOID
dogoto()
	{
	char			goto_label[NAMEMAX + 1];
	struct	id_node		*label_sym;
	char 			lblbuf[MAXLABEL + 1];

	if(symname(goto_label))
		{
		label_sym = add_label(goto_label,getlabel(),FALSE);
		instruction(TOP_J,REL,strlabel(label_sym->val.label,lblbuf),
			NULL,(INT32) 0,NULL);
		}
	else
		{
		error("Illegal goto symbol");
		junk();
		}
	ns();
	}

/*
 *	Process the if (xx) ... else ... statement.
 */
PRIVATE	VOID
doif(locptr)
	INT32			locptr;
	{
	LABEL			endlab;
	struct	dvrt		*false_statement;
	LABEL			falselab;
	struct	op_node		*if_expression;
	char			lblbuf[NAMEMAX + 1];
	int			temp_line1;
	int			temp_line2;
	struct	dvrt		*true_statement;

	if_expression = test_expression();

	temp_line1 = Lineno;

	dvrt_stack();
	statement(locptr);
	true_statement = dvrt_unstack();

	if(amatch("else"))
		{
		temp_line2 = Lineno;

		dvrt_stack();
		statement(locptr);
		false_statement = dvrt_unstack();
		}
	else
		{
		false_statement = NULL;
		temp_line2 = 0;		/* For lint */
		}
/*
 *	Check for constant "if" expression.
 */
	if(ISCONSTANT(if_expression))
		{
		if(if_expression->right.ptr.terminal->val.integral != 0L)
			{
			dvrt_append(true_statement);
			dvrt_free(false_statement);
			}
		else
			{
			dvrt_append(false_statement);
			dvrt_free(true_statement);
			}
		free_op_node(if_expression);
		return;
		}
	lineline(temp_line1);
/*
 *	Wasn't a constant "if", do the real thing.
 */
	expr_out(if_expression,NEED_VALUE | RELAX_BOOL);
	endlab = getlabel();
	falselab = ((false_statement == NULL) ? endlab : getlabel());
	instruction(TOP_CJ,REL,strlabel(falselab,lblbuf),NULL,(INT32) 0,NULL);
	dvrt_append(true_statement);
	if(false_statement != NULL)
		{
		lineline(temp_line2);
		instruction(TOP_J,REL,strlabel(endlab,lblbuf),NULL,(INT32) 0,
			NULL);
		padlabel(falselab);
		dvrt_append(false_statement);
		}
	padlabel(endlab);
	}

/*
 *	Process return statement.
 */
PRIVATE VOID
doreturn()
	{
	struct	op_node		*return_expression;
	INT32			size;
	struct	id_node		*sym;

	if(! endst())
		{
		return_expression = expression();
		if(! same_type(return_expression->node_type,Func_type,
			/* ptr_eq_array = */ TRUE,/* void_ptr_ok = */ TRUE))
			{
/*
 *	If the return type is different from the function type, check to see
 *	if they are numeric types such that applying the usual unary
 *	promotions will make them equivalent.
 */
			if(((Func_type->type & ARITH_NODE) == 0) ||
				(basetype(return_expression->
					node_type->type) !=
						basetype(Func_type->type)))
				{
/*
 *	Note that if the function type is a pointer, then it is OK to have a
 *	NULL constant as the return value.
 */
				if(! ((Func_type->type & PTYPE_NODE) &&
					ISCONSTANT(return_expression) &&
					ISINTEGRAL(return_expression) &&
					(return_expression->
						right.ptr.terminal->
							val.integral == 0)))
					{
					warning(3,"Inconsistent return type");
					}
				}
			return_expression = cast_it(return_expression,
				Func_type);
			}
		if(EXPRTYPE(return_expression) & SU_NODE)
			{
			if((sym = lookup_node(RSLINK,NS_VAR)) == NULL)
				fatal("Compiler error: doreturn");
			if(return_expression->operator == OP_CALL)
				{
				/* Avoid duplicate copying of structure */
				modify_rslink(return_expression,
					lterminal(sym));
				expr_out(return_expression,NEED_VALUE);
				}
			else
				{
				expr_out(return_expression,
					(NEED_VALUE | EVAL_ADDR));
				size = return_expression->node_type->
					val.integral;
				if(size <= Sizeof_int)
					abs_oper(TOP_LDNL,(INT32) 0);
				instruction(TOP_LDL,ABS,WORKSPACE,NULL,
					-(word(-sym->address.offset)),
#ifdef	DEBUG
				(Debug ? "struct/union function return ptr" :
					NULL)
#else
						 NULL
#endif	/* DEBUG */
							);
				if(size <= Sizeof_int)
					abs_oper(TOP_STNL,(INT32) 0);
				else
					{
					loadconstant(size);
					no_oper(TOP_MOVE);
					}
				}
			instruction(TOP_LDL,ABS,WORKSPACE,NULL,
				-(word(-sym->address.offset)),
#ifdef	DEBUG
			(Debug ? "struct/union function return ptr" : NULL)
#else
					 NULL
#endif	/* DEBUG */
						);
			}
		else
			expr_out(return_expression,NEED_VALUE);
		}
	else
		{
		if(! same_type(Voidtype,Func_type,/* ptr_eq_array = */ FALSE,
			/* void_ptr_ok = */ FALSE))
			{
			warning(2,"Inconsistent return type");
			}
		}
	ns();

	if(Traceout)			/* Do function exit trace fn	*/
		instruction(TOP_CALL,REL,Troname,NULL,(INT32) 0,NULL);

	instruction(TOP_RETF,ABS,WORKSPACE,NULL,(INT32) 0,NULL);
	}

/*
 *	Process the switch statement.
 */
PRIVATE VOID
doswitch(locptr)
	INT32 locptr;
	{
	BOOL			cdup;
	char			lblbuf[NAMEMAX + 1];
	struct	switem		*sl;
	int			sw_count;
	struct	swheader	switch_header;
	struct	dvrt		*switch_body;
	TEMP			temp;

	needpunc('(');
	switch_header.sh_operand = expression();
	needpunc(')');

	switch_header.sh_swqueue.exit = getlabel(); /* Exit (break) label */
	switch_header.sh_swqueue.loop =
		((Swqueue != NULL) ? Swqueue->loop : (LABEL) 0);

	addswq(&switch_header.sh_swqueue);	/* Add sw/loop queue elem */
	switch_header.sh_default = (LABEL) 0;
	switch_header.sh_caselist = NULL;	/* Case list */
	addswheader(&switch_header);		/* Add to switch list */

	if(! ISINTEGRAL(switch_header.sh_operand))
		error("Switch expression not integral type");

	dvrt_stack();
	statement(locptr);		/* Get switch body */
	switch_body = dvrt_unstack();

	expr_out(switch_header.sh_operand,NEED_VALUE);
/*
 *	Figure out what template to generate the switch statement with
 *	depending on whether we have 3 or more cases (use "C" register self
 *	duplication feature), or less (use distributed re-fetch of switch
 *	value).
 */
	for(sw_count = 0, sl = switch_header.sh_caselist;
		((sl != NULL) && (sw_count < 4)); sw_count++, sl = sl->sw_next)
		;
	if(sw_count < 4)
		cdup = FALSE;		/* Use distributed re-fetch scheme */
	else
		cdup = TRUE;		/* Use "C" register auto-dup feature */

	if(cdup)
		{			/* Fill all registers */
		if(CPU_DUP)
			{
			no_oper(TOP_DUP);	/* Set value of C register */
			no_oper(TOP_DUP);
			temp = 0;	/* For lint */
			}
		else
			{
			temp = get_temp();
			store_temp(Inttype,temp);
			load_temp(Inttype,temp);
			load_temp(Inttype,temp);
			load_temp(Inttype,temp);	/* Set value of C */
			}
		}
	else
		{		/* Fill only A reg. (+temp if no dup) */
		if(CPU_NO_DUP && (sw_count > 1))
			{
			temp = get_temp();
			store_temp(Inttype,temp);
			}
		else
			temp = 0;	/* For lint */
		}
/*
 *	For future versions, we could sort the case list in a tree order to
 *	build a binary-executable tree to select the appropriate case label.
 *	For now, take the simple (cowardly) way out.
 */
	/* Emit simple compare/jump clauses */
	for(sl = switch_header.sh_caselist; sl != NULL; sl = sl->sw_next)
		{
		if(! cdup)
			{
			if(CPU_NO_DUP && (sw_count > 1))
				load_temp(Inttype,temp);
			else
				{
				if(sl->sw_next != NULL)
					no_oper(TOP_DUP);
				}
			}
		if(sl->sw_value != (INT32) 0)
			abs_oper(TOP_ADC,(INT32) -sl->sw_value);
		instruction(TOP_CJ,REL,strlabel(sl->sw_label,lblbuf),NULL,
			(LABEL) 0, NULL);
		}
	if(switch_header.sh_default != (LABEL) 0)
		{
		instruction(TOP_J,REL,
			strlabel(switch_header.sh_default,lblbuf),
				NULL,(INT32) 0,NULL);
		}
	else
		{
		instruction(TOP_J,REL,
			strlabel(switch_header.sh_swqueue.exit,lblbuf),
				NULL,(INT32) 0,NULL);
		}

	if(CPU_NO_DUP && (sw_count > 1))
		free_temp(temp);

	dvrt_append(switch_body);	/* Output body of switch statement */
	padlabel(switch_header.sh_swqueue.exit);
	delswq();			/* Delete the active sw/loop element */
	delswheader();			/* And the switch header */
	}

/*
 *	Process the while... statement.
 */
PRIVATE VOID
dowhile(locptr)
	INT32			locptr;
	{
	char			lblbuf[NAMEMAX + 1];
	struct	dvrt		*while_body;
	struct	op_node		*while_expression;
	struct	swq		while_queue;	/* Element for sw/loop queue */

	while_queue.loop = getlabel();
	while_queue.exit = getlabel();	/* Add labels to queue element */

	addswq(&while_queue);		/* Add this level to while queue */

	while_expression = test_expression();

	lineline(Lineno);

	dvrt_stack();
	statement(locptr);		/* Capture body of while body */
	while_body = dvrt_unstack();
/*
 *	Test for a FALSE constant expression "while".
 */
	if(ISCONSTANT(while_expression) &&
		(while_expression->right.ptr.terminal->val.integral == 0L))
		{
		dvrt_free(while_body);
		free_op_node(while_expression);
		}
/*
 *	The following code handles the general case of the while as well as
 *	the "while(TRUE)" case.
 */
	else
		{
		padlabel(while_queue.loop);
		if(! ISCONSTANT(while_expression))
			{
			expr_out(while_expression,(NEED_VALUE | RELAX_BOOL));
			instruction(TOP_CJ,REL,
				strlabel(while_queue.exit,lblbuf),NULL,
					(INT32) 0,NULL);
			}
		dvrt_append(while_body);
		instruction(TOP_J,REL,strlabel(while_queue.loop,lblbuf),NULL,
			(INT32) 0,NULL);
		padlabel(while_queue.exit);
		}

	delswq();
	}

/*
 *	Syntax requires a semicolon here.
 */
PUBLIC	VOID
ns()
	{
	if(! matchc(';'))
		error("Missing semicolon");
	}

/*
 *	Evaluate a expression for a boolean result.  If expression is a
 *	floating point constant convert to the equivalent integral value.
 *	If a floating point expression paste on a != 0.0
 */
PRIVATE	struct	op_node	*
bool_expr()
	{
	struct	op_node		*op;

	op = expression();
	if(EXPRTYPE(op) & FP_NODE)
		{
		if(ISCONSTANT(op))
			{
			if(op->right.ptr.terminal->val.fvalue != 0.0)
				{
				free_op_node(op);
				op = make_const((INT32) 1,Inttype);
				}
			else
				{
				free_op_node(op);
				op = make_const((INT32) 0,Inttype);
				}
			}
		else
			{
			op = binaryop((int) OP_NE,op,
				make_fconst(0.0,op->node_type));
			}
		}
	return (op);
	}

/*
 *	Recognize a "(", boolean expression, and ")".
 */
PRIVATE	struct	op_node	*
test_expression()
	{
	struct	op_node		*op;

	needpunc('(');
	op = bool_expr();
	needpunc(')');

	return (op);
	}

/* End of TCX3.C */
