/*
 *	TCX Compiler (4 of 11)
 *
 *	THIS FREE SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 *	WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *	SPECIFICALLY DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR OTHER
 *	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *	TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 *	OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *	THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 *	USE OF THIS SOFTWARE OR ANY DERIVATIVE WORK, EVEN IF ADVISED OF THE
 *	POSSIBILITY OF SUCH DAMAGE.
 *
 *	This program is hereby placed in the public domain.  In
 *	contrast to other claims of "public domain", this means no
 *	copyright is claimed and you may do anything you like with TCX,
 *	including selling it!  As a gesture of courtesy, please retain
 *	the authorship information in the source code and documentation.
 *
 *	Contact:	Kirk Bailey (bailey@ieee.org)
 *			P.O. Box 1702
 *			Corvallis, OR 97339 (USA)
 *
 *	03/17/87
 *
 *	Each level of the expression evaluator (except primary()) call
 *	succeeding lower levels, until a primary is detected.  Upon
 *	backtracking out, each level detects its own set of operators, and
 *	binds together op_nodes that indicate the operator and the one or
 *	two operands for each.
 */

#include	<stdio.h>
#include	<stdlib.h>
#include	<ctype.h>
#include	<string.h>
#include	"tcx.h"
#include	"tcxdecl.h"
#include	"tcxext.h"

/*
 *	Local function declarations.
 */
PRIVATE	VOID			const_add();
PRIVATE	VOID			const_sub();
PRIVATE	VOID			const_mul();
PRIVATE	VOID			const_div();
PRIVATE	VOID			const_eq();
PRIVATE	VOID			const_ne();
PRIVATE	VOID			const_lt();
PRIVATE	VOID			const_le();
PRIVATE	VOID			const_gt();
PRIVATE	VOID			const_ge();
PRIVATE	VOID			const_umin();
PRIVATE	VOID			const_lnot();
PRIVATE	VOID			const_cnvt();
PRIVATE	BOOL			lvalue();
PRIVATE	struct	id_node		*new_type();
PRIVATE	struct	id_node		*select_type();
PRIVATE	struct	op_node		*hier1();
PRIVATE	int			matchassign();
PRIVATE	VOID			boolify();
PRIVATE	struct	op_node		*hier4();
PRIVATE	struct	op_node		*hier5();
PRIVATE	int			matchor();
PRIVATE	struct	op_node		*hier6();
PRIVATE	int			matchxor();
PRIVATE	struct	op_node		*hier7();
PRIVATE	int			matchand();
PRIVATE	struct	op_node		*hier8();
PRIVATE	int			matcheqlop();
PRIVATE	struct	op_node		*hier9();
PRIVATE	int			matchrelop();
PRIVATE	struct	op_node		*hier10();
PRIVATE	int			matchshift();
PRIVATE	struct	op_node		*hier11();
PRIVATE	int			matchadd();
PRIVATE	struct	op_node		*hier12();
PRIVATE	int			matchmult();
PRIVATE	struct	op_node		*hier13();
PRIVATE	struct	op_node		*hier14();
PRIVATE	struct	op_node		*hier15();
PRIVATE	struct	op_node		*param_help();
PRIVATE	struct	op_node		*parameters();
PRIVATE	struct	id_node		*ismember();
PRIVATE	struct	op_node		*primary();
PRIVATE	BOOL			needlvalue();
PRIVATE	VOID			struct_error();
PRIVATE	struct	op_node		*get_op_node();

/*
 *	Local storage declarations.
 */
PRIVATE	struct op_node	*free_oplist =	NULL;	/* Chain of free op_nodes */

#ifdef	DEBUG
PUBLIC	int		Alloc_oplist =	0;	/* Total number allocated */
PUBLIC	int		Nfree_oplist =	0;	/* Total currently free	  */
#endif	/* DEBUG */

#define	islvalue(op)	(op->lvalue)

/*
 *	Handle combining binary operators.
 */

#ifdef	MSC_OPT
#pragma	optimize("e",off)		/* Disable global reg optimizing */
#pragma	optimize("g",off)		/* Disable global common subs */
#pragma	optimize("l",off)		/* Disable loop optimizations */
#endif	/* MSC_OPT */

PUBLIC	struct	op_node	*
binaryop(op,lop,rop)
	int			op;
	struct	op_node		*lop;
	struct	op_node		*rop;
	{
	register struct	op_node	*newop;
	struct	id_node		*newtype;
	struct	op_node		*tempop;
	union	value		value;

	newop = NULL;			/* Not assigned yet */

	if(((lop == NULL) || ISCONSTANT(lop) || (op == OP_CONVERT)) &&
		((rop == NULL) || ISCONSTANT(rop)) &&
		(op != OP_TSEL) &&
		(op != OP_MEMBER) &&
		(op != OP_INDIRECT) &&
		(op != OP_PARAM))
		{
/*
 *	Note that constants cannot use the assignment operators.
 */
		switch((unsigned int) op)
			{
/*
 *	Dyadic operators.
 */
			case	OP_MUL:
				const_mul(lop,rop,&value);
				break;

			case	OP_DIV:
				const_div(lop,rop,&value);
				break;

			case	OP_MOD:
				if((EXPRTYPE(lop) & FP_NODE) ||
					(EXPRTYPE(rop) & FP_NODE))
					{
					error("\"%\" invalid for float types");
					value.integral = 0;
					}
				else
					{
					if(rop->right.ptr.terminal->
						val.integral == 0)
						{
						error("Division by zero");
						value.integral = 0;
						}
					else if(ISSIGNED(lop) || ISSIGNED(rop))
						{
						value.integral =
					lop->right.ptr.terminal->val.integral %
					rop->right.ptr.terminal->val.integral;
						}
					else
						{
						value.uintegral =
					lop->right.ptr.terminal->val.integral %
					rop->right.ptr.terminal->val.integral;
						}
					}
				break;

			case	OP_LSHIFT:
				if((EXPRTYPE(lop) & FP_NODE) ||
					(EXPRTYPE(rop) & FP_NODE))
					{
				error("\"<<\" invalid for float types");
					value.integral = 0;
					}
				else
					{
					value.integral =
				lop->right.ptr.terminal->val.integral <<
				rop->right.ptr.terminal->val.integral;
					}
				break;

			case	OP_RSHIFT:
				if((EXPRTYPE(lop) & FP_NODE) ||
					(EXPRTYPE(rop) & FP_NODE))
					{
				error("\">>\" invalid for float types");
					value.integral = 0;
					}
				else if(ISUNSIGNED(lop) || (! Cpu_srshift) ||
				(lop->right.ptr.terminal->val.integral >= 0))
					{
/*
 *	If unsigned or signed shifts are unsigned, zero fill.
 */
#if	H_OPSYS != XENIX	/* Xenix compiler is a bit flimsy */
					value.integral =
				(lop->right.ptr.terminal->val.integral >>
				rop->right.ptr.terminal->val.integral) &
				((1L <<
				(32 - rop->right.ptr.terminal->val.integral))
						- 1);
#else
						{
						UINT32	tmpval;
						INT32	targ =
					rop->right.ptr.terminal->val.integral;

						tmpval =
						((1L << (32 - targ)) - 1);
						value.integral =
				(lop->right.ptr.terminal->val.integral >>
				rop->right.ptr.terminal->val.integral) &
							tmpval;
						}
#endif	/* H_OPSYS != XENIX */
					}
				else
					{
/*
 *	Signed right shifts -- extend sign forcibly.
 */
#if	H_OPSYS != XENIX	/* Xenix compiler still a bit flimsy */
					value.integral =
				(lop->right.ptr.terminal->val.integral >>
				rop->right.ptr.terminal->val.integral) |
				(- (1L <<
				(32 - rop->right.ptr.terminal->val.integral)));
#else
						{
						UINT32	tmpval;
						INT32	targ =
					rop->right.ptr.terminal->val.integral;

						tmpval =
						(- (1L << (32 - targ)));
						value.integral =
				(lop->right.ptr.terminal->val.integral >>
				rop->right.ptr.terminal->val.integral) |
							tmpval;
						}
#endif	/* H_OPSYS != XENIX */
					}
				break;

			case	OP_ADD:
				const_add(lop,rop,&value);
				break;

			case	OP_SUB:
				const_sub(lop,rop,&value);
				break;

			case	OP_BAND:
				if((EXPRTYPE(lop) & FP_NODE) ||
					(EXPRTYPE(rop) & FP_NODE))
					{
					error("\"&\" invalid for float types");
					value.integral = 0;
					}
				else
					{
					value.integral =
					lop->right.ptr.terminal->val.integral &
					rop->right.ptr.terminal->val.integral;
					}
				break;

			case	OP_BXOR:
				if((EXPRTYPE(lop) & FP_NODE) ||
					(EXPRTYPE(rop) & FP_NODE))
					{
					error("\"^\" invalid for float types");
					value.integral = 0;
					}
				else
					{
					value.integral =
					lop->right.ptr.terminal->val.integral ^
					rop->right.ptr.terminal->val.integral;
					}
				break;

			case	OP_BOR:
				if((EXPRTYPE(lop) & FP_NODE) ||
					(EXPRTYPE(rop) & FP_NODE))
					{
					error("\"|\" invalid for float types");
					value.integral = 0;
					}
				else
					{
					value.integral =
					lop->right.ptr.terminal->val.integral |
					rop->right.ptr.terminal->val.integral;
					}
				break;

			case	OP_SEQ:
				/* Just pass back the next expression */
				(VOID) memcpy((char *) &value,
					(char *) &rop->right.ptr.terminal->val,
						sizeof(union value));
				break;

			case	OP_EQ:
				const_eq(lop,rop,&value);
				break;

			case	OP_NE:
				const_ne(lop,rop,&value);
				break;

			case	OP_LT:
				const_lt(lop,rop,&value);
				break;

			case	OP_LE:
				const_le(lop,rop,&value);
				break;

			case	OP_GT:
				const_gt(lop,rop,&value);
				break;

			case	OP_GE:
				const_ge(lop,rop,&value);
				break;
/*
 *	Monadic operators.
 */
			case	OP_BNOT:
				if(EXPRTYPE(rop) & FP_NODE)
					{
					error("\"~\" invalid for float types");
					value.integral = 0;
					}
				else
					{
					value.uintegral = ~
					rop->right.ptr.terminal->val.uintegral;
					}
				break;

			case	OP_UMIN:
				const_umin(NULL,rop,&value);
				break;

			case	OP_UPLUS:
				(VOID) memcpy((char *) &value,
					(char *) &rop->right.ptr.terminal->val,
						sizeof(union value));
				break;

			case	OP_LNOT:
				const_lnot(NULL,rop,&value);
				break;

			case	OP_CONVERT:
				const_cnvt(lop,rop,&value);
				break;

			default:
				error("Invalid constant expression");
				value.integral = 0;
				break;
			}
/*
 *	Make a new constant.
 */
		newtype = new_type(op,lop,rop);
		if(newtype->type & FP_NODE)
			newop = make_fconst(value.fvalue, newtype);
		else
			newop = make_const(value.integral, newtype);

		if(lop != NULL)
			free_op_node(lop);

		if(rop != NULL)
			free_op_node(rop);	/* Don't need the right op */
		}
	else
		{
/*
 *	Generate an error if:
 *		1) Left operand is a union/struct and operator is not
 *			a) Assign
 *			b) Member-of
 *			c) Offset (internal Member-of)
 *			d) Ternary select (:)
 *			e) Ternary (?)
 *			f) Sequence (,)
 *			g) Function call
 *			h) Parameter
 *			i) Convert (illegal but caught in "new_type")
 *		or
 *		2) Right operand is a union/struct and operator is not
 *			a) Assign
 *			b) Member-of
 *			c) Offset
 *			d) Address-of
 *			e) Ternary select (:)
 *			f) Ternary (?)
 *			g) Sequence (,)
 *			h) Parameter
 *			i) Convert (illegal but caught in "new_type")
 */
		if(((lop != NULL) &&
			(EXPRTYPE(lop) & SU_NODE) &&
			((op != OP_ASSIGN) && (op != OP_MEMBER) &&
			(op != OP_OFFSET) && (op != OP_TSEL) &&
			(op != OP_TERN) && (op != OP_SEQ) &&
			(op != OP_CALL) && (op != OP_PARAM) &&
			(op != OP_CONVERT))) ||
		((rop != NULL) &&
			(EXPRTYPE(rop) & SU_NODE) &&
			((op != OP_ASSIGN) && (op != OP_MEMBER) &&
			(op != OP_OFFSET) && (op != OP_ADDRESS) &&
			(op != OP_TSEL) && (op != OP_TERN) &&
			(op != OP_SEQ) && (op != OP_PARAM) &&
			(op != OP_CONVERT))))
			{
			error("Invalid operation on struct/union");
			}
/*
 *	Disallow assignment between structures/unions that are
 *	not the same type/size.  Also note that if the right operand is
 *	the result of a function call no explicit assignment needs to be
 *	done if the return struct/union link pointer is monkeyed with
 *	appropriately.
 */
		if((op == OP_ASSIGN) &&
			((EXPRTYPE(lop) | EXPRTYPE(rop)) & SU_NODE))
			{
			if(! same_type(lop->node_type,rop->node_type,
				/* ptr_eq_array = */ FALSE,
					/* void_ptr_ok = */ FALSE))
				{
				error("Incompatible struct/union assignment");
				}
			else if(rop->operator == OP_CALL)
				{
/*
 *	First make sure that the return value monkeying hasn't
 *	already been done at a lower level of struct assignment.
 */
				if(((tempop = LEFT(rop)) != NULL) &&
					((tempop = RIGHT(tempop)) != NULL) &&
					(ISTERMINAL(tempop)) &&
					(tempop->right.ptr.terminal->
						info.storage_class ==
					SC_RSTEMP))
					{
/*
 *	Link in a address-of operator so that the address of the
 *	destination gets computed.
 */
					modify_rslink(rop,
						binaryop(OP_ADDRESS,NULL,lop));
					newop = rop;
					}
				}
			}
		if((op & OP_COMM) && (! (op & OP_AOP)))
			{
/*
 *	If a constant operand put in right slot for commutative ops.
 */
			if(ISCONSTANT(lop))
				{
				tempop = rop;
				rop = lop;
				lop = tempop;
				}
			}
		if((op == (int) OP_ADD) || (op == (int) OP_AADD) ||
			(op == OP_POSTADD) ||
			(((op == OP_SUB) || (op == OP_ASUB) ||
				(op == OP_POSTSUB)) &&
			isptrclass(lop) && !isptrclass(rop)))
			{
			if(isptrclass(rop))
				{
/*
 *	Interchange to keep pointer in lop.
 */
				tempop = rop;
				rop = lop;
				lop = tempop;
				}
			if(isptrclass(lop))
				{
/*
 *	Adding/Subtracting to a pointer.  Right must be value.  Produce
 *	scaling node.
 */
				if(lop->node_type->type_ptr.base_type->type &
					FUNC_NODE)
					{
					error("Math on pointer-to-func()");
					}
				if(lop->node_type->type_ptr.base_type->
					val.integral ==	((INT32) 0))
					{
				error("Size of pointer object is unknown");
					}
				rop = binaryop((int) OP_MUL,rop,
					make_const(lop->node_type->
						type_ptr.base_type->
					val.integral,Inttype));
				}
			}
		else if(op == OP_MEMBER)
			{
/*
 *	See if membering an already membered item.
 */
			if(lop->operator == OP_MEMBER)
				{
/*
 *	Just add offset into member constant already created.
 */
				RIGHT(lop)->right.ptr.terminal->val.integral +=
					rop->right.ptr.terminal->
						address.offset;

				newop = lop;	/* Just use left op as final */
				}
			else
				{
/*
 *	First member-of applied to a node.
 */
				newop = get_op_node();
				newop->operator = op;
				LEFT(newop) = lop;

				RIGHT(newop) = 
					make_const(rop->right.ptr.terminal->
						address.offset,
					Inttype);
				}
			newop->node_type = new_type(op,lop,rop);
			free_op_node(rop);	/* Get rid of right */
			newop->lvalue = lvalue(op,newop);
			}
		if(((lop != NULL) && (EXPRTYPE(lop) & FP_NODE)) ||
			((rop != NULL) && (EXPRTYPE(rop) & FP_NODE)))
			{
			switch ((unsigned int) (op & (~OP_AOP)))
				{
/*
 *	Dyadic operators.
 */
				case	OP_MOD:
					error("\"%\" invalid for float types");
					break;

				case	OP_LSHIFT:
				error("\"<<\" invalid for float types");
					break;

				case	OP_RSHIFT:
				error("\">>\" invalid for float types");
					break;

				case	OP_BAND:
					error("\"&\" invalid for float types");
					break;

				case	OP_BXOR:
					error("\"^\" invalid for float types");
					break;

				case	OP_BOR:
					error("\"|\" invalid for float types");
					break;
/*
 *	Monadic operators.
 */
				case	OP_BNOT:
					error("\"~\" invalid for float types");
					break;

				default:
					break;
				}
			}
		if(newop == NULL)
			{
			newop = get_op_node();

			newop->operator = op;
			LEFT(newop) = lop;
			RIGHT(newop) = rop;
			newop->node_type = new_type(op,lop,rop);
			newop->lvalue = lvalue(op,newop);
			}

		if(op == OP_CONVERT)
			{
/*
 *	Convert uses the left pointer to import the type for casting.  We
 *	don't need it anymore, so free that lower node.
 */
			LEFT(newop) = NULL;
			free_op_node(lop);
			lop = NULL;
			}
		else if((op == OP_SUB) && isptrclass(rop) && isptrclass(lop))
			{
/*
 *	Do the subtract and then bind a divide by size.
 */

			newop = binaryop(OP_DIV,newop,
				make_const(lop->node_type->
					type_ptr.base_type->val.integral,
						Inttype));
			}
		}
	return (newop);
	}

#ifdef	MSC_OPT
#pragma	optimize("e",on)		/* Enable global reg optimizing */
#pragma	optimize("g",on)		/* Enable global common subs */
#pragma	optimize("l",on)		/* Enable loop optimizing */
#endif	/* MSC_OPT */

/*
 *	Perform constant addition.
 */
PRIVATE VOID
const_add(lop,rop,results)
	struct	op_node		*lop;
	struct	op_node		*rop;
	union	value		*results;
	{
	double			leftf;
	double			rightf;
	struct	op_node		*tempop;

	if(isptrclass(rop))
		{
		/* Interchange to keep pointer in lop */
		tempop = rop;
		rop = lop;
		lop = tempop;
		}
	if((EXPRTYPE(lop) & FP_NODE) || (EXPRTYPE(rop) & FP_NODE))
		{
/*
 *	Perform calculation in floating point and yield a float.
 */
		if(EXPRTYPE(lop) & FP_NODE)
			leftf = lop->right.ptr.terminal->val.fvalue;
		else
			leftf = l_to_f(lop);

		if(EXPRTYPE(rop) & FP_NODE)
			rightf = rop->right.ptr.terminal->val.fvalue;
		else
			rightf = l_to_f(rop);

		results->fvalue = leftf + rightf;
		}
	else
		{
		results->integral = lop->right.ptr.terminal->val.integral +
			((isptrclass(lop) ? lop->node_type->
			type_ptr.base_type->val.integral : 1) *
			rop->right.ptr.terminal->val.integral);
		}
	}

/*
 *	Perform constant subtraction.
 */
PRIVATE	VOID
const_sub(lop,rop,results)
	struct	op_node		*lop;
	struct	op_node		*rop;
	union	value		*results;
	{
	double			leftf;
	double			rightf;

	if((EXPRTYPE(lop) & FP_NODE) || (EXPRTYPE(rop) & FP_NODE))
		{
/*
 *	Perform calculation in floating point and yield a float.
 */
		if(EXPRTYPE(lop) & FP_NODE)
			leftf = lop->right.ptr.terminal->val.fvalue;
		else
			leftf = l_to_f(lop);

		if(EXPRTYPE(rop) & FP_NODE)
			rightf = rop->right.ptr.terminal->val.fvalue;
		else
			rightf = l_to_f(rop);

		results->fvalue = leftf - rightf;
		}
	else
		{
		if(isptrclass(lop))
			{
			if(isptrclass(rop))
				{
				results->integral = (lop->right.ptr.terminal->
					val.integral - rop->
					right.ptr.terminal->val.integral) /
					lop->node_type->type_ptr.base_type->
					val.integral;
				}
			else
				{
				results->integral =
					lop->right.ptr.terminal->val.integral -
					(lop->node_type->type_ptr.base_type->
					val.integral *
					rop->right.ptr.terminal->val.integral);
				}
			}
		else
			{
			results->integral =
				lop->right.ptr.terminal->val.integral -
					rop->right.ptr.terminal->val.integral;
			}
		}
	}

/*
 *	Perform constant multiplication.
 */
PRIVATE	VOID
const_mul(lop,rop,results)
	struct	op_node		*lop;
	struct	op_node		*rop;
	union	value		*results;
	{
	double			leftf;
	double			rightf;

	if((EXPRTYPE(lop) & FP_NODE) || (EXPRTYPE(rop) & FP_NODE))
		{
/*
 *	Perform calculation in floating point and yield a float.
 */
		if(EXPRTYPE(lop) & FP_NODE)
			leftf = lop->right.ptr.terminal->val.fvalue;
		else
			leftf = l_to_f(lop);

		if(EXPRTYPE(rop) & FP_NODE)
			rightf = rop->right.ptr.terminal->val.fvalue;
		else
			rightf = l_to_f(rop);

		results->fvalue = leftf * rightf;
		}
	else
		{
		results->integral = lop->right.ptr.terminal->val.integral *
			rop->right.ptr.terminal->val.integral;
		}
	}

/*
 *	Perform constant division.
 */
PRIVATE	VOID
const_div(lop,rop,results)
	struct	op_node		*lop;
	struct	op_node		*rop;
	union	value		*results;
	{
	double			leftf;
	double			rightf;

	if((EXPRTYPE(lop) & FP_NODE) || (EXPRTYPE(rop) & FP_NODE))
		{
/*
 *	Perform calculation in floating point and yield a float.
 */
		if(EXPRTYPE(lop) & FP_NODE)
			leftf = lop->right.ptr.terminal->val.fvalue;
		else
			leftf = l_to_f(lop);

		if(EXPRTYPE(rop) & FP_NODE)
			rightf = rop->right.ptr.terminal->val.fvalue;
		else
			rightf = l_to_f(rop);

		if(rightf == 0.0)
			{
			error("Division by zero");
			results->fvalue = 0.0;
			}
		else
			results->fvalue = leftf / rightf;
		}
	else
		{
		if(rop->right.ptr.terminal->val.integral == 0)
			{
			error("Division by zero");
			results->integral = 0;
			}
		else if(ISSIGNED(lop) && ISSIGNED(rop))
			{
			results->integral = lop->right.ptr.terminal->
				val.integral /
					rop->right.ptr.terminal->val.integral;
			}
		else
			{
			results->integral = lop->right.ptr.terminal->
				val.uintegral /
					rop->right.ptr.terminal->val.uintegral;
			}
		}
	}

/*
 *	Perform constant equality comparison.
 */
PRIVATE VOID
const_eq(lop,rop,results)
	struct	op_node		*lop;
	struct	op_node		*rop;
	union	value		*results;
	{
	double			leftf;
	double			rightf;

	if((EXPRTYPE(lop) & FP_NODE) || (EXPRTYPE(rop) & FP_NODE))
		{
/*
 *	Perform calculation in floating point and yield a float.
 */
		if(EXPRTYPE(lop) & FP_NODE)
			leftf = lop->right.ptr.terminal->val.fvalue;
		else
			leftf = l_to_f(lop);

		if(EXPRTYPE(rop) & FP_NODE)
			rightf = rop->right.ptr.terminal->val.fvalue;
		else
			rightf = l_to_f(rop);

		results->integral = leftf == rightf;
		}
	else
		{
		results->integral = lop->right.ptr.terminal->val.uintegral ==
			rop->right.ptr.terminal->val.uintegral;
		}
	}

/*
 *	Perform constant inequality comparison.
 */
PRIVATE	VOID
const_ne(lop,rop,results)
	struct	op_node		*lop;
	struct	op_node		*rop;
	union	value		*results;
	{
	double			leftf;
	double			rightf;

	if((EXPRTYPE(lop) & FP_NODE) || (EXPRTYPE(rop) & FP_NODE))
		{
/*
 *	Perform calculation in floating point and yield a float.
 */
		if(EXPRTYPE(lop) & FP_NODE)
			leftf = lop->right.ptr.terminal->val.fvalue;
		else
			leftf = l_to_f(lop);

		if(EXPRTYPE(rop) & FP_NODE)
			rightf = rop->right.ptr.terminal->val.fvalue;
		else
			rightf = l_to_f(rop);

		results->integral = leftf != rightf;
		}
	else
		{
		results->integral = lop->right.ptr.terminal->val.uintegral !=
			rop->right.ptr.terminal->val.uintegral;
		}
	}

/*
 *	Perform constant less than comparison.
 */
PRIVATE	VOID
const_lt(lop,rop,results)
	struct	op_node		*lop;
	struct	op_node		*rop;
	union	value		*results;
	{
	double			leftf;
	double			rightf;

	if((EXPRTYPE(lop) & FP_NODE) || (EXPRTYPE(rop) & FP_NODE))
		{
/*
 *	Perform calculation in floating point and yield a float.
 */
		if(EXPRTYPE(lop) & FP_NODE)
			leftf = lop->right.ptr.terminal->val.fvalue;
		else
			leftf = l_to_f(lop);

		if(EXPRTYPE(rop) & FP_NODE)
			rightf = rop->right.ptr.terminal->val.fvalue;
		else
			rightf = l_to_f(rop);

		results->integral = leftf < rightf;
		}
	else if(ISSIGNED(lop))
		{
		if(ISSIGNED(rop))
			{
			results->integral =
				lop->right.ptr.terminal->val.integral <
					rop->right.ptr.terminal->val.integral;
			}
		else
			{
			results->integral = (UINT32)
				lop->right.ptr.terminal->val.integral <
					rop->right.ptr.terminal->val.uintegral;
			}
		}
	else
		{
		if(ISSIGNED(rop))
			{
			results->integral =
				lop->right.ptr.terminal->val.uintegral <
					(UINT32) rop->right.ptr.terminal->
						val.integral;
			}
		else
			{
			results->integral =
				lop->right.ptr.terminal->val.integral <
					rop->right.ptr.terminal->val.integral;
			}
		}
	}

/*
 *	Perform constant less than or equal comparison.
 */
PRIVATE	VOID
const_le(lop,rop,results)
	struct	op_node		*lop;
	struct	op_node		*rop;
	union	value		*results;
	{
	double			leftf;
	double			rightf;

	if((EXPRTYPE(lop) & FP_NODE) || (EXPRTYPE(rop) & FP_NODE))
		{
/*
 *	Perform calculation in floating point and yield a float.
 */
		if(EXPRTYPE(lop) & FP_NODE)
			leftf = lop->right.ptr.terminal->val.fvalue;
		else
			leftf = l_to_f(lop);

		if(EXPRTYPE(rop) & FP_NODE)
			rightf = rop->right.ptr.terminal->val.fvalue;
		else
			rightf = l_to_f(rop);

		results->integral = leftf <= rightf;
		}
	else if(ISSIGNED(lop))
		{
		if(ISSIGNED(rop))
			{
			results->integral =
				lop->right.ptr.terminal->val.integral <=
					rop->right.ptr.terminal->val.integral;
			}
		else
			{
			results->integral = (UINT32)
				lop->right.ptr.terminal->val.integral<=
					rop->right.ptr.terminal->val.uintegral;
			}
		}
	else
		{
		if(ISSIGNED(rop))
			{
			results->integral =
				lop->right.ptr.terminal->val.uintegral <=
					(UINT32) rop->right.ptr.terminal->
						val.integral;
			}
		else
			{
			results->integral =
				lop->right.ptr.terminal->val.integral <=
					rop->right.ptr.terminal->val.integral;
			}
		}
	}

/*
 *	Perform constant greater than comparison.
 */
PRIVATE	VOID
const_gt(lop,rop,results)
	struct	op_node		*lop;
	struct	op_node		*rop;
	union	value		*results;
	{
	double			leftf;
	double			rightf;

	if((EXPRTYPE(lop) & FP_NODE) || (EXPRTYPE(rop) & FP_NODE))
		{
/*
 *	Perform calculation in floating point and yield a float.
 */
		if(EXPRTYPE(lop) & FP_NODE)
			leftf = lop->right.ptr.terminal->val.fvalue;
		else
			leftf = l_to_f(lop);

		if(EXPRTYPE(rop) & FP_NODE)
			rightf = rop->right.ptr.terminal->val.fvalue;
		else
			rightf = l_to_f(rop);

		results->integral = leftf > rightf;
		}
	else if(ISSIGNED(lop))
		{
		if(ISSIGNED(rop))
			{
			results->integral =
				lop->right.ptr.terminal->val.integral >
					rop->right.ptr.terminal->val.integral;
			}
		else
			{
			results->integral = (UINT32)
				lop->right.ptr.terminal->val.integral >
					rop->right.ptr.terminal->val.uintegral;
			}
		}
	else
		{
		if(ISSIGNED(rop))
			{
			results->integral =
				lop->right.ptr.terminal->val.uintegral >
					(UINT32) rop->right.ptr.terminal->
						val.integral;
			}
		else
			{
			results->integral =
				lop->right.ptr.terminal->val.integral >
					rop->right.ptr.terminal->val.integral;
			}
		}
	}

/*
 *	Perform constant greater than equal comparison.
 */
PRIVATE	VOID
const_ge(lop,rop,results)
	struct	op_node		*lop;
	struct	op_node		*rop;
	union	value		*results;
	{
	double			leftf;
	double			rightf;

	if((EXPRTYPE(lop) & FP_NODE) || (EXPRTYPE(rop) & FP_NODE))
		{
/*
 *	Perform calculation in floating point and yield a float.
 */
		if(EXPRTYPE(lop) & FP_NODE)
			leftf = lop->right.ptr.terminal->val.fvalue;
		else
			leftf = l_to_f(lop);

		if(EXPRTYPE(rop) & FP_NODE)
			rightf = rop->right.ptr.terminal->val.fvalue;
		else
			rightf = l_to_f(rop);

		results->integral = leftf >= rightf;
		}
	else if(ISSIGNED(lop))
		{
		if(ISSIGNED(rop))
			{
			results->integral =
				lop->right.ptr.terminal->val.integral >=
					rop->right.ptr.terminal->val.integral;
			}
		else
			{
			results->integral = (UINT32)
				lop->right.ptr.terminal->val.integral>=
					rop->right.ptr.terminal->val.uintegral;
			}
		}
	else
		{
		if(ISSIGNED(rop))
			{
			results->integral =
				lop->right.ptr.terminal->val.uintegral >=
					(UINT32) rop->right.ptr.terminal->
						val.integral;
			}
		else
			{
			results->integral =
				lop->right.ptr.terminal->val.integral >=
					rop->right.ptr.terminal->val.integral;
			}
		}
	}

/*
 *	Perform constant unary minus.
 */

#ifdef	__TURBOC__
#pragma	warn	-par
#endif	/* __TURBOC__ */

PRIVATE	VOID
const_umin(lop,rop,results)
	struct	op_node		*lop;
	struct	op_node		*rop;
	union	value		*results;
	{
	if(EXPRTYPE(rop) & FP_NODE)
		results->fvalue = -rop->right.ptr.terminal->val.fvalue;
	else
		results->integral = -rop->right.ptr.terminal->val.integral;
	}

#ifdef	__TURBOC__
#pragma	warn	.par
#endif	/* __TURBOC__ */

/*
 *	Perform constant logical not.
 */

#ifdef	__TURBOC__
#pragma	warn	-par
#endif	/* __TURBOC__ */

PRIVATE	VOID
const_lnot(lop,rop,results)
	struct	op_node		*lop;
	struct	op_node		*rop;
	union	value		*results;
	{
	if(EXPRTYPE(rop) & FP_NODE)
		results->integral = !rop->right.ptr.terminal->val.fvalue;
	else
		results->integral = !rop->right.ptr.terminal->val.integral;
	}

#ifdef	__TURBOC__
#pragma	warn	.par
#endif	/* __TURBOC__ */

/*
 *	Perform constant conversion.
 */
PRIVATE	VOID
const_cnvt(lop,rop,results)
	struct	op_node		*lop;
	struct	op_node		*rop;
	union	value		*results;
	{
	if(EXPRTYPE(lop) & FP_NODE)
		{
		if(EXPRTYPE(rop) & FP_NODE)
			results->fvalue = rop->right.ptr.terminal->val.fvalue;
		else
			results->fvalue = l_to_f(rop);
		}
	else if(EXPRTYPE(lop) & CHAR_NODE)
		{
		if(EXPRTYPE(rop) & FP_NODE)
			{
			results->uintegral = ((UINT32)
				(rop->right.ptr.terminal->val.fvalue)) & 0xFF;
			}
		else
			{
			results->uintegral = rop->right.ptr.terminal->
				val.uintegral & 0xFF;
			}
		if(ISSIGNED(lop))
			{
			if(results->uintegral > 127)
				results->uintegral |= ~0xFF;	/* Sign ext */
			}
		}
	else
		{
		if(EXPRTYPE(rop) & FP_NODE)
			{
			results->integral = f_to_l(rop->right.ptr.terminal->
				val.fvalue,TRUE);
			}
		else
			{
			results->integral = rop->right.ptr.terminal->
				val.integral;
			}
		}
	}

/*
 *	Return TRUE if node is an lvalue.
 */
PRIVATE	BOOL
lvalue(op,newop)
	int			op;
	struct	op_node		*newop;
	{
/*
 *	Choose if new op_node is an lvalue.
 */
	switch((int) op)
		{
		case OP_CONVERT:
			return (RIGHT(newop)->lvalue);
		case OP_INDIRECT:
			return (TRUE);
		case OP_MEMBER:
			return (LEFT(newop)->lvalue);
		}
	return (FALSE);
	}

/*
 *	Determine the type of an expression involving one or two
 *	operands and an operator.
 */

#ifdef	MSC_OPT
#pragma	optimize("e",off)		/* Disable global reg optimizing */
#endif	/* MSC_OPT */

PRIVATE struct	id_node	*
new_type(op,lop,rop)
	int			op;
	struct	op_node		*lop;
	struct	op_node		*rop;
	{
	BOOL			leftptr;
	BOOL			leftvp;
	struct id_node		*new;
	BOOL			rightptr;
	BOOL			rightvp;
	BOOL			vserror;

	leftptr = isptrclass(lop);
	rightptr = isptrclass(rop);

	switch((unsigned int) op)
		{
		case OP_INDIRECT:
			new = rop->node_type->type_ptr.base_type;
			break;

		case OP_ADDRESS:
			new = ptrto(rop->node_type);
			break;

		case OP_OFFSET:
		case OP_MEMBER:
		case OP_TERN:
		case OP_SEQ:
		case OP_UPLUS:
		case OP_PARAM:
			new = rop->node_type;
			break;

		case OP_ADD:
			/* Check for addition of a pointer/array and offset */
			if(leftptr)
				{
				if(rightptr)
					error("Pointer addition");
				new = lop->node_type;
				}
			else if(rightptr)
				new = rop->node_type;
			else
				new = greatest_type(lop,rop);
			break;

		case OP_SUB:
			if(leftptr)
				{
				if(rightptr)
					{
					if(! same_type(lop->node_type,
						rop->node_type,
						/* ptr_eq_array = */ TRUE,
						/* void_ptr_ok = */ FALSE))
						{
			    error("Subtracting pointers of different types");
						}
					new = Inttype;
					}
				else
					new = lop->node_type;
				}
			else if(rightptr)
				{
			error("Pointer not allowed as right operand only");
				new = rop->node_type;
				}
			else
				new = greatest_type(lop, rop);
			break;

		case OP_MUL:
		case OP_DIV:
		case OP_MOD:
		case OP_BAND:
		case OP_BXOR:
		case OP_BOR:
			if(leftptr || rightptr)
				error("Pointer not allowed");
			new = greatest_type(rop,lop);
			break;

		case OP_UMIN:
		case OP_BNOT:
			if(rightptr)
				error("Pointer not allowed");
			new = rop->node_type;
			break;

		case OP_AADD:
		case OP_ASUB:
		case OP_POSTADD:
		case OP_POSTSUB:
			if(rightptr)
				error("Pointer not allowed");
			new = lop->node_type;
			break;

		case OP_RSHIFT:
		case OP_LSHIFT:
		case OP_AMUL:
		case OP_ADIV:
		case OP_AMOD:
			if(leftptr || rightptr)
				error("Pointer not allowed");
			new = lop->node_type;
			break;

		case OP_ASSIGN:
			if(! same_type(lop->node_type,rop->node_type,
		   		/* ptr_eq_array = */ TRUE,
					/* void_ptr_ok = */ TRUE))
				{
				if(leftptr && rightptr)
					{
			    warning(1,"Different pointer types across \"=\"");
					}
				else if(leftptr && (! ISCONSTANT(rop) ||
					(EXPRTYPE(rop) & FP_NODE) ||
				(rop->right.ptr.terminal->val.integral != 0)))
					{
				warning(1,"Invalid assignment to pointer");
					}
				else if(rightptr)
					{
				warning(1,"Assigning pointer to non-pointer");
					}
				}
			/* Fall through to use left operand */
		case OP_ARSHIFT:
		case OP_ALSHIFT:
		case OP_ABAND:
		case OP_ABXOR:
		case OP_ABOR:
			new = lop->node_type;
			break;

		case OP_LT:
		case OP_GT:
		case OP_LE:
		case OP_GE:
		case OP_EQ:
		case OP_NE:
		case OP_LAND:
		case OP_LOR:
		case OP_LNOT:
		case OP_BOOL:
			/* Boolean expressions yield "integers" */
			new = Inttype;
			break;

		case OP_TSEL:
			new = select_type(lop,rop);
			break;

		case OP_CALL:
		case OP_ICALL:
			new = rop->node_type->type_ptr.base_type;
			break;

		case OP_CONVERT:
			if(lop->right.ptr.terminal->type & SU_NODE)
				error("Cannot cast to struct/union");
			if((rop->node_type->type & SU_NODE) &&
				(! (lop->right.ptr.terminal->type &
					VOID_NODE)))
				{
				error("Cannot cast struct/union");
				}
			new = lop->right.ptr.terminal;	/* Contains type ptr */
			break;

		default:
			fatal("Compiler error: new_type");
		}
/*
 *	Do case analysis to ensure correct use of "void *" constructs.
 */
	if(leftptr || rightptr)
		{
		leftvp = FALSE;		/* Assume not void pointer initially */
		if(leftptr)
			leftvp = ((EXPRTYPE(lop) & PTR_NODE) &&
				(lop->node_type->type_ptr.base_type->type &
					VOID_NODE));
		rightvp = FALSE;	/* Assume not void pointer initially */
		if(rightptr)
			rightvp = ((EXPRTYPE(rop) & PTR_NODE) &&
				(rop->node_type->type_ptr.base_type->type &
					VOID_NODE));
		vserror = FALSE;	/* Assume OK for now */
		switch((unsigned int) op)
			{
			case OP_UPLUS:
			case OP_ADD:
			case OP_SUB:
			case OP_MUL:
			case OP_DIV:
			case OP_MOD:
			case OP_BAND:
			case OP_BXOR:
			case OP_BOR:
			case OP_UMIN:
			case OP_BNOT:
			case OP_AADD:
			case OP_ASUB:
			case OP_POSTADD:
			case OP_POSTSUB:
			case OP_RSHIFT:
			case OP_LSHIFT:
			case OP_AMUL:
			case OP_ADIV:
			case OP_AMOD:
			case OP_ARSHIFT:
			case OP_ALSHIFT:
			case OP_ABAND:
			case OP_ABXOR:
			case OP_ABOR:
			case OP_LAND:
			case OP_LOR:
			case OP_BOOL:
				if(leftvp || rightvp)
					vserror = TRUE;
				break;

			case OP_TERN:
				if(leftvp)
					vserror = TRUE;
				break;
			default:
				break;
			}
		if(vserror)
			error("Illegal operation on void pointer");
		}
/*
 *	Check for correct use of "void" constructs.
 */
	if(op == OP_SEQ)
		{
		if((rop != NULL) && (EXPRTYPE(rop) & VOID_NODE))
			{
		error("void type may only be on left side of comma operator");
			}
		}
	else if((op != OP_CONVERT) &&
		(((lop != NULL) && (EXPRTYPE(lop) & VOID_NODE)) ||
		((rop != NULL) && (EXPRTYPE(rop) & VOID_NODE))))
		{
		error("Illegal operation on void type");
		}
	return (new);
	}

#ifdef	MSC_OPT
#pragma	optimize("e",on)		/* Enable global reg optimizing */
#endif	/* MSC_OPT */

/*
 *	Return type of a select (:) operator.
 */
PRIVATE	struct	id_node	*
select_type(lop,rop)
	struct	op_node		*lop;
	struct	op_node		*rop;
	{
	if(same_type(lop->node_type,rop->node_type,/* ptr_eq_array = */ TRUE,
		/* void_ptr_ok = */ FALSE))
		{
		return (lop->node_type);	/* Same: return left type */
		}
	else if(isptrclass(lop) && ISCONSTANT(rop) &&
		(! (EXPRTYPE(rop) & FP_NODE)) &&
			rop->right.ptr.terminal->val.integral == 0)
		{
		return (lop->node_type);
		}
	else if(isptrclass(rop) && ISCONSTANT(lop) &&
		(! (EXPRTYPE(lop) & FP_NODE)) &&
			lop->right.ptr.terminal->val.integral == 0)
		{
		return (rop->node_type);
		}
	else if((EXPRTYPE(lop) & ARITH_NODE) && (EXPRTYPE(rop) & ARITH_NODE))
		return (greatest_type(lop,rop));
	else
		{
		error("Incompatible types within select");
		return (Inttype);
		}
	}

/*
 *	Return the greatest promoted type of the two operands.
 */
PUBLIC	struct	id_node	*
greatest_type(lop,rop)
	struct	op_node		*lop;
	struct	op_node		*rop;
	{
	INT32			temp_type;	/* Damn buggy MSC 6.00A! */
/*
 *	Return id_node pointer to type of bits.
 */
/*
 *	Note that if we are in "pre-ANSI" mode we must always promote floating
 *	point operands to be at least "double" size prior to performing
 *	operations on them to ensure that full precision is retained.
 */
	temp_type = EXPRTYPE(lop) | EXPRTYPE(rop);
	if((Float_mode == FLOAT_PREANSI) && (temp_type & FP_NODE))
		{
		return (basetype(temp_type | DOUBLE_NODE));
		}
/*
 *	If FLOAT_ANSI or FLOAT_SINGLE ("single" and "double" are the same
 *	size), we avoid the automatic promotion.
 */
	else
		return (basetype(basetype(EXPRTYPE(lop))->type |
			basetype(EXPRTYPE(rop))->type));
	}

/*
 *	Return a basetype pointer to an id_node corresponding to the
 *	type bits passed as parameter.
 */
PUBLIC	struct	id_node	*
basetype(type)
	INT32			type;
	{
	if(type & LDOUBLE_NODE)
		return (Ldoubletype);
	else if(type & DOUBLE_NODE)
		return (Doubletype);
	else if(type & FLOAT_NODE)
		return (Floattype);
	else if(type & LONG_NODE)
		{
		if(type & UNSIGNED_NODE)
			return (Ulongtype);
		else
			return (Longtype);
		}
	else if(type & (INT_NODE | SHORT_NODE | ENUM_NODE))
		{
		if(type & UNSIGNED_NODE)
			return (Uinttype);
		else
			return (Inttype);
		}

	return (Inttype);		/* Default */
	}

/*
 *	Evaluate an expression and return a pointer to a tree of op_nodes.
 */
PUBLIC	struct	op_node	*
expression()
	{
	return(hier1());
	}

/*
 *	Sequence operator (,).
 */
PRIVATE	struct	op_node	*
hier1()
	{
	struct	op_node		*lop;

	lop = hier2();

	while(matchc(','))
		lop = binaryop(OP_SEQ,lop,hier2());

	return (lop);
	}

/*
 *	Assignment operator helper function.
 */
PRIVATE	int
matchassign()
	{
	struct	id_node		*op;
	char			operator[4];
	int			value;

	value = 0;

	blanks();
	if(strchr("+|^&*-></%=", ch()))
		{
		operator[0] = ch();

		if(operator[0] != '=')
			{
			operator[1] = Lptr[1];

			if((operator[1] == '<') || (operator[1] == '>'))
				{
				operator[2] = Lptr[2];
				operator[3] = '\0';
				}
			else
				operator[2] = '\0';
			}
		else
			operator[1] = '\0';

		if((op = lookup_node(operator,NS_CKEYWORD)) != NULL)
			{
			value = (int) (op->val.integral);
/*
 *	It would be nice to remove ALL references to Lptr and do this
 *	"functionally" instead.  For now, this is a kludge.
 */
			Lptr += strlen(operator);
			}
		}
	return (value);
	}

/*
 *	Assignment operators.
 */
PUBLIC	struct	op_node	*
hier2()
	{
	struct	op_node		*lop;
	BOOL			lval;
	int			operator;
	struct	op_node		*rop;
	INT32			val;

	lop = hier3();

	if((operator = matchassign()) != 0)
		{
		lval = needlvalue(lop);		/* Must be lvalue */

		rop = hier2();

		if(lval)
			{
			if(ISCONSTANT(rop) && (EXPRTYPE(rop) & FP_NODE))
				{
				val = f_to_l(rop->right.ptr.terminal->
					val.fvalue,FALSE);
				if(((double) val) == rop->right.ptr.terminal->
					val.fvalue)
					{
/*
 *	Reach down and change constant to "INT32" and value to an integral
 *	value.
 */
					free_op_node(rop);
					rop = make_const(val,Inttype);
					}
				}
			lop = binaryop(operator,lop,rop);
			}
		else
			{
/*
 *	Can't use lop - free it and use rop instead.
 */
			free_op_node(lop);
			lop = rop;
			}
		}
	return (lop);
	}

/*
 *	Ternary operator:  (condition) ? (thenpart) : (elsepart)
 *
 *	A slight bug:  This code permits the construct:
 *		TRUE ? 17 : c
 *	where c is not a constant as a constant expression.  This is not
 *	correct.  One possible way of handling this is to have a bit in
 *	the operand block like LD_CONSTANT (which says we have a constant
 *	value) but this other bit says this is a constant expression,
 *	which would get turned off here.  The high level code would then
 *	be responsible for checking to see that it really received a
 *	constant expression, as opposed to a constant value.
 *
 *	Is this complaint still valid?  The new "C" references seem to allow
 *	this sort of thing (plus perhaps other stuff)?  KAB 10/14/87
 */
PUBLIC	struct	op_node	*
hier3()
	{
	struct	op_node		*opcond;
	struct	op_node		*opelse;
	struct	op_node		*opthen;

	opcond = hier4();

	if(! matchc('?'))
		return (opcond);

	opthen = hier3();

	needpunc(':');

	opelse = hier3();

	if(ISCONSTANT(opcond))
		{
		if(((EXPRTYPE(opcond) & FP_NODE) &&
			(opcond->right.ptr.terminal->val.fvalue != 0.0)) ||
				((! (EXPRTYPE(opcond) & FP_NODE)) &&
					(opcond->right.ptr.terminal->
						val.integral != 0)))
			{		/* Use first part */
			free_op_node(opelse);
			free_op_node(opcond);
			opcond = opthen;
			}
		else
			{		/* Use second part */
			free_op_node(opthen);
			free_op_node(opcond);
			opcond = opelse;
			}
		}
	else
		{
		opcond = binaryop(OP_TERN,opcond,
			binaryop(OP_TSEL,opthen,opelse));
		}

	return (opcond);
	}

/*
 *	Helping function for hier4 & hier5.  Makes scalar constants into
 *	canonical True and False.
 */
PRIVATE	VOID
boolify(op)
	struct	op_node		*op;
	{
	if(ISCONSTANT(op))
		{
		if(((EXPRTYPE(op) & FP_NODE) &&
			(op->right.ptr.terminal->val.fvalue == 0.0)) ||
				((! (EXPRTYPE(op) & FP_NODE)) &&
					(op->right.ptr.terminal->
						val.integral == 0)))
			{
			op->right.ptr.terminal = False;	/* Make canonical */
			op->node_type = False->type_ptr.base_type;
			}
		else
			{
			op->right.ptr.terminal = True;	/* Make canonical */
			op->node_type = True->type_ptr.base_type;
			}
		}
	}

/*
 *	Parse a logical OR expression (e1 || e2).
 */
PRIVATE	struct	op_node	*
hier4()
	{
	BOOL			canonical;
	BOOL			done;
	BOOL			had_one;
	struct	op_node		*lop;
	struct	op_node		*rop;

	done = had_one = canonical = FALSE;

	lop = hier5();

	while(match("||"))
		{
		boolify(lop);
		had_one = TRUE;
		boolify(rop = hier5());
		if(done)
			free_op_node(rop);
		else if(ISCONSTANT(lop))
			{
			if(lop->right.ptr.terminal->val.integral == TRUE)
				free_op_node(rop);	/* Return lop */
			else
				{
				free_op_node(lop);
				lop = rop;
				}
			}
		else if(ISCONSTANT(rop) &&
			(rop->right.ptr.terminal->val.integral == FALSE))
			{
			free_op_node(rop);	/* Return lop */
			}
		else
			{
			lop = binaryop(OP_LOR,lop,rop);
			canonical = TRUE;
			if(ISCONSTANT(rop) && (rop->right.ptr.terminal->
				val.integral == TRUE))
				{
				done = TRUE;
				}
			}
		}
/*
 *	Force the value to be canonical {0,1} if needed.
 */
	if(had_one && (! canonical) && (! ISCONSTANT(lop)))
		lop = binaryop((int) OP_BOOL,NULL,lop);

	return (lop);
	}

/*
 *	Parse a logical AND expression (e1 && e2).
 */
PRIVATE	struct	op_node	*
hier5()
	{
	BOOL			canonical;
	BOOL			done;
	BOOL			had_one;
	struct	op_node		*lop;
	struct	op_node		*rop;

	done = had_one = canonical = FALSE;

	lop = hier6();

	while(match("&&"))
		{
		boolify(lop);
		had_one = TRUE;
		boolify(rop = hier6());
		if(done)
			free_op_node(rop);
		else if(ISCONSTANT(lop))
			{
			if(lop->right.ptr.terminal->val.integral == FALSE)
				free_op_node(rop);	/* Return lop */
			else
				{
				free_op_node(lop);
				lop = rop;
				}
			}
		else if(ISCONSTANT(rop) &&
			(rop->right.ptr.terminal->val.integral == TRUE))
			{
			free_op_node(rop);	/* Return lop */
			}
		else
			{
			lop = binaryop(OP_LAND,lop,rop);
			canonical = TRUE;
			if(ISCONSTANT(rop) && (rop->right.ptr.terminal->
				val.integral == FALSE))
				{
				done = TRUE;
				}
			}
		}
/*
 *	Force the value to be canonical {0,1} if needed.
 */
	if(had_one && (! canonical) && (! ISCONSTANT(lop)))
		{
		lop = binaryop((int) OP_BOOL,NULL,lop);
		}

	return (lop);
	}

/*
 *	Bitwise inclusive OR (e1 | e2).
 */
PRIVATE	BOOL
matchor()
	{
	blanks();
	return ((nch() != '|') && (nch() != '=') && matchc('|'));
	}

PRIVATE	struct	op_node	*
hier6()
	{
	struct	op_node		*lop;

	lop = hier7();

	while(matchor())
		lop = binaryop((int) OP_BOR,lop,hier7());

	return (lop);
	}

/*
 *	Bitwise exclusive OR (e1 ^ e2).
 */
PRIVATE	BOOL
matchxor()
	{
	blanks();
	return ((nch() != '=') && matchc ('^'));
	}

PRIVATE	struct	op_node	*
hier7()
	{
	struct	op_node		*lop;

	lop = hier8();

	while(matchxor())
		lop = binaryop((int) OP_BXOR,lop,hier8());

	return (lop);
	}

/*
 *	Bitwise AND (e1 & e2).
 */
PRIVATE	BOOL
matchand()
	{
	blanks();
	return ((nch() != '&') && (nch() != '=') && matchc ('&'));
	}

PRIVATE	struct	op_node	*
hier8()
	{
	struct	op_node		*lop;

	lop = hier9();

	while(matchand())
		lop = binaryop((int) OP_BAND,lop,hier9());

	return (lop);
	}

/*
 *	Equality comparisons (==, !=).
 */
PRIVATE	int
matcheqlop()
	{
	blanks();
	if(match("=="))
		return ((int) OP_EQ);
	else if(match("!="))
		return ((int) OP_NE);
	else
		return (0);
	}

PRIVATE	struct	op_node	*
hier9()
	{
	struct	op_node		*lop;
	int			operator;

	lop = hier10();

	while((operator = matcheqlop()) != 0)
		lop = binaryop(operator,lop,hier10());

	return (lop);
	}

/*
 *	Relational comparisons (<, <=, >=, >).
 */
PRIVATE	int
matchrelop()
	{
	blanks();
	if(match("<="))
		return ((int) OP_LE);
	else if(match(">="))
		return ((int) OP_GE);
	else if((nch() != '<') && matchc('<'))
		return ((int) OP_LT);
	else if((nch() != '>') && matchc('>'))
		return ((int) OP_GT);
	else
		return ((int) 0);
	}

PRIVATE	struct	op_node	*
hier10()
	{
	struct	op_node		*lop;
	int			operator;

	lop = hier11();

	while((operator = matchrelop()) != 0)
		lop = binaryop(operator,lop,hier11());

	return (lop);
	}

/*
 *	Shift operators.
 */
PRIVATE	int
matchshift()
	{
	blanks();
	if(streq(Lptr,">>=") || streq(Lptr,"<<="))
		return (0);
	else if(match("<<"))
		return ((int) OP_LSHIFT);
	else if(match(">>"))
		return ((int) OP_RSHIFT);
	else
		return ((int) 0);
	}

PRIVATE	struct	op_node	*
hier11()
	{
	struct	op_node		*lop;
	int			operator;

	lop = hier12();

	while((operator = matchshift()) != 0)
		lop = binaryop(operator,lop,hier12());

	return (lop);
	}

/*
 *	Addition and subtraction.
 */
PRIVATE	int
matchadd()
	{
	blanks();
	if(nch() == '=')
		return (0);
	else if(matchc('+'))
		return ((int) OP_ADD);
	else if(matchc('-'))
		return ((int) OP_SUB);
	else
		return ((int) 0);
	}

PRIVATE	struct	op_node	*
hier12()
	{
	struct	op_node		*lop;
	int			operator;

	lop = hier13();

	while((operator = matchadd()) != 0)
		lop = binaryop(operator,lop,hier13());

	return (lop);
	}

/*
 *	Multiplicative operators (*, /, and %).
 */
PRIVATE	int
matchmult()
	{
	blanks();
	if(nch() == '=')
		return (0);
	else if(matchc('*'))
		return ((int) OP_MUL);
	else if(matchc('/'))
		return ((int) OP_DIV);
	else if(matchc('%'))
		return ((int) OP_MOD);
	else
		return ((int) 0);
	}

PRIVATE	struct	op_node	*
hier13()
	{
	struct	op_node		*lop;
	int			operator;

	lop = hier14();

	while((operator = matchmult()) != 0)
		lop = binaryop(operator,lop,hier14());

	return (lop);
	}

/*
 *	Handle explicit and implicit cast operations.
 */
PUBLIC	struct	op_node	*
cast_it(op,type)
	struct	op_node		*op;
	struct	id_node		*type;
	{
	struct	op_node		*op1;
/*
 *	A bit of work here: we make a terminal node out of the "type" so that
 *	we can pass it into binaryop().  The terminal op_node is soon freed by
 *	binaryop().
 */
	op1 = terminal(type);
/*
 *	Fix the node_type pointer.  terminal() does an implicit dereference by
 *	going through the base_type pointer of the id_node.  Normally this
 *	just gets the base type of the symbol, etc.  However for the id_node
 *	created for the (casted) type, it causes an EXPLICIT dereference,
 *	which is NOT desired.
 */
	op1->node_type = op1->right.ptr.terminal;	/* Fix node_type */

	return (binaryop(OP_CONVERT,op1,op));
	}

/*
 *	Prefix unary operators.
 */
PRIVATE	struct	op_node	*
hier14()
	{
	int			need_paren;
	struct	op_node		*op;
	struct	op_node		*op1;
	INT32			size;
	struct	id_node		*type;

	struct_error(NULL,NULL);	/* No structure syntax errors yet */

	op = NULL;			/* No expression yet */

	if(match("++"))
		{
		op1 = hier14();

		if(needlvalue(op1))
			{
			/* Turn it into an op += 1 */
			op = binaryop((int) OP_AADD,op1,
				make_const((INT32) 1,Inttype));
			}
		else
			op = op1;
		}
	else if(match("--"))
		{
		op1 = hier14();

		if(needlvalue(op1))
			{
			/* Turn it into an op -= 1 */
			op = binaryop(OP_ASUB,op1,
				make_const((INT32) 1,Inttype));
			}
		else
			op = op1;
		}
	else if(matchc('-'))
		op = binaryop(OP_UMIN,NULL,hier14());
	else if(matchc('+'))
		op = binaryop(OP_UPLUS,NULL,hier14());
	else if(matchc('*'))
		{			/* Indirection */
		op = hier14();		/* Evaluate expression */
		if(! isptrclass(op))
			error("Not a pointer");
		else if(op->operator == OP_ADDRESS)
			{
			/* '*' of an '&' just gets ignored */
			op1 = op;	/* Save copy so we can free it */
			op = RIGHT(op);	/* Remove '&' node */
			RIGHT(op1) = LEFT(op1) = NULL;
			free_op_node(op1);
			}
		else
			{
			/* Otherwise bind on an indirection operator */
			op = binaryop(OP_INDIRECT,NULL,op);
			}
		}
	else if(matchc('!'))
		op = binaryop(OP_LNOT,NULL,hier14());
	else if(matchc('~'))
		op = binaryop(OP_BNOT,NULL,hier14());
	else if(matchc('&'))
		{			/* Address */
		op = hier14();
		if(! islvalue(op) || (ISTERMINAL(op) &&
			(op->right.ptr.terminal->info.storage_class ==
				SC_REG)))
			{
			error("Illegal address");
			}
		else if(op->operator == OP_INDIRECT)
			{
			/* '&' of an '*' just gets ignored */
			op1 = op;	/* Save copy so we can free it */
			op = RIGHT(op);	/* Remove '&' node */
			RIGHT(op1) = LEFT(op1) = NULL;
			free_op_node(op1);
			}
		else
			op = binaryop(OP_ADDRESS,NULL,op);
		}
	else if(amatch("sizeof"))
		{
		need_paren = matchc('(');
		if((type = abstdecl()) != NULL)
			{
			/* sizeof abstract declarator */
			size = type->val.integral;
			if(type->type & FUNC_NODE)
				error("Cannot take 'sizeof' a function");

			if(need_paren)
				needpunc(')');
			}
		else
			{
			/* sizeof a variable */
			if(need_paren)
				putback('(');

			op1 = hier14();

			size = op1->node_type->val.integral;

			free_op_node(op1);	/* We don't need it */
			}

		if(size < 0)
			size = 0;	/* Unknown size */

		if(size == 0)
			warning(1,"'sizeof' object is zero or unknown");

		op = make_const(size,Uinttype);
		}
	else if(matchc('('))
		{
		/* Check casting operator */
		if((type = abstdecl()) != NULL)
			{
			/* An abstract declarator */
			needpunc(')');	/* Must have closing paren */
			op = hier14();	/* Get expression */
			if((type->type & (SU_NODE | ARRAY_NODE | FUNC_NODE)) ||
				(EXPRTYPE(op) & FUNC_NODE) ||
					((EXPRTYPE(op) & SU_NODE) &&
						(! (type->type & VOID_NODE))))
				{
				struct_error("Illegal cast",NULL);
				}
			else
				op = cast_it(op,type);
			}
		else
			putback('(');
		}

	if(op == NULL)
		op = hier15();		/* Look for a primary expression */

	return (op);
	}

/*
 *	Postfix unary, function calls, arrays, etc.
 */
PRIVATE	struct	op_node	*
hier15()
	{
	struct	id_node		*base_type;
	char			buffer1[NAMEMAX + 1];
	register int		dot;
	char			mem_name[NAMEMAX + 1];
	register int		name_space;
	struct	op_node		*op;
	struct	op_node		*op1;
	INT32			op_t_code;
	struct	id_node		*op_type;
	struct	prototype	*proto;
	struct	id_node		*sym;

	op = primary();
	for(;;)
		{
		op_type = op->node_type;
		op_t_code = op_type->type;

		if(match("++"))
			{
			if(needlvalue(op))
				{
				op = binaryop(OP_POSTADD,op,
					make_const((INT32) 1,Inttype));
				}
			}
		else if(match("--"))
			{
			if(needlvalue(op))
				{
				op = binaryop(OP_POSTSUB,op,
					make_const((INT32) 1,Inttype));
				}
			}
		else if(matchc('('))
			{
			if((op_t_code & PTR_NODE) &&
				(op_type->type_ptr.base_type->type &
					FUNC_NODE))
				{
/*
 *	"op" is a pointer to function -- treat as if called with *ptr().
 */
				base_type = op_type->type_ptr.base_type->
					type_ptr.base_type;
				proto = op_type->type_ptr.base_type->
					proto_list;
				op = binaryop(OP_INDIRECT,NULL,op);
				}
			else if((op_t_code & FUNC_NODE) == 0)
				{
				base_type = Inttype;
				proto = NULL;
				error("Not a function");
				}
			else
				{
				base_type = op_type->type_ptr.base_type;
				proto = op_type->proto_list;
				}

			if(proto == NULL)
				warning(3,"No prototype in scope");

			op1 = parameters(proto != NULL,proto);
/*
 *	If the static link data model is in effect and the function being
 *	called isn't a "intrinsic", add the "static link" to the function
 *	parameter list.
 */
			if(Static_link && (ISNON_TERMINAL(op) ||
				(op->right.ptr.terminal->val.funcnum == 0)))
				{
				if((sym = lookup_node(SLBNAME,NS_VAR)) == NULL)
					fatal("Compiler error: hier15a");
				op1 = binaryop(OP_PARAM,op1,lterminal(sym));
				}
/*
 *	For functions which return structs or unions, add a pointer to the
 *	return destination to the parameter list.
 */
			if(base_type->type & SU_NODE)	/* Struct/union ret */
				{
				if((sym = add_loc(strlabel(getlabel(),buffer1),
					Inttype,SC_RSTEMP,NS_VAR,
						(INT_NODE | SIGNED_NODE))) !=
							NULL)
					{
					sym->val.integral =
						Inttype->val.integral;
					sym->address.tempsize =
						base_type->val.integral;
					}
				else
					fatal("Compiler error: hier15b");
				op1 = binaryop(OP_PARAM,op1,lterminal(sym));
				}

			if(ISTERMINAL(op) &&
				(op->right.ptr.terminal->val.funcnum != 0))
				{
				op = binaryop(OP_ICALL,op1,op);
				}
			else
				op = binaryop(OP_CALL,op1,op);

			needpunc(')');
			}
		else if(((dot = matchc('.')) != 0) || match("->"))
			{
			if(dot)		/* Structure reference */
				{
				if(op_t_code & SU_NODE)
					name_space = op_type->name_space;
				else
					{
					struct_error("Must be structure/union",
						NULL);
					name_space = 0;	/* For lint */
					}
				}
			else		/* Pointer to structure reference */
				{
				if((op_t_code & (PTR_NODE | ARRAY_NODE)) &&
					((base_type = op_type->
						type_ptr.base_type) != NULL) &&
					(base_type->type & SU_NODE))
					{
					name_space = base_type->name_space;
					}
				else
					{
		struct_error("Must be pointer-to-structure/union",NULL);
					name_space = 0;	/* For lint */
					}
				}

			if((sym = ismember(mem_name,name_space)) == NULL)
				{
				struct_error("Not a member of structure/union",
					mem_name);
				op->node_type = Inttype; /* Try to avoid err */
				op->lvalue = TRUE;
				}
			else
				{
				if(! dot)
					op = binaryop(OP_INDIRECT,NULL,op);

				op = binaryop(OP_MEMBER,op,terminal(sym));

				op->node_type = sym->type_ptr.base_type;
				}
			}
		else if(matchc('['))	/* Check for array reference */
			{
/*
 *	It would be better to accept the expression, then test for legality of
 *	subscripting.  The expression would be ignored and not added to the
 *	expression tree when it would cause an illegal subscript formation.
 *	22-Jun-87 GO
 */
			op1 = expression();

			if((! ISINTEGRAL(op1)) ||
				(! (isptrclass(op) ^ isptrclass(op1))))
				{
				error("Can't subscript");
				free_op_node(op1);	/* Won't need expr */
				}
			else
				{
				op = binaryop((int) OP_INDIRECT,NULL,
					binaryop((int) OP_ADD,op,op1));
				}

			needpunc(']');
			}
		else
			break;
		}

	return (op);
	}

/*
 *	Helping function for "parameters()".  Attach parameter to node list
 *	and perform appropriate magic to make function calls which return a
 *	struct/union and are themselves parameters to a higher level function
 *	copy the returned functions into the correct spot in the parameter
 *	list for the higher level function call (avoid a duplicate copy of
 *	the return struct/union value).
 */
PRIVATE	struct	op_node	*
param_help(lop,rop)
	struct	op_node		*lop;
	struct	op_node		*rop;
	{
	struct	op_node		*op;
	struct	id_node		*sym;

	if((EXPRTYPE(rop) & SU_NODE) && (rop->operator == OP_CALL))
		{
		if(((op = LEFT(rop)) == NULL) || (op->operator != OP_PARAM) ||
			((op = RIGHT(op)) == NULL))
			{
			fatal("Compiler error: param_help #1");
			}
		if(ISTERMINAL(op))
			{
			if(sym = op->right.ptr.terminal,
				sym->info.storage_class != SC_RSTEMP)
				{
				fatal("Compiler error: param_help #2");
				}
			sym->address.tempsize = 0;	/* Disable dup copy */
			}
		}
	return (binaryop(OP_PARAM,lop,rop));
	}

/*
 *	Parse a function call parameter set.
 */
PRIVATE	struct op_node *
parameters(has_proto,proto_list)
	BOOL			has_proto;
	struct	prototype	*proto_list;
	{
	struct	op_node		*op;
/*
 *	Any parameters?  This check handles the case of no parameters at all,
 *	The check at the end of the function handles the case of "no more"
 *	parameters (since we are recursive).  These checks can't be combined
 *	since the syntax tree has no "OP_PARAM" nodes for the special case of
 *	zero parameters.
 */
	if(nextc(')'))
		{
		if((proto_list != NULL) && (proto_list->type != Voidtype))
			warning(1,"Insufficient parameters");
		return (NULL);
		}
/*
 *	Check parameter type with function prototype.
 */
	op = hier2();
	if(has_proto && (proto_list != NULL) &&
		(proto_list->type != NULL /*ellipsis*/))
		{
		if(! same_type(op->node_type, proto_list->type,
			/* ptr_eq_array = */ TRUE,/* void_ptr_ok = */ TRUE))
			{
/*
 *	Note that assigning NULL to any type of pointer is legit.
 */
			if(! ((proto_list->type->type & PTYPE_NODE) &&
				(ISCONSTANT(op) && ISINTEGRAL(op) &&
					(op->right.ptr.terminal->
						val.integral == 0))))
				{
				if(isptrclass(op) && ((proto_list->
					type->type & PTYPE_NODE) != 0))
					{
					/* Error if incompatible ptr types */
					warning(2,
						"Illegal pointer conversion");
					}
				else if(isptrclass(op) ||
					((proto_list->type->type &
						PTYPE_NODE) != 0))
					{
					/* Error if ptr -vs- non-pointer */
					warning(1,
						"Illegal pointer conversion");
					}
				else
					{
/*
 *	If the parameter type is non numeric, or is numeric and is different
 *	from the prototype after the application of the usual unary promotions,
 *	then optionally flag a warning.
 */
					if(((op->node_type->type&ARITH_NODE) ==
							0) ||
						(basetype(op->node_type->
							type) !=
						basetype(proto_list->type->
							type)))
						{
					warning(3,"Parameter type conversion");
						}
					}


				/* Cast to the appropriate type */
				op = cast_it(op,proto_list->type);
				}
			}
		}
	else if(EXPRTYPE(op) & FP_NODE)
		{
		/* No prototype, FP params -> "double" */
		if(! same_type(op->node_type,Doubletype,
			/* ptr_eq_array = */ FALSE,/* void_ptr_ok = */ FALSE))
			{
	warning(2,"No function prototype, float parameter cast to double");
			op = cast_it(op,Doubletype);
			}
		}

	if(matchc(','))
		{
		if((proto_list != NULL) &&
			(proto_list->type != NULL /*ellipsis*/))
			{
			proto_list = proto_list->next;
			if(proto_list == NULL)
				{
		warning(1,"More parameters used than the function allows");
				}
			}
		op = param_help(parameters(has_proto,proto_list),op);
		}
	else
		{
		if((proto_list != NULL) &&
			(proto_list->type != NULL /*ellipsis*/))
			{
			proto_list = proto_list->next;
			if((proto_list != NULL) &&
				(proto_list->type != NULL /*ellipsis*/))
				{
				warning(1,"Insufficient parameters");
				}
			}
		op = param_help(NULL,op);
		}

	return (op);
	}

/*
 *	See if the next token is a member of the specified structure/union
 *	name space.  If found, return pointer to id_node.
 */
PRIVATE	struct	id_node	*
ismember(name,name_space)
	char			*name;
	int			name_space;
	{
	struct	id_node		*sym;

	if(! symname(name))
		{
		illname();
		sym = NULL;
		}
	else
		sym = lookup_node(name,name_space);

	return (sym);
	}

/*
 *	Evaluate a primary expression.
 */
PRIVATE	struct	op_node	*
primary()
	{
	struct	id_node		*isym;
	char			name[NAMEMAX + 1];
	struct	op_node		*op;
	struct	id_node		*sym;
	struct	id_node		*type;

	op = get_op_node();

	if(matchc('('))
		{
		free_op_node(op);	/* We don't need it */
		op = hier1();		/* Start a new expression */
		needpunc(')');		/* One level down */
		}
	else if(chksym(name) != NULL)
		{			/* See if next token is legal name */
		skip();

		if((isym = lookup_node(name,NS_IFUNC)) != NULL)
			{
			if((! Builtin) || (! (Cpu_type & isym->val.funcnum)))
				{
				strcpy(name,isym->name + strlen("_BUILTIN_"));
				isym = NULL;	/* Do non-internal func call */
				}
			else if((sym = lookup_node(name,NS_VAR)) == NULL)
				{
		error("No type information supplied for _BUILTIN_ function");
				/* Error recovery, give it terminal value */
				op->right.ptr.terminal = True;
				op->node_type = True->type_ptr.base_type;
				}
			else
				{
				/* Internal func # */
				sym->val.funcnum = isym->val.funcnum;
				op->right.ptr.terminal = sym;
				op->node_type = sym->type_ptr.base_type;

				if(! nextc('('))
					{
			error("No address exists for a _BUILTIN_ function");
					}
				}
			}

		if(isym == NULL)
			{
			if((sym = lookup_node(name,NS_VAR)) != NULL)
				{
				op->right.ptr.terminal = sym;
				op->node_type = sym->type_ptr.base_type;

				if(sym->type & FUNC_NODE)
					{
					if(! nextc('('))
						{
						op = binaryop(OP_ADDRESS,NULL,
							op);
						op->node_type = ptrto(sym->
							type_ptr.base_type);
						}
					}
				else if((sym->type & ARRAY_NODE) == 0)
					op->lvalue = TRUE;
				}
			else if(nextc('('))
				{
				/* Appears to be function() reference */
	warnname(3,"Undefined function, 'int' return type assumed",name);
				op->right.ptr.terminal =
					add_loc(name,Ifunctype,SC_EXTERN,
						NS_VAR,(INT32) FUNC_NODE);

				op->node_type = Ifunctype;
				}
			else
				{
				errname("Undefined variable",name);
				type = (nextc('[') ? ptrto(Inttype) : Inttype);

				op->right.ptr.terminal =
					add_loc(name,type,SC_EXTERN,NS_VAR,
						(INT32) INT_NODE);

				op->node_type = type;
				/* Pretend LVALUE so no errors */
				op->lvalue = TRUE;
				}
			}
		}
	else if((! constant(op)) && (! string(op)))
		{			/* If not, it's an error */
		error("Invalid expression");
		if(ch() != ';')
			junk();

		/* Error recovery attempt */
		op->right.ptr.terminal = True;	/* Give it a terminal value */
		op->node_type = True->type_ptr.base_type;
		}

	return (op);
	}

/*
 *	Check for LVALUE and bitch if not present.
 */
PRIVATE	BOOL
needlvalue(op)
	struct	op_node		*op;
	{
	if(islvalue(op))
		return (TRUE);
	else
		{
		error("Must be lvalue");
		return (FALSE);
		}
	}

/*
 *	Give an error message about structures, but only do it once.
 */
PRIVATE VOID
struct_error(msg,name)
	char			*msg;
	char			*name;
	{
	static	error_flag = FALSE;	/* Record err to suppress multiples */
	if(msg == NULL)			/* Reset the flag */
		error_flag = FALSE;
	else if(! error_flag)
		{
		if(name == NULL)
			error(msg);
		else
			errname(msg,name);
		error_flag = TRUE;
		}
	}

/*
 *	Return a terminal node with pointer to sym
 */
PUBLIC	struct	op_node	*
terminal(sym)
	struct	id_node		*sym;
	{
	struct	op_node		*op;

	op = get_op_node();

	op->right.ptr.terminal = sym;
	op->node_type = sym->type_ptr.base_type;

	return (op);
	}

/*
 *	Return a terminal node with pointer to sym and the "lvalue" flag set.
 */
PUBLIC	struct	op_node	*
lterminal(sym)
	struct	id_node		*sym;
	{
	struct	op_node		*op;

	op = terminal(sym);
	op->lvalue = TRUE;

	return (op);
	}

/*
 *	Make a constant of type and value specified
 */
PUBLIC	struct	op_node	*
make_const(value,type)
	INT32			value;
	struct	id_node		*type;
	{
	struct	op_node		*op;

	op = terminal(add_type(type->type,type,value));

	op->right.ptr.terminal->info.storage_class = SC_CONSTANT;

	return (op);
	}

/*
 *	Make a floating constant of type and value specified
 */
PUBLIC	struct	op_node	*
make_fconst(value,type)
	double			value;
	struct	id_node		*type;
	{
	struct	op_node		*op;

	op = terminal(add_type(type->type,type,(INT32) 0));
	op->right.ptr.terminal->val.fvalue = value;

	op->right.ptr.terminal->info.storage_class = SC_CONSTANT;

	return (op);
	}

/*
 *	Get a free op_node element.
 */
PRIVATE	struct	op_node	*
get_op_node()
	{
	int			n;
	struct	op_node		*op;

	if(free_oplist == NULL)
		{
		op = (struct op_node *) malloc(OP_ALLOC *
			sizeof(struct op_node));
		if(op != NULL)
			{
#ifdef	DEBUG
			Alloc_oplist += OP_ALLOC;	/* Total # requested */
#endif	/* DEBUG */
			for(n = 0; n < OP_ALLOC; ++n)
				{
				/* Free all in this chunk */
				RIGHT(op) = free_oplist;
				free_oplist = op++;
#ifdef	DEBUG
				++Nfree_oplist;
#endif	/* DEBUG */
				}
			}
		else
			fatal("Out of op_nodes");
		}

	op = free_oplist;
	free_oplist = RIGHT(free_oplist);
#ifdef	DEBUG
	--Nfree_oplist;			/* One more removed */
#endif	/* DEBUG */
	op->operator = OP_TERM;		/* No operator yet */
	op->left.ptr.terminal = op->right.ptr.terminal = NULL;
	op->node_type = NULL;
	op->lvalue = FALSE;		/* Not an lvalue */

	return (op);
	}

/*
 *	Recursively free an operand node tree.
 */
PUBLIC	VOID
free_op_node(tree)
	struct	op_node		*tree;
	{
	if(tree != NULL)
		{
		if(ISNON_TERMINAL(tree))
			{
			free_op_node(LEFT(tree));
			free_op_node(RIGHT(tree));
			}

		RIGHT(tree) = free_oplist;
		free_oplist = tree;
#ifdef	DEBUG
		++Nfree_oplist;
#endif	/* DEBUG */
		}
	}

PUBLIC	INT32
word(byte_offset)
	INT32			byte_offset;
	{
	return (byte_offset / Sizeof_int);
	}

/*
 *	Predicate that tests to see if two types are the same/compatible.
 */
PUBLIC	BOOL
same_type(left,right,ptr_eq_array,void_ptr_ok)
	struct	id_node		*left;
	struct	id_node		*right;
	BOOL			ptr_eq_array;
	BOOL			void_ptr_ok;
	{
	struct	prototype	*lptr;
	BOOL			ok;
	struct	prototype	*rptr;
/*
 *	Follow type chains until no more type information to compare, or a
 *	"significant" difference is found.
 */
	while(left != NULL)
		{
		if(right == NULL)
			return (FALSE);	/* Ran out too soon */
		ok = FALSE;
/*
 *	If the two types are arrays/pointers, we must take into account two
 *	special argument flags:
 *
 *	"ptr_eq_array"	If this flag is TRUE, then pointers to objects, and
 *			arrays of the same objects, compare equal.
 *
 *	"void_ptr_ok"	If this flag is TRUE, then a void pointer is allowed
 *			to compare equal to any other type of pointer.
 */
		if((left->type & PTYPE_NODE) &&	(right->type & PTYPE_NODE))
			{
			if(void_ptr_ok && (((left->type_ptr.base_type->type |
				right->type_ptr.base_type->type) &
					VOID_NODE) != 0))
				{
				return (TRUE);
				}
			else
				{
/*
 *	If we aren't dealing with a "void_ptr_ok" and a void pointer, then
 *	determine if the size of the currently pointed-to objects are the
 *	same, and, modified by the setting of the "ptr_eq_array" parameter,
 *	set the "ok" flag as appropriate for this stage in the type comparison.
 */
				ok = left->type_ptr.base_type->val.integral ==
					right->type_ptr.base_type->
						val.integral;
				if(! (ptr_eq_array ||
					((left->type & PTYPE_NODE) ==
						(right->type & PTYPE_NODE))))
					{
					ok = FALSE;
					}
				}
			}
/*
 *	If the types aren't pointers/arrays, the comparison is very simple:  If
 *	the type bits for this stage match, and the type isn't a function, then
 *	if the basic item sizes match, the types at this stage compare equal.
 *	If the type is a function, then ensure that any provided function
 *	prototypes also match.
 */
		else if((left->type & (TYPE_MASK|SIGNED_NODE|UNSIGNED_NODE)) ==
			(right->type & (TYPE_MASK|SIGNED_NODE|UNSIGNED_NODE)))
			{
			if(left->type & FUNC_NODE)
				{
				ok = TRUE;
				lptr = left->proto_list;
				rptr = right->proto_list;
/*
 *	lptr or rptr == NULL means a non-prototype declaration/definition.
 *	We also test the type field within a prototype list since the list
 *	might only be holding the names of the parameters in a non-prototype-
 *	form function definition.
 */
				if((lptr != NULL) && (lptr->type != NULL) &&
					(rptr != NULL) && (rptr->type != NULL))
					{
/*
 *	If they both have prototype lists compare them to the letter.  If the
 *	"void_ptr_ok" parameter is TRUE, indicating void pointers are 
 *	equivalent to any other pointer, pass the indication on down.
 */
					for(; (lptr != NULL) || (rptr != NULL);
						lptr = lptr->next,
							rptr = rptr->next)
						{
						if(((lptr == NULL) ||
							(rptr == NULL)) ||
							(! same_type(lptr->
							type,rptr->type,
							/*ptr_eq_array*/TRUE,
							void_ptr_ok)))
							{
							ok = FALSE;
							break;
							}
						}
					}
				}
			else
				ok = left->val.integral == right->val.integral;
			}
/*
 *	We have completed the comparison for this stage, if they compared
 *	equal, check the next stage, if not return the news.
 *
 *	Note that if we	are at a struct/union node, we don't bother to go
 *	through and check the individual components if the sizes are the same.
 *	This is a hole in the error detection which should be eventually fixed.
 */
		if(ok)
			{
			if(left->type & SU_NODE)
				return (TRUE);	/* They matched (we hope)! */

			left = left->type_ptr.base_type;
			right = right->type_ptr.base_type;
			}
		else
			return (FALSE);	/* They didn't match */
		}
	return (right == NULL);		/* TRUE if same types */
	}
	    
/* End of TCX4.C */
