/*
 *	TCX Compiler (8 of 11)
 *
 *	THIS FREE SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 *	WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *	SPECIFICALLY DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR OTHER
 *	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *	TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 *	OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *	THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 *	USE OF THIS SOFTWARE OR ANY DERIVATIVE WORK, EVEN IF ADVISED OF THE
 *	POSSIBILITY OF SUCH DAMAGE.
 *
 *	This program is hereby placed in the public domain.  In
 *	contrast to other claims of "public domain", this means no
 *	copyright is claimed and you may do anything you like with TCX,
 *	including selling it!  As a gesture of courtesy, please retain
 *	the authorship information in the source code and documentation.
 *
 *	Contact:	Kirk Bailey (bailey@ieee.org)
 *			P.O. Box 1702
 *			Corvallis, OR 97339 (USA)
 *
 *	03/17/87
 */

#include	<stdio.h>
#include	<stdlib.h>
#include	<stddef.h>
#include	<ctype.h>
#include	<string.h>
#include	"tcx.h"
#include	"tcxdecl.h"
#include	"tcxext.h"

/*
 *	Local function declarations.
 */
PRIVATE	VOID			cantop();
PRIVATE	VOID			errfile();
PRIVATE	VOID			warnfile();
PRIVATE	VOID			errptr();
PRIVATE	int			astreq();
PRIVATE	VOID			preprocess();
PRIVATE	VOID			ppcline();
PRIVATE	VOID			doline();
PRIVATE	VOID			keepch();
PRIVATE	VOID			ppaline();

/*
 *	Local data definitions.
 */
PRIVATE	char	*mptr	=	NULL;	/* Input line pointer */

/*
 *	This section handles I/O and preprocessor commands.
 */

/*
 *	Get the next input file.
 */
PUBLIC	VOID
getinfil()
	{
	if(Ninfils--)
		{			/* Any more input files? */
		getfilename(*Nextfil++,Infil,"pp");	/* Default PP file */
		if((Input = fopen(Infil, "r")) == NULL)
			cantop(Infil);
		Lineno = 0;
		Newfile = TRUE;
		(VOID) new_name_space(FALSE);	/* Rewind member name space */
		if(Verbose)
			fprintf(STDERR,"*** Read    %s\n",Infil);
		}
	else
		Eof = TRUE;
	}

/*
 *	Open code output and temporary files.
 */
PUBLIC	VOID
getoutfil()
	{
	char			*cptr;
	int			i;
/*
 *	Open inter-pass temp file if required.
 */
	if(Inter_file)
		{
		Ipfil[0] = '\0';
		if((cptr = getenv(TEMP_PATH)) != NULL)
			{
			if(strlen(cptr) < FILNAMSIZE)
				strcpy(Ipfil,cptr);
			}
/*
 *	Ensure temp directory has trailing delimiter and prune directory name
 *	from input file.
 */
#if	H_OPSYS == MPW
		if(((i = strlen(Ipfil)) > 0) && (Ipfil[i - 1] != ':'))
			strcat(Ipfil,":");
		if((cptr = strrchr(Infil,':')) == NULL)
#else	/* !MPW */
		if(((i = strlen(Ipfil)) > 0) && (Ipfil[i - 1] != '/') &&
			(Ipfil[i - 1] != '\\'))
			{
			strcat(Ipfil,"/");
			}
		if(((cptr = strrchr(Infil,'/')) == NULL) &&
			((cptr = strrchr(Infil,'\\')) == NULL))
#endif	/* MPW */
			{
			cptr = &Infil[-1];
			}
/*
 *	Add input filename to any existing temp directory path.
 */
		if((strlen(Ipfil) + strlen(++cptr)) < FILNAMSIZE)
			strcat(Ipfil,cptr);
		else
			strcpy(Ipfil,cptr);
/*
 *	Prune the filename extension from the temp version of the input file
 *	and substitute the temp file version.
 */
		if((cptr = strrchr(Ipfil,'.')) != NULL)
			*cptr = '\0';
		if((strlen(Ipfil) + strlen(TEMP_NAME)) < FILNAMSIZE)
			strcat(Ipfil,TEMP_NAME);
		else
			strcpy(Ipfil,TEMP_NAME);
#if	CRLF
		if((Interfp = fopen(Ipfil,"w+b")) == NULL)
#else	/* !CRLF */
		if((Interfp = fopen(Ipfil,"w+")) == NULL)
#endif	/* CRLF */
			cantop(Ipfil);
		}
/*
 *	Open output file.
 */
	if((Coutput = fopen(Coutfil, "w")) == NULL)
		cantop(Coutfil);
	header();
	}

/*
 *	Flush the output file.
 */
PUBLIC	VOID
flushout()
	{
	trailer();
	dvrt_out(dvrt_get());		/* Write current diversion out */
	}

/*
 *	Close the code output and temporary files.
 */
PUBLIC	VOID
closeout()
	{
	if(fclose(Coutput) == EOF)
		fatal("Can't close output file");
	
	if(Inter_file)
		{
		if(fclose(Interfp) == EOF)
			fatal("Can't close temporary file");
		if(remove(Ipfil) == -1)
			fatal("Can't remove temporary file");
		}
	}

/*
 *	Remove output file.
 */
PUBLIC	VOID
deleteout()
	{
	if(remove(Coutfil) == -1)
		fatal("Can't remove output file");
	}

/*
 *	Get file name, optionally append extension.
 */
PUBLIC	VOID
getfilename(s,name,ext)
	char			*s;
	char			*name;
	char			*ext;
	{
	register	int	i;
	register	char	*p;

	p = name;
	i = 0;
/*
 *	Following loop also terminates on trailing " of a string constant.
 */
	while(*s && (*s != '\"') && (++i <= FILNAMSIZE))
		*p++ = *s++;
	*p = '\0';
#if	H_OPSYS == VMS
/*
 *	Later - should use VMS filename parsing routines.
 */
	if((strrchr(name,';') == NULL) &&
/*
 *	Don't add .ext if has ;version.
 */
 		(strchr(((strrchr(name,']') == NULL) ? name :
/*
 *	Ignore '.' in directory names.
 */
			strrchr(name,']')),'.') == NULL)
		&& (++i < FILNAMSIZE))
#endif
#if	((H_OPSYS == UNIX) || (H_OPSYS == BSD) || (H_OPSYS == XENIX))
	if((strchr(((strrchr(name,'/') == NULL) ? name : strrchr(name,'/')),
		'.') == NULL)		/* Ignore '.' in directory names */
		&& (++i < FILNAMSIZE))
#endif
#if	H_OPSYS == MPW
	if((strchr(((strrchr(name,':') == NULL) ? name : strrchr(name,':')),
		'.') == NULL)		/* Ignore '.' in directory names */
		&& (++i < FILNAMSIZE))
#endif
#if	H_OPSYS == MS_DOS
	if((strchr(((strrchr(name,'\\') == NULL) ? name : strrchr(name,'\\')),
		'.') == NULL)		/* Ignore '.' in directory names */
		&& (++i < FILNAMSIZE))
#endif
/*
 *	`{' placed here to allow VI's brace-match function to work.
 */
		{
		*p++ =	'.';		/* No extension given */
		while((*ext != '\0') && (++i <= FILNAMSIZE))
			*p++ = *ext++;
		*p = '\0';
		}
	}

/*
 *	Cant open filename error message.
 */
PRIVATE	VOID
cantop(filename)
	char			*filename;
	{
	fprintf(STDERR,"Can't open: %s\n",filename);
	exit(BADEXIT);
	}

/*
 *	Get the next line of input.
 */
PUBLIC	VOID
in_line()
	{
	register	int	c;

	for(;;)
		{			/* Find a non-empty line or EOF */
		if(Eof)
			return;
		kill_line();		/* Clear the input line */
		++Lineno;
		while((c = getc(Input)) != EOF)
			{
			if(c == '\n')
				break;
			else if((Lptr - Line) >= LINEMAX)
				{
				error("Input line too long");
				break;
				}
			else
				*Lptr++ = c;
			}
		*Lptr = '\0';
		if(Lptr != Line)	/* Anything on this line? */
			{
			Lptr = Line;	/* Reset line to beginning */
			return;
			}
		else if(c == EOF)
			{
			fclose(Input);
			getinfil();	/* Get the next input file */
			}
		}
	}

/*
 *	Write a line of assembly language output.
 */
PUBLIC	VOID
outline(opcode,operand,comment)
	int			opcode;
	register	char	*operand;
	register	char	*comment;
	{
/*
 *	Write opcode.
 */
	dvrt_write(opcode);
/*
 *	Write operand field.
 */
	if(operand != NULL)
		{
		while(*operand)
			dvrt_write((int) *operand++);
		}
	dvrt_write('\0');
/*
 *	Write comment field.
 */
	if(comment != NULL)
		{
		while(*comment)
			dvrt_write((int) *comment++);
		}
	dvrt_write('\0');
	}

/*
 *	Out of memory error message.
 */
PUBLIC	VOID
mem_err()
	{
	fatal("Out of memory");
	}

/*
 *	Output a "#line" line if desirable.
 */
PUBLIC	VOID
lineline(line_num)
	int			line_num;
	{
	static	char		lastfile[FILNAMSIZE + 1];
	static	int		lastline;
	char			buffer[FILNAMSIZE + 7 + 20 + 1];

	if(Source_info)
		{
		if(Newfile && strcmp(Infil,lastfile))
			strcpy(lastfile,Infil);
		else
			Newfile = FALSE;
		if(Newfile || (line_num > lastline))
			{
			if(Newfile)
				{
				sprintf(buffer,"#line\t%d\t\"%s\"",line_num,
					Infil);
				}
			else
				sprintf(buffer,"#line\t%d",line_num);
			lastline = line_num;
			Newfile = FALSE;
			outcomment(buffer);
			}
		}
	}

/*
 *	Write a .VAL pseudo-op.
 */
PUBLIC	VOID
assign(name,oldname,plus,offset)
	char			*name;	/* Name of the symbol	*/
	char			*oldname;	/* Used if in another sym */
	int			plus;	/* TRUE if adding symbols */
	INT32			offset;	/* Value to set	*/
	{
	char			buf[LINESIZE];

	sprintf(buf,"%s,%s%s%ld",name,((oldname != NULL) ? oldname : ""),
		((oldname != NULL) ? ((plus) ? "+" : "-") : ""),
		(long) offset);
	outline(TOP_VAL,buf,NULL);
	}

/*
 *	Print an erroneous line with a pointer (^) to the location where the
 *	error was discovered and a message describing the error.
 */
PUBLIC	VOID
error(mesg)
	char			*mesg;
	{
	errfile();
	fprintf(STDERR,"%s\n",mesg);
	errptr();
	}

/*
 *	Print an erroneous line with a pointer (^) to the location where the
 *	warning was discovered and a message describing the warning.
 */
PUBLIC	VOID
warning(level,mesg)
	int			level;
	char			*mesg;
	{
	if(level <= Warn_level)
		{
		warnfile();
		fprintf(STDERR,"%s\n",mesg);
		errptr();
		}
	}

/*
 *	Print an erroneous line with a pointer (^) to the location where the
 *	warning was discovered and a message describing the warning.
 */
PUBLIC	VOID
warnname(level,mesg,name)
	int			level;
	char			*mesg;
	char			*name;
	{
	if(level <= Warn_level)
		{
		warnfile();
		fprintf(STDERR,"%s: %s\n",mesg,name);
		errptr();
		}
	}

/*
 *	Print fatal error message.
 */
PUBLIC	VOID
fatal(mesg)
	char			*mesg;
	{
	errfile();
	fprintf(STDERR,"%s\n",mesg);
	exit(BADEXIT);
	}

/*
 *	"not-yet-implemented" message generator.
 */
PUBLIC	VOID
notyet(mesg)
	char			*mesg;
	{
	errfile();
	fprintf(STDERR,"%s - not yet implemented\n",mesg);
	errptr();
	}

/*
 *	Print error message.
 */
PUBLIC	VOID
errname(mesg,name)
	char			*mesg;
	char			*name;
	{
	errfile();
	fprintf(STDERR,"%s: %s\n",mesg,name);
	errptr();
	}

/*
 *	Print the name of the file which contains the error.
 */
PRIVATE	VOID
errfile()
	{
	++Errcnt;
	fprintf(STDERR,"<%s> @ %d: ",Infil,Lineno);
	}

/*
 *	Print the name of the file which contains the warning.
 */
PRIVATE	VOID
warnfile()
	{
	fprintf(STDERR,"<%s> @ %d: WARNING: ",Infil,Lineno);
	}

/*
 *	Print the input line with a pointer (^) under the error.
 */
PRIVATE	VOID
errptr()
	{
	register	char	*p;

	p = Line;
	if(*p != '\0')
		{
		fprintf(STDERR,"%s\n",p);	/* Print erroneous line */
		while(p < Lptr)
			{		/* Get to error location */
/*lint -e516 Does ANSI know what it is doing? */
			putc(((*p++ == '\t') ? '\t': ' '),STDERR);
/*lint +e516 */
			}
		fprintf(STDERR,"^\n");		/* Print error pointer ^ */
		}
	}

/*
 *	Convert a double to a long and bitch if it can't be done.
 */
PUBLIC	INT32
f_to_l(f,msg)
	double			f;
	BOOL			msg;
	{
/*
 *	Note that the following test is not perfect since ideally you should
 *	allow floating point constants of > 2**31-1 to convert to unsigned
 *	integers.  This is not done since several of the "C" compilers which
 *	are used to host TCX don't handle this very well (everyone gets tired
 *	of exception traps eventually...).  TCX internally handles this
 *	correctly, now if we could just get a few other folks to fix their
 *	code generators... (Microsoft et. al. are you listening?)
 */
	if((f < (double) ((INT32) MIN_LONG_32B)) ||
		(f > (double) ((INT32) MAX_LONG_32B)))
		{
		if(msg)
			{
		error("Floating point number out of signed integral range");
			}
		return ((INT32) 0);
		}
	return ((INT32) f);
	}

/*
 *	Invalid storage class error.
 */
PUBLIC	VOID
scerr()
	{
	error("Invalid storage class");
	}

/*
 *	Invalid type error.
 */
PUBLIC	VOID
typerr()
	{
	error("Invalid type");
	}

/*
 *	Illegal symbol name error.
 */
PUBLIC	VOID
illname()
	{
	error("Illegal symbol name");
	junk();
	}

/*
 *	Name collision error.
 */
PUBLIC	VOID
multidef(name)
	char			*name;
	{
	errname("Already defined",name);
	}

/*
 *	Attempted struct/union redefinition error.
 */
PUBLIC	VOID
redefmem(name)
	char			*name;
	{
	errname("Redefinition of structure/union member",name);
	}

/*
 *	Undeclared struct/union tag error.
 */
PUBLIC	VOID
tagerr(tag)
	char			*tag;
	{
	errname("Undeclared tag",tag);
	}

/*
 *	If the specified punctuation isn't present, bitch.
 */
PUBLIC	VOID
needpunc(c)
	int			c;
	{
	static	char		s[2];

	s[1] = '\0';
	if(! matchc((int) (s[0] = c)))
		errname("Missing punctuation -- assumed present",s);
	}

/*
 *	Illegal use of keyword error.
 */
PUBLIC	VOID
keyerr(keyword)
	char			*keyword;
	{
	errname("Illegal use of a keyword",keyword);
	}

/*
 *	Array size must be specified error.
 */
PUBLIC	VOID
sizerr()
	{
	error("Need explicit array size");
	}

/*
 *	Initializers must be constant expressions error.
 */
PUBLIC	VOID
initerr()
	{
	error("Initializer must be constant expression");
	}

/*
 *	Constant expression required error.
 */
PUBLIC	VOID
cexperr()
	{
	error("Must be constant expression");
	}

/*
 *	Test parse line to see if lit matches the next sequence of characters
 *	and skip over them if it does.
 */
PUBLIC	BOOL
match(lit)
	char			*lit;
	{
	register	int	k;

	blanks();			/* Skip white space */
	if((k = streq(Lptr,lit)) != 0)
		{
		Lptr += k;
		return (TRUE);
		}
	return (FALSE);
	}

/*
 *	Test parse line to see if c matches the next character and skip it.
 */
PUBLIC	BOOL
matchc(c)
	int			c;
	{
	blanks();			/* Skip white space */
	if(*Lptr == c)
		{
		++Lptr;
		return (TRUE);
		}
	return (FALSE);
	}

/*
 *	Test parse line to see if 'c' is the next char.
 */
PUBLIC	BOOL
nextc(c)
	int			c;
	{
	blanks();
	return (*Lptr == c);
	}

/*
 *	Test parse line to see if lit matches the next token and skip over it
 *	if a match is found.
 */
PUBLIC	BOOL
amatch(lit)
	char			*lit;
	{
	blanks();
	if(chks(lit))
		{
		skip();
		return (TRUE);
		}
	return (FALSE);
	}

/*
 *	Skip white space in parse line.
 */
PUBLIC	VOID
blanks()
	{
	register	int	c;

	for(;;)
		{
		while((c = ch()) == '\0')
			{		/* At eol? */
			if(Stopeol)
				return;	/* OK to pass eol? */
			in_line();	/* Get next line */
			if(Eof)
				return;
			preprocess();
			}

		if(isspace(c))
			(VOID) gch();
		else
			return;
		}
	}

/*
 *	Match string against next token in parsing buffer.
 */
PUBLIC	BOOL
chks(s)
	char			*s;
	{
	return (astreq(Lptr, s, (int) strlen(s)));
	}

/*
 *	Test two strings for equality up to the end of the second string.
 */
PUBLIC	int
streq(s1,s2)
	char			*s1;
	char			*s2;
	{
	register	int	len;

	len = 0;
	while(*s2)
		{
		if(*s1++ != *s2++)
			return (0);
		++len;
		}
	return (len);
	}

/*
 *	Test two strings for equality of contents up to len (signed int).
 */
PRIVATE	int
astreq(s1,s2,len)
	char			*s1;
	char			*s2;
	int			len;
	{
	register	int	cnt;

	for(cnt = 0; cnt < len; ++cnt)
		{
		if(*s2 == '\0')
			break;
		if(*s1++ != *s2++)
			return (0);
		}
	return ((issym(*s1)) ? 0 : cnt);
	}

/*
 *	Skip token in input line.
 */
PUBLIC	VOID
skip()
	{
	while(issym(*Lptr))
		++Lptr;			/* Skip all a/n chars */
	}

/*
 *	Throw away garbage in input line.
 */
PUBLIC	VOID
junk()
	{
	static	char		*lastlptr = NULL;

	if(Lptr == lastlptr)
		{
		if(ch() == '\0')
			{
			Stopeol = FALSE;
			blanks();
			}
		else
			(VOID) gch();
		}

	while((! issym(ch())) || (! chkeyword()))
		{
		if(endst() || (ch() == '}'))
			break;		/* Up to EOL or ';' or '}' */
		(VOID) gch();		/* Flush non-a/n characters */
		}

	lastlptr = Lptr;
	}

/*
 *	Is 'c' valid as the first character in a C symbol?
 */
PUBLIC	BOOL
issymf(c)
	register	int	c;
	{
	return (isalpha(c) || c == '_');
	}

/*
 *	Is 'c' valid as any but the first character in a C symbol?
 */
PUBLIC	BOOL
issym(c)
	register	int	c;
	{
#if	H_OPSYS == VMS
	return (isalpha(c) || isdigit(c) || c == '_' || c == '$');
#else	/* H_OPSYS != VMS */
	return (isalpha(c) || isdigit(c) || c == '_');
#endif	/* H_OPSYS == VMS */
	}

/*
 *	Test for end of statement.
 */
PUBLIC	BOOL
endst()
	{
	blanks();			/* Skip white space */
	return ((ch() == ';') || (ch() == '\0'));
	}

/*
 *	Look at current character in parsing buffer.
 */
PUBLIC	int
ch()
	{
	return ((int) *Lptr);
	}

/*
 *	Look at next character in parsing buffer.
 */
PUBLIC	int
nch()
	{
	return (*Lptr ? *(Lptr+1) : 0);
	}

/*
 *	Get current character in parsing buffer.
 */
PUBLIC	int
gch()
	{
	return (*Lptr ? *Lptr++ : 0);
	}

/*
 *	Put a character back in the parsing buffer.
 */
PUBLIC	VOID
putback(c)
	int			c;
	{
	unsigned	int	length;
	register	char	*ptr1;
	register	char	*ptr2;

	if(Lptr > Line)
		*--Lptr = c;
	else
		{
		if((length = strlen(Line)) < LINEMAX)
			{
			ptr2 = &Line[length];
			ptr1 = ptr2 + 1;
			while(ptr1 != &Line[0])
				*ptr1-- = *ptr2--;
			Line[0] = c;
			}
		else
			fatal("Compiler error: putback");
		}
	}

/*
 *	Clear current line from parsing buffer.
 */
PUBLIC	VOID
kill_line()
	{
	Lptr = Line;
	*Lptr = '\0';
	}

/*
 *	Check for #preprocessor commands.
 */
PRIVATE	VOID
preprocess()
	{
	if(! matchc('#'))
		{			/* If it's not a #command... */
		Lptr = Line;		/* Start at beginning of line */
		ppcline();		/* Process the input line */
		}
	else
		{			/* Do a #preprocessor command */
		Stopeol = TRUE; 	/* Look only on this line */
		if(amatch("asm"))
			doasm();	/* Copy assembler thru */
		else if(amatch("line"))
			doline();	/* Process #line statement */
		else if(amatch("pragma"))
			;		/* Ignore unknown "#pragma's" */
		else
			error("Unknown #preprocessor command");
		Stopeol = FALSE;	/* Don't stop at end-of-line */
		kill_line();
		}
	}

/*
 *	Preprocess the "C" input line.
 */
PRIVATE	VOID
ppcline()
	{
	register int		c;
	register char		*savelptr;

	mptr = savelptr = Lptr;		/* Remember where we are on the line */
	while((c = ch()) != 0)
		{
		if(c == ' ' || c == '\t')
			{		/* Eliminate excess whitespace */
			keepch(' ');
			while((ch() == ' ') || (ch() == '\t'))
				(VOID) gch();
			}
		else
			{
			blanks();
			if(issymf(c))
				{		/* Is it a symbol? */
				while(issym(ch()))
					keepch(gch());	/* Keep text */
				}
			else if((c == '/') && (*(Lptr+1) == '*'))
				{
				(VOID) gch();
				(VOID) gch();	/* Discard comments */
				while((*Lptr != '*') || (*(Lptr+1) != '/'))
					{
					if(gch() == 0)	/* If EOL get */
						in_line();	 /* nxt line */
					if(Eof)
						{
						error("Unclosed comment");
						break;
						}
					}
				(VOID) gch();
				(VOID) gch();	/* Toss comment end */
				}
			else if(c == '\"')
				{
/* Pass string through */	keepch(gch());
/* Unchanged */			while(ch() != '\"')
					{
					if(ch() == 0)
						{
/* Backup and use last cell again */		--mptr;
						error("Missing quote");
						break;
						}
					if(ch() == '\\')
/* Pass an escape character */			{
						if(nch())	/* not EOL */
							keepch(gch());
						else
/* More long lines */					{
							in_line();
							continue;
							}
						}
					keepch(gch());
					}
				(VOID) gch();
				keepch('\"');
				}
			else if(c == '\'')
				{
/* Pass char constants */	keepch(gch());
/* Unchanged */			while(ch() != '\'')
					{
					if(ch() == 0)
						{
/* Backup and use last cell again */		--mptr;
						error("Missing apostrophe");
						break;
						}
/* Pass escape character */			if(ch() == '\\')
						keepch(gch());
					keepch(gch());
					}
				(VOID) gch();
				keepch('\'');
				}
			else
				keepch(gch());	/* Pass all else unchanged */
			}
		}
	*mptr = '\0';
	Lptr = savelptr;		/* Restore original position */
	}

/*
 *	Process input "#line" statements.
 */
PRIVATE	VOID
doline()
	{
	ppcline();		/* Strip comments */
	blanks();
	if(! isdigit(ch()))	/* New line number? */
		error("#line number must be decimal");
	else
		{
		Lineno = 0;
		Newfile = TRUE;	/* Emit an entire line */
		while(isdigit(ch()))
			Lineno = 10 * Lineno + gch() - '0';
		--Lineno;	/* Adjust for next line */
		skip();
		if(! endst())	/* Optional filename? */
			{
			if(matchc('\"'))
				getfilename(Lptr,Infil,"pp");
			else
				error("Illegal #line format");
			}
		}
	}

/*
 *	Put a char in input buffer
 */
PRIVATE	VOID
keepch(c)
	int			c;
	{
	*mptr++ = c;
	}

/*
 *	Maximum length checked "strcat".  Returns updated pointer.
 */
PUBLIC	char	*
safecat(cur,str,end)
	register char		*cur;
	register char		*str;
	register char		*end;
	{
	while((cur < end) && (*str != '\0'))
		*cur++ = *str++;
	return (cur);
	}

/*
 *	Like "safecat" but appends the string equivalent of an INT32.
 */
PUBLIC	char	*
safedec(cur,num,end)
	char			*cur;
	INT32			num;
	char			*end;
	{
	char			longbuf[80];

	sprintf(longbuf,"%ld",(long) num);
	return (safecat(cur,longbuf,end));
	}

/*
 *	Like "safecat" but appends the string equivalent of an double.
 */
PUBLIC	char	*
safefloat(cur,num,end)
	char			*cur;
	double			num;
	char			*end;
	{
	char			floatbuf[80];

	sprintf(floatbuf, "%.20g", num);
	return (safecat(cur,floatbuf,end));
	}

/*
 *	Convert signed or unsigned integral constant into floating point.
 *	Input is a pointer to the "constant" non-terminal leaf node which
 *	points to the terminal holding the actual constant.
 */
PUBLIC	double
l_to_f(iconstant)
	struct	op_node		*iconstant;
	{
	if(iconstant->right.ptr.terminal->type & SIGNED_NODE)
		return ((double) iconstant->right.ptr.terminal->val.integral);
	else
		return ((double) iconstant->right.ptr.terminal->val.uintegral);
	}

/*
 *	Output a string as an assembly language comment.
 */
PUBLIC	VOID
outcomment(str1)
	char			*str1;
	{
	outline(TOP_COMMENT,NULL,str1);
	}

/*
 *	Format up and output and assembly language instruction.
 */
PUBLIC	VOID
instruction(opcode,absrel,symone,symtwo,offset,comment)
	int			opcode;
	char			*symone;
	char			*symtwo;
	int			absrel;
	INT32			offset;
	char			*comment;
	{
	char			cbuf[NAMEMAX];
	char			buf[LINESIZE];
	char			*cur;
	char			*end;

	cur = buf;
	end = buf + (sizeof(buf) - 1);

	if((symone != NULL) || (symtwo != NULL) || (offset != 0))
		{
		if(symone != NULL)
			{
			if(absrel == REL)
				cur = safecat(cur,"@",end);
			cur = safecat(cur,symone,end);
			}
		if(symtwo != NULL)
			{
			cur = safecat(cur,"-",end);
			if(absrel == REL)
				cur = safecat(cur,"@",end);
			cur = safecat(cur,symtwo,end);
			}
		if(offset > 0)
			cur = safecat(cur,"+",end);
		if(offset != 0)
			cur = safedec(cur,offset,end);
		}
	else if(absrel != NONE)
		cur = safecat(cur,"0",end);
	*cur = '\0';

	if(comment != NULL)		/* Add comment delimiter to field */
		{
		cur = cbuf;
		end = cbuf + (sizeof(cbuf) - 1);
		cur = safecat(cur,comment,end);
		*cur = '\0';
		comment = cbuf;
		}
	outline(opcode,buf,comment);
	}

/*
 *	Print assembler info before any code is generated.
 */
PUBLIC	VOID
header()
	{
	char			buf[80];

	sprintf(buf,";\tCompiled by TCX %s",Version);
	outcomment(buf);

	if(Cpu_type & (T800|T8EX))
		no_oper(TOP_T800);
	else
		{
		if(Cpu_type & (T212|T2EX))
			no_oper(TOP_T212);
		else if(Cpu_type & (T414|T4EX))
			no_oper(TOP_T414);
		else
			no_oper(TOP_ALL);
		no_oper(TOP_EMULATE);
		}
	if(Rel_only)
		no_oper(TOP_REL);
	if(Static_link)
		instruction(TOP_EXT,ABS,SLBNAME,NULL,(INT32) 0,NULL);

	segment(SEG_UNINIT);		/* Default to uninit. data module */
	}

/*
 *	Print any assembler stuff needed after all code.
 */
PUBLIC	VOID
trailer()
	{
	no_oper(TOP_END);
	}

/*
 *	"#asm" preprocessor command.
 */
PUBLIC	VOID
doasm()
	{
	for(;;)
		{
		in_line();		/* Get next line */
		if(Eof)
			break;
		if(matchc('#'))
			{
			if(amatch("endasm"))
				break;
			else if(amatch("line"))
				{
				doline();
				continue;
				}
			}
/*
 *	Output assembly language line as comment to code generator.
 */
		lineline(Lineno);

		dvrt_write((int) TOP_COMMENT);	/* Indicate it is a comment */
		dvrt_write('\0');	/* Dummy operand field */
		ppaline();		/* Output asm line in comment field */
		dvrt_write('\0');	/* Terminate comment field */
		}
	}

/*
 *	Preprocess the assembly input line and output the result.
 */
PRIVATE	VOID
ppaline()
	{
	register int		c;
	register char		*cptr;
	char			*cur;
	char			*end;
	char			name[NAMEMAX+1];
	INT32			offset;
	int			sc;
	struct	id_node		*sym;

	Lptr = Line;			/* Restart at beginning of line */
	while((c = ch()) != 0)
		{
		if((c == ' ') || (c == '\t'))
			{		/* Eliminate excess whitespace */
			dvrt_write((int) ' ');
			while((ch() == ' ') || (ch() == '\t'))
				(VOID) gch();
			}
		else if(c == '[')
			{		/* Substitute local variable offsets */
			(VOID) gch();	/* Toss opening bracket */
			cptr = name;
			if(ch() == '?')
				*cptr++ = gch();	/* Allow "?slb" */
			if(((cptr = chksym(cptr)) == NULL) || (*cptr != ']'))
				{
				error("Badly formatted variable name");
				break;
				}
			if((sym = lookup_node(name,NS_VAR)) == NULL)
				{
				errname("Undefined variable",name);
				break;
				}
			sc = sym->info.storage_class;
			if((sc != SC_AUTO) && (sc != SC_ARG) && (sc != SC_REG))
				{
			error("Variable must be argument/auto/register");
				break;
				}
/*
 *	Compute workspace relative expression to access variable.
 */
			cur = name;
			end = name + (sizeof(name) - 1);
			cur = safecat(cur,"(",end);
			cur = safecat(cur,WORKSPACE,end);
			if((offset = -word(-sym->address.offset)) >= 0L)
				cur = safecat(cur,"+",end);
			cur = safedec(cur,offset,end);
			cur = safecat(cur,")",end);
			*cur = '\0';
/*
 *	Substitute for variable name.
 */
			cptr = name;
			while(*cptr)
				dvrt_write((int) *cptr++);
			while(ch() != ']')
				(VOID) gch();	/* Toss local variable name */
			(VOID) gch();	/* Toss trailing ']' */
			}
		else if(c == '\"')
			{		/* Pass strings on through */
			dvrt_write((int) gch());
			while(ch() != '\"')
				{	/* Unchanged */
				if(ch() == 0)
					{
					error("Missing quote");
					break;
					}
				if(ch() == '\\')
					{	 /* Pass an escape char */
					dvrt_write((int) gch());
					}
				if(ch())
					dvrt_write((int) gch());
				}
			if(ch())
				dvrt_write((int) gch());	/* Final " */
			}
		else if(c == '\'')
			{		/* Pass char constants on through */
			dvrt_write((int) gch());
			while(ch() != '\'')
				{	/* Unchanged */
				if(ch() == 0)
					{
					error("Missing apostrophe");
					break;
					}
				if(ch() == '\\')	/* Pass escape char */
					dvrt_write((int) gch());
				if(ch())
					dvrt_write((int) gch());
				}
			if(ch())
				dvrt_write((int) gch());	/* Final ' */
			}
		else if(c == ';')
			{		/* Rest of line is a comment */
			cptr = Lptr;
			while(*cptr)
				dvrt_write((int) *cptr++);
			break;
			}
		else			/* Pass all else unchanged */
			dvrt_write((int) gch());
		}
	}

/*
 *	Load constant value on to the top of the stack.
 */
PUBLIC	VOID
loadconstant(value)
	INT32			value;
	{
	abs_oper(TOP_LDC,(INT32) value);
	}

/*
 *	Print a global label for a variable or function name.
 */
PUBLIC	VOID
glblabel(name,sc)
	char			*name;
	int			sc;
	{
	if((sc == SC_GLOBAL) || (sc == SC_EXTERN))
		instruction(TOP_PUB,ABS,name,NULL,(INT32) 0,NULL);
	outline(TOP_LABEL,name,NULL);
	}

/*
 *	Change to specified segment.
 */
PUBLIC	VOID
segment(seg)
	int			seg;
	{
	static	int 		pseg = SEG_NONE;	/* Previous seg */

	if((seg != pseg) &&
		((pseg == SEG_NONE) || (SegMap[seg] != SegMap[pseg])))
		{
		abs_oper(TOP_MOD,(INT32) SegMap[seg]);
		}
	pseg = seg;
	}

/*
 *	Align assembler's location counter to next word boundary.
 */
PUBLIC	VOID
align()
	{
	no_oper(TOP_ALIGN);
	}

/*
 *	Default initialization in a data segment.
 */
PUBLIC	INT32
definit(amt)
	INT32			amt;
	{
	if(amt > 0)
		abs_oper(TOP_DS,(INT32) amt);

	return (amt);
	}

/* End of TCX8.C */
