/*
 *	TCX Compiler Environment Header File
 *
 *	THIS FREE SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 *	WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *	SPECIFICALLY DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR OTHER
 *	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *	TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 *	OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *	THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 *	USE OF THIS SOFTWARE OR ANY DERIVATIVE WORK, EVEN IF ADVISED OF THE
 *	POSSIBILITY OF SUCH DAMAGE.
 *
 *	This program is hereby placed in the public domain.  In
 *	contrast to other claims of "public domain", this means no
 *	copyright is claimed and you may do anything you like with TCX,
 *	including selling it!  As a gesture of courtesy, please retain
 *	the authorship information in the source code and documentation.
 *
 *	Contact:	Kirk Bailey (bailey@ieee.org)
 *			P.O. Box 1702
 *			Corvallis, OR 97339 (USA)
 *
 *	03/17/87
 *
 *	This file defines the host and target environments for
 *	the TCX compiler and library.  It is included in the
 *	other source files to enable environment-dependent code
 *	to be selected where necessary.  The definitions of the
 *	names currently used are:
 *
 *		H_OPSYS	- the host operating system that TCX will
 *			  run under
 *
 *	The host environments are:
 *
 *		H_OPSYS
 *		-------
 *		MS_DOS	- TCX will run under MSDOS
 *		VMS	- TCX will run under VAX/VMS
 *		UNIX	- TCX will run under Unix
 *		BSD	- TCX will run under "new" BSD Unix variants
 *		MPW	- TCX will run under MPW on the Mac
 *		XENIX	- TCX will run under Microsoft XENIX on PC's
 *
 *	The target environment is always assumed to be a Inmos Transputer
 *	processor.
 *
 */

#define	MS_DOS		1
#define	VMS		2
#define	UNIX		3
#define	BSD		4
#define	MPW		5
#define	XENIX		6

/*
 *	Define host environment.
 */

#define	H_OPSYS		MS_DOS

/*
 *	Some versions of BSD are missing functions
 */

#if	H_OPSYS == BSD

#define	strchr			index
#define	strrchr			rindex
#define	memcpy(dst, src, len)	bcopy(src, dst, len)

#endif	/* BSD */

/*
 *	The parameters defined here tailor the TCX compiler to the
 *	characteristics of the host environment where it will run.  This
 *	includes initial values for global variables to fit the size of the
 *	environment and features which may be included in TCX.
 */

#define	TEMP_NAME	".1"		/* Filename ext. for temp file	*/
#define	TEMP_PATH	"TMP"		/* Env. var. for temp directory	*/

/*
 *	CRLF defines whether the system needs to convert CR,LF pairs into
 *	linefeeds for text files (needed on many micro's).  The effect of this
 *	switch (if TRUE), is to make "fopen" calls contain a trailing "b" for
 *	binary files.  For most UNIX installations this	switch should be FALSE.
 */

#if	((H_OPSYS == MS_DOS) || (H_OPSYS == MPW))
#define	CRLF		1
#else
#define	CRLF		0
#endif	/* MS_DOS || MPW */

#if	H_OPSYS == MS_DOS

typedef	long		INT32;		/* 32 bit integer		*/
typedef unsigned long	UINT32;		/* 32 bit unsigned integer	*/
typedef	int		BOOL;		/* Boolean variable		*/

#define FILNAMSIZE	64		/* Size of largest pathname	*/
#endif	/* MS_DOS */

#if	((H_OPSYS == VMS) || (H_OPSYS == UNIX) || (H_OPSYS == MPW) || (H_OPSYS == BSD) || (H_OPSYS == XENIX))

typedef	long		INT32;		/* 32 bit integer		*/
typedef unsigned long	UINT32;		/* 32 bit unsigned integer	*/
typedef	int		BOOL;		/* Boolean variable		*/

#define FILNAMSIZE	128		/* Size of largest pathname */
#endif	/* VMS || UNIX || MPW || BSD || XENIX */

/*
 *	The parameters defined here are used to generate output acceptable to
 *	the target assembler.
 */

#define ASM_FEXT	"tal"		/* Lower case file extension	*/
#define ASM_UNIQ	'\0'		/* Prefix/Suffix to make asm	*/
					/* 	names unique.		*/
#define	WORKSPACE	"?ws"		/* Symbol for val. of WS size	*/
#define	RSLINK		"?rslink"	/* Struct return link symbol	*/
#define	STRUCTBASE	"?sb"		/* Symbol for offset from WS to	*/
					/* struct func ret temp space	*/
#define	TEMPBASE	"?tb"		/* Symbol for offst frm WS->tmp	*/

#define	SLBMODULE	250		/* First of 2 static link mods	*/
#define	SLBNAME		"?slb"		/* Global static link basename	*/

#define	SCOPE_END	"?se"		/* Debugging scope end symbol	*/
#define	SCOPE_LABEL	"?sl"		/* Label scope symbol		*/
#define	SCOPE_START	"?ss"		/* Debugging scope start symbol	*/

#define	TYPE_END	"?te"		/* Debugging type end symbol	*/
#define	TYPE_START	"?ts"		/* Debugging type start symbol	*/

/* End TCXENV.H */
