/************************************************************************/
/*									*/
/*	File:	tcxtypes.h						*/
/*									*/
/*	System data type definitions for the Transputer compiler (TCX),	*/
/*	and debugger (NV).						*/
/*									*/
/*		Written by Kirk Bailey of Logical Systems.		*/
/*									*/
/*	THIS FREE SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 *	WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *	SPECIFICALLY DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR OTHER
 *	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *	TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 *	OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *	THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 *	USE OF THIS SOFTWARE OR ANY DERIVATIVE WORK, EVEN IF ADVISED OF THE
 *	POSSIBILITY OF SUCH DAMAGE.
 *
 *	This program is hereby placed in the public domain.  In
 *	contrast to other claims of "public domain", this means no
 *	copyright is claimed and you may do anything you like with TCX,
 *	including selling it!  As a gesture of courtesy, please retain
 *	the authorship information in the source code and documentation.
 *
 *	Contact:	Kirk Bailey (bailey@ieee.org)
 */
/*									*/
/*				02/01/92				*/
/*									*/
/************************************************************************/

/*
 *	Architecture-specific information for sizes of base types.
 */

#define	SIZEOF_CHAR	1		/* Size of a character		*/
#define	SIZEOF_FLOAT	4		/* Size of a float type		*/
#define	SIZEOF_S_DBLE	4		/* Size of 32 bit double	*/
#define	SIZEOF_S_LDBLE	4		/* Size of 32 bit long double	*/
#define	SIZEOF_L_DBLE	8		/* Size of 64 bit double	*/
#define	SIZEOF_L_LDBLE	8		/* Size of 64 bit long double	*/

/*
 *	Define sizes for target machine.
 */

#define	WORD_SIZE_16B	2		/* Word size (bytes) for T212..	*/
#define MAX_INT_16B	0x00007FFFL	/* Largest signed int (2**15)-1 */
#define	MAX_UINT_16B	0x0000FFFFL	/* Largest unsigned (2**16)-1	*/
#define MIN_INT_16B	0xFFFF8000L	/* Smallest signed int -(2**15) */
#define MAX_LONG_16B	MAX_INT_16B	/* Largest signed long (2**15)-1*/
#define	MIN_LONG_16B	MIN_INT_16B	/* Smallest signed long -(2**15)*/

#define	WORD_SIZE_32B	4		/* Default word size in bytes	*/
#define MAX_INT_32B	0x7FFFFFFFL	/* Largest signed int (2**31)-1 */
#define	MAX_UINT_32B	0xFFFFFFFFL	/* Largest unsigned (2**32)-1	*/
#define MIN_INT_32B	0x80000000L	/* Smallest signed int -(2**31) */
#define MAX_LONG_32B	MAX_INT_32B	/* Largest signed long (2**31)-1*/
#define	MIN_LONG_32B	MIN_INT_32B	/* Smallest signed long -(2**31)*/

/*
 *	Storage classes.
 */

#define SC_NONE 	0		/* No explicit storage class	*/
#define	SC_CONSTANT	1		/* Constant value		*/
#define SC_GLOBAL	2		/* Unspecified external		*/
#define SC_ST_GLB	3		/* "static" ... (global)	*/
#define SC_EXTERN	4		/* "extern" data/function decl.	*/
#define	SC_ST_EXT	5		/* "static" function decl.	*/
#define SC_STATIC	6		/* "static" ... (local)		*/
#define SC_AUTO 	7		/* "auto" ... or misc local	*/
#define SC_ARG		8		/* Like "auto", but for args	*/
#define SC_REG		9		/* "register" ...		*/
#define	SC_TYPEDEF	10		/* Typedef "storage class"	*/
#define	SC_LITERAL	11		/* For tagging literals		*/
#define	SC_RSTEMP	12		/* For structure return scratch	*/

/*
 *	Name spaces
 *
 *	All symbols belong to some name space.  Some of the name spaces are
 *	"well known" such as keywords and global symbols.  Others are private,
 *	like symbol tags and such.
 */

#define	NS_CKEYWORD	1		/* C keywords			*/
#define	NS_CTYPES	2		/* C type keywords		*/
#define	NS_CSTORAGE	3		/* C storage class keywords	*/
#define	NS_TAG		4		/* Structure/union/enum tags	*/
#define	NS_LABEL	5		/* Local labels for "goto"	*/
#define	NS_OTHER	6		/* Vars/funcs/typedefs/enum	*/
#define	NS_EXTERNS	7		/* Referenced externals		*/
#define	NS_FP32LITS	8		/* For 32 bit FP literals 	*/
#define	NS_FP64LITS	9		/* For 64 bit FP literals	*/
#define	NS_IFUNC	10		/* "_BUILTIN_" function names	*/
/* not used		11						*/
#define	NS_MEMBER	12		/* First member namespace	*/

#define	NS_VAR		NS_OTHER	/* Variables are also 'other'	*/
#define	NS_FUNC		NS_OTHER	/* Function names		*/
#define	NS_ENUM		NS_OTHER	/* Enumerations			*/
#define	NS_TYPEDEF	NS_OTHER	/* Typedefs			*/

/*
 *	Type field encodings.
 */

#define	LAST_NODE	0x80000000L	/* TRUE if last node in list	*/
#define	UNSIGNED_NODE	0x40000000L	/* TRUE if unsigned		*/
#define	SIGNED_NODE	0x20000000L	/* TRUE if signed		*/
#define	CONST_NODE	0x10000000L	/* TRUE if const node		*/

#define	VOID_NODE	0x08000000L	/* Type is "void"		*/
#define	CHAR_NODE	0x04000000L	/* Type is "char"		*/
#define	SHORT_NODE	0x02000000L	/* Type is "short"		*/
#define	INT_NODE	0x01000000L	/* Type is "int"		*/
#define	LONG_NODE	0x00800000L	/* Type is "long"		*/
#define	FLOAT_NODE	0x00400000L	/* Type is "float"		*/
#define	DOUBLE_NODE	0x00200000L	/* Type is "double"		*/
#define	LDOUBLE_NODE	0x00100000L	/* Type is "long double"	*/
#define	PTR_NODE	0x00080000L	/* Type is pointer-to		*/
#define	STRUCT_NODE	0x00040000L	/* Type is "struct"		*/
#define	UNION_NODE	0x00020000L	/* Type is "union"		*/
#define	MEMBER_NODE	0x00010000L	/* Type is "member"		*/
#define	BITFIELD_NODE	0x00008000L	/* Type is a "bit field"	*/
#define	ENUM_NODE	0x00004000L	/* Type if "enum"		*/
#define	ARRAY_NODE	0x00002000L	/* Type is "array"		*/
#define	FUNC_NODE	0x00001000L	/* Type is "function"		*/

#define	LABEL_NODE	0x00000800L	/* Node represents a goto label	*/
#define	LABEL_DEFINED	0x00000400L	/* The label has been defined	*/

#define	TENTATIVE_NODE	0x00000200L	/* Tentative node for func pars	*/
#define	REFERENCED_NODE	0x00000100L	/* EXTERN symbol has been refd	*/
#define	OFFSET_ASSIGNED	0x00000080L	/* Set when stk offset assigned	*/
#define	INIT_NODE	0x00000040L	/* Node has been initialized	*/
#define	LITERAL_NODE	0x00000020L	/* Set if a literal node	*/

#define	INTEGRAL_NODE	(CHAR_NODE|SHORT_NODE|INT_NODE|LONG_NODE|ENUM_NODE|BITFIELD_NODE)
#define	FP_NODE		(FLOAT_NODE|DOUBLE_NODE|LDOUBLE_NODE)
#define	ARITH_NODE	(INTEGRAL_NODE|FP_NODE)
#define	SCALAR_NODE	(ARITH_NODE|PTR_NODE)
#define	PTYPE_NODE	(ARRAY_NODE|PTR_NODE)
#define	SU_NODE		(STRUCT_NODE|UNION_NODE)
#define	TYPE_MASK	(INTEGRAL_NODE|FP_NODE|PTYPE_NODE|SU_NODE|VOID_NODE|FUNC_NODE)
