/************************************************************************/
/*									*/
/*	File:	make.h							*/
/*									*/
/*	Include file for the MAKE program.				*/
/*									*/
/*	Written by Greg Yachuk.  Placed in public domain 10/1/88.	*/
/*	Modeled after a MAKE written by Neil Russell which in turn was	*/
/*	modeled after UN*X MAKE...					*/
/*									*/
/*	Modified/maintained by Kirk Bailey of Willamette RF, Inc.	*/
/*									*/
/*	THIS FREE SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR	*/
/*	IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED	*/
/*	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR	*/
/*	PURPOSE ARE SPECIFICALLY DISCLAIMED. IN NO EVENT SHALL THE	*/
/*	AUTHOR OR OTHER CONTRIBUTORS BE LIABLE FOR ANY DIRECT,		*/
/*	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL	*/
/*	DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF		*/
/*	SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR	*/
/*	BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF	*/
/*	LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT	*/
/*	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF	*/
/*	THE USE OF THIS SOFTWARE OR ANY DERIVATIVE WORK, EVEN IF	*/
/*	ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.			*/
/*									*/
/*	This program is hereby placed in the public domain.  In		*/
/*	contrast to other claims of "public domain", this means no	*/
/*	copyright is claimed and you may do anything you like with	*/
/*	MAKE, including selling it!  As a gesture of courtesy, please	*/
/*	retain the authorship information in the source code and	*/
/*	documentation.							*/
/*									*/
/*	Contact:	Kirk Bailey (bailey@ieee.org)			*/
/*									*/
/************************************************************************/

/*
 *	Macro definitions.
 */

#define	VERSION		"Willamette RF, Inc. Version 02.1"
#define MAKEINI 	"default.mk"


#define	READBUFSIZE	2048		/* Makefile input read buffer size */
#define	MAXFNSIZE	80		/* Maximum filename size */
#define	MAXLNSIZE	512		/* Maximum input line size */
#define	MAXNEGTIME	((long) 0x80000000L)
#define	MAXSYMSIZE	512		/* Maximum macro symbol size */
#define	REGISTER	register

#ifndef	TRUE
#define	TRUE		1
#define	FALSE		0
#endif

#ifdef	__WATCOMC__
#undef	MSDOS				/* Assume Windows not DOS */
#endif	/* __WATCOMC__ */

#if	defined(MSDOS) || defined(__TURBOC__) || defined(__WATCOMC__)
#define	FILE_SEPARATOR	"/\\"
#define	PATH_SEPARATOR	";"
#define	STDERR		stdout
#else	/* !(MSDOS || __TURBOC__ || __WATCOMC__) */
#define	FILE_SEPARATOR	"/"
#define	PATH_SEPARATOR	":"
#define	STDERR		stderr
#endif	/* MSDOS || __TURBOC__ || __WATCOMC__ */

#define	append_preq(t,p) (fileptr*)append_node((char**)(t),(char**)(p),(int)sizeof(fileptr*))
#define	append_shell(t,s) (shellptr*)append_node((char**)(t),(char**)(s),(int)sizeof(shellptr*))
#define	equal(s,t)	(!strcmp((s),(t)))
#define	get_target(t)	hash_target((t),NULL)
#define	get_file(f)	hash_file((f),NULL)
#define	tfree(t)	free((void *) t)
#define	tnew(t)		((t *) talloc((int) sizeof(t)))
#define	tstrcat(s,p)	(strcat(strcpy(talloc((int) (strlen(s)+strlen(p)+1)),(s)),(p)))
#define	tstrcpy(s)	(strcpy(talloc((int) (strlen(s)+1)), (s)))

#if	! (defined(MSDOS) || defined(__TURBOC__) || defined(__WATCOMC__))
#define	P_WAIT		1
#define	P_NOWAIT	2
#define	P_OVERLAY	3
#endif	/* ! (MSDOS || __TURBOC__ || __WATCOMC__) */

/*
 *	Macro priority selection.
 */

#define	COMMAND_MACRO	4		/* Command line (highest priority) */
#define	HIGH_MACRO	3		/* Either user file or environment */
#define	LOW_MACRO	2		/* Either user file or environment */
#define	DEFAULT_MACRO	1		/* Default makefile (low priority) */

/*
 *	Type definitions.
 */

typedef	unsigned short t_mask;

typedef	struct	targnode
	{
	t_mask		tmask;		/* Mask to avoid string compares */
	struct	targnode *tnext;	/* Next target in global target list */
	struct	filenode *tfile;	/* File node for this target */
	struct	filenode **tpreq;	/* Pre-req list for this target */
	struct	shellnode **tshell;	/* Command list for this target */
	} targnode,*targptr;

typedef	struct	filenode
	{
	t_mask		fmask;		/* Mask to avoid string compares */
	char		*fname;		/* Name of this file (targ, preq...) */
	long		ftime;		/* Last MODIFY time for this file */
	struct	filenode *fnext;	/* Next file node in global file lst */
	} filenode,*fileptr;

typedef	struct	shellnode
	{
	char		*scmd;		/* Text of command */
	unsigned	s_silent:1;	/* Don't echo before executing */
	unsigned	s_ignore:1;	/* Ignore exit status */
	unsigned	s_shell:1;	/* Force spawning of command.com */
	unsigned	s_force:1;	/* Force execution */
	struct	shellnode *snext;	/* Next shell node for a target */
	struct	shellnode *slink;	/* Next shell node in global list */
	} shellnode,*shellptr;

typedef	struct	symnode
	{
	t_mask		smask;		/* Mask to avoid string compares */
	char		*sname;		/* Name of a symbol */
	char		*svalue;	/* Value of a symbol */
	int		sprior;		/* Macro value priority */
	struct	symnode *snext;		/* Next symbol node in global list */
	} symnode,*symptr;

struct	msi				/* Make program state information */
	{
	int		depend;		/* -d option */
	int		dispcount;	/* Used for -D option */
	int		display;	/* -D option */
	fileptr		file_list;	/* List of file nodes */
	targptr		first_targ;	/* 1st target, if nothing explicit */
	char		**flist;	/* Command line make files */
	int		fmax;		/* Max size of flist */
	int		ignore;		/* -i option */
	int		inlen;		/* Input char counter in "input" */
	char		*input;		/* Current input line pointer */
	int		mac_env;	/* Environment macro priority */
	int		mac_usr;	/* User makefile macro priority */
	char		**mlist;	/* Command line macros */
	int		mmax;		/* Max size of mlist */
	int		noexec;		/* -n option */
	long		now;		/* Time at startup */
	char		*olds;		/* Token internal pointer */
	int		print;		/* -p option */
	int		readdef;	/* -r option */
	shellptr	shell_list;	/* List of shell nodes */
	int		silent;		/* -s option */
	targptr		suffix_targ;	/* .SUFFIXES target pointer */
	symptr		symbol_list;	/* List of symbol nodes */
	targptr		target_list;	/* List of target nodes */
	char		**tlist;	/* Command line targets */
	int		tmax;		/* Max size of tlist */
	int		touch;		/* -t option */
	jmp_buf		up_one;		/* 'longjmp' target */
	};

struct statitem				/* Cached file time information */
	{
	struct statitem	*next;
	long		filetime;
	char		fname[1];	/* Extra space malloc'd as needed */
	};

/*
 *	External data definitions.
 */

extern	struct	msi	ms;		/* Make program internal state */
extern	int		nesting;	/* Nesting level for recursive MAKEs */
extern	char		*shell_cmds[];	/* List of commands requiring shell */

/*
 *	External function definitions.
 */
extern	fileptr		add_file(/* char *name */);
extern	void		add_metas(/* char *basename,char *preqname,
				char *targname */);
extern	shellptr	add_shell(/* char *input */);
extern	void		add_symbol(/* int priority,char *name,
				char *value */);
extern	targptr		add_target(/* char *name */);
extern	char		**append_node(/* char **node,char **adds,
				int size */);
extern	char		*breakout(/* char *input */);
extern	void		build(/* targptr targp,char *targname */);
extern	long		curr_time(/* void */);
extern	int		default_rule(/* char *targname,int worry,
				int mustbuild */);
extern	void		display_prereq(/* char *targname,long targtime,
				char *preqname,long preqtime */);
extern	void		dump_data(/* void */);
extern	long		file_time(/* char *fname */);
extern	FILE		*fopenp(/* char *fname,char *type */);
extern	symptr		get_symbol(/* char *name */);
extern	char		**grow_list(/* char **list,int *len */);
extern	fileptr		hash_file(/* char *name,t_mask *maskp */);
extern	t_mask		hash_name(/* char *name */);
extern	targptr		hash_target(/* char *name,t_mask *maskp */);
extern	void		link_targs(/* targptr *targs,fileptr *preqs,
				shellptr *shells */);
extern	void		main(/* int argc,char **argv */);
extern	void		make_args(/* int argc,char **argv */);
extern	int		new_make(/* char **argv */);
extern	void		nexit(/* int n */);
/*lint -function(exit,nexit) */
#if	! (defined(MSDOS) || defined(__TURBOC__) || defined(__WATCOMC__))
extern	int		spawnvp(/* int mode,char *path,char **args */);
#endif	/* ! (MSDOS || __TURBOC__ || __WATCOMC__) */
extern	struct statitem *statlook(/* char *fname */);
extern	void		terror(/* char *s1,char *s2 */);
/*lint -function(exit,terror) */
extern	char		*tgets(/* FILE *fd */);
extern	char		*token(/* char *s,char *sep */);
extern	char		**tokenize(/* char *input */);
extern	void		touch_file(/* char *targname */);
extern	char   		*trealloc(/* char *s,int n */);
extern	char   		*tstrncpy(/* char *s,int n */);
extern	void		usage(/* void */);

/*
 *	The following functions may be called only for side effects.
 */
/*lint +fvr */

extern	int		add_macro(/* int priority,char *input */);
extern	int		make(/* char *targname,int worry */);
extern	int		parse(/* int priority,FILE *fd */);
extern	char		*talloc(/* int n */);

/*lint -fvr */
