/************************************************************************/
/*									*/
/*	File:	make1.c							*/
/*									*/
/*	Top level control for MAKE operation.				*/
/*									*/
/*	Written by Greg Yachuk.  Placed in public domain 10/1/88.	*/
/*	Modeled after a MAKE written by Neil Russell which in turn was	*/
/*	modeled after UN*X MAKE...					*/
/*									*/
/*	Modified/maintained by Kirk Bailey of Willamette RF, Inc.	*/
/*									*/
/*	Modified: 9-Apr-90. G. Oliver					*/
/*		Caches stat() calls into a local structure so redundant	*/
/*		calls to the operating system are eliminated.  touch()	*/
/*		function updates the time in the data structure.	*/
/*									*/
/*	THIS FREE SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR	*/
/*	IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED	*/
/*	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR	*/
/*	PURPOSE ARE SPECIFICALLY DISCLAIMED. IN NO EVENT SHALL THE	*/
/*	AUTHOR OR OTHER CONTRIBUTORS BE LIABLE FOR ANY DIRECT,		*/
/*	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL	*/
/*	DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF		*/
/*	SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR	*/
/*	BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF	*/
/*	LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT	*/
/*	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF	*/
/*	THE USE OF THIS SOFTWARE OR ANY DERIVATIVE WORK, EVEN IF	*/
/*	ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.			*/
/*									*/
/*	This program is hereby placed in the public domain.  In		*/
/*	contrast to other claims of "public domain", this means no	*/
/*	copyright is claimed and you may do anything you like with	*/
/*	MAKE, including selling it!  As a gesture of courtesy, please	*/
/*	retain the authorship information in the source code and	*/
/*	documentation.							*/
/*									*/
/*	Contact:	Kirk Bailey (bailey@ieee.org)			*/
/*									*/
/************************************************************************/

#include	<stdio.h>
#include	<errno.h>
#include	<fcntl.h>
#include	<stdlib.h>
#include	<string.h>
#include	<setjmp.h>
#include	<time.h>
#include	<sys/types.h>
#include	<sys/stat.h>

#ifdef __TURBOC__

#include	<process.h>
#include	<dos.h>
#include	<dir.h>
#include	<io.h>
#define		_getdrive	getdisk
extern	int	utime(/* char *fname, void *ptr */);
unsigned int	_stklen = 32000;

#else	/* !__TURBOC__ */

#ifdef	__WATCOMC__

#include	<process.h>
#include	<io.h>
#include	<sys/utime.h>
/*lint -wlib(0) Inhibit windows.h warnings */
#include	<nt/windows.h>
//void	FreeEnvironmentStrings(char *);
//char	*GetEnvironmentStrings(void);

#else	/* !__WATCOMC__ */

#ifdef	MSDOS

#include	<process.h>
#include	<dos.h>
#include	<io.h>
#include	<sys/utime.h>

#else	/* !MSDOS */

#include	<sys/errno.h>
#include	<sys/wait.h>

#endif	/* MSDOS */
#endif	/* __WATCOMC__ */
#endif	/* __TURBOC__ */

#include	"make.h"

struct	statitem *stat_list = NULL;	/* File time info cache structure */

struct	msi	ms	=	{0};	/* Make program state information */
int	nesting		=	0;	/* Nesting level for recursive MAKEs */
char	*shell_cmds[]	=		/* List of commands requiring shell */
	{
#if	defined(MSDOS) || defined(__TURBOC__)
xyzzy*rtr
	"dir", "type", "rem", "pause", "date", "time", "ren", "rename",
	"ver", "vol", "break", "verify", "mkdir", "md", "exit", "ctty",
	"echo", "if", "cls", "chdir", "cd", "rmdir", "rd", "copy", "del",
	"erase", "set", "for",
	/* Hack alert.  Changing disks is done directly */
	"a:", "b:", "c:", "d:", "e:", "f:", "g:", "h:", "i:", "j:",
	"k:", "l:", "m:", "n:", "o:", "p:", "q:", "r:", "s:", "t:",
	"u:", "v:", "w:", "x:", "y:", "z:",
#endif	/* MSDOS || __TURBOC__ */
	NULL,
	};

/*
 *	main -
 */
void
main(argc,argv,envp)
	int			argc;
	char			**argv;
	char			**envp;
	{
	char			buf[MAXSYMSIZE + 1];
#ifdef	__WATCOMC__
	char			*eblock;
	char			*eptr;
#endif	/* __WATCOMC__ */
	REGISTER int		i;
	struct	statitem	*stat_ptr;
	symptr			symp;
	REGISTER targptr	targp;

	/* Initialize the state of MAKE */

	ms.depend	=	0;	/* -d option */
	ms.dispcount	=	0;	/* Used for -D option */
	ms.display	=	0;	/* -D option */
	ms.file_list	=	NULL;	/* List of file nodes */
	ms.first_targ	=	NULL;	/* 1st target, if nothing explicit */
	ms.flist	=	NULL;	/* Command line make files */
	ms.ignore	=	0;	/* -i option */
	ms.inlen	=	0;	/* Input char counter in "input" */
	ms.input	=	NULL;	/* Char input line pointer */
	ms.mac_env	=	LOW_MACRO;	/* Low relative priority */
	ms.mac_usr	=	HIGH_MACRO;	/* High relative priority */
	ms.mlist	=	NULL;	/* Command line macros */
	ms.noexec	=	0;	/* -n option */
	ms.now		=	0L;	/* Time at startup */
	ms.olds		=	NULL;	/* Token internal pointer */
	ms.print	=	0;	/* -p option */
	ms.readdef	=	1;	/* -r option */
	ms.shell_list	=	NULL;	/* List of shell nodes */
	ms.silent	=	0;	/* -s option */
	ms.suffix_targ	=	NULL;	/* .SUFFIXES target pointer */
	ms.symbol_list	=	NULL;	/* List of symbol nodes */
	ms.target_list	=	NULL;	/* List of target nodes */
	ms.tlist	=	NULL;	/* Command line targets */
	ms.touch	=	0;	/* -t option */

	/* Allocate space for command line targets, files and macros */

	ms.tlist = grow_list(NULL,&ms.tmax);
	ms.flist = grow_list(NULL,&ms.fmax);
	ms.mlist = grow_list(NULL,&ms.mmax);

	/* Return storage possibly previously allocated for the "stat" cache */

	while(stat_list != NULL)
		{
		stat_ptr = stat_list->next;
		tfree(stat_list);
		stat_list = stat_ptr;
		}

	make_args(argc,argv);		/* Process command line options */

	if(ms.noexec)
		{
		ms.silent = 0;		/* -n always displays */
		ms.touch = 0;		/* -n never touches */
		}
/*
 *	Read and process "default.mk".
 */
	if(ms.dispcount > 1)
		ms.display = 1;		/* Display `default.mk' on -DD */

	ms.first_targ = NULL;		/* Used in parse() */

	if(ms.readdef)			/* Read in `default.mk' */
		parse(DEFAULT_MACRO,fopenp(MAKEINI,"r"));
/*
 *	Process command line macros.
 */
	for(i = 0; ms.mlist[i] != NULL; i++)
		add_macro(COMMAND_MACRO,ms.mlist[i]);	/* Cmd line macros */

	tfree(ms.mlist);		/* All done with command line macros */
/*
 *	Process environment macros.  "-e" environment priority control is
 *	handled by "ms.mac_env"/"ms.mac_usr" relative priority.  Be sure to
 *	export macro names from the command line which match environment
 *	variables.
 */
#ifdef	__WATCOMC__	/* Get environment strings from Windows */
	if((eptr = eblock = GetEnvironmentStrings()) != NULL)
		{
		while(*eptr != '\0')
			{
			if((*eptr != '=') && (strlen(eptr) < MAXSYMSIZE))
				{
/* A safe copy to modify */	strcpy(buf,eptr);
				add_macro(ms.mac_env,buf);
/* '='->'\0' in add_macro */	symp = get_symbol(buf);
/* Command line? */		if(symp->sprior == COMMAND_MACRO)
/* Yes, update environment */		{
					strcat(buf,"=");
					strcat(buf,symp->svalue);
					if(putenv(tstrcpy(buf)))
						{
				terror("Insufficient environment memory","");
						}
					}
				}
			while(*eptr++ != '\0')	/* Skip to next string */
				;
			}
		if(! FreeEnvironmentStrings(eblock))
			terror("Unable to free Windows environment block","");
		}

#else	/* !__WATCOMC__*/	/* Get env. strings from non-Windows OS */
	for(i = 0; envp[i] != NULL; i++)
		{
		if(strlen(envp[i]) > MAXSYMSIZE)
			continue;
		strcpy(buf,envp[i]);	/* Since "add_macro" modifies */
		add_macro(ms.mac_env,buf);
		symp = get_symbol(buf);	/* '=' replaced by '\0' by add_macro */
		if(symp->sprior == COMMAND_MACRO)	/* Command line? */
			{
			strcat(buf,"=");	/* Yes, update environment */
			strcat(buf,symp->svalue);
			if(putenv(tstrcpy(buf)))
				terror("Insufficient environment memory","");
			}
		}
#endif	/* __WATCOMC__ */
#ifdef	__TURBOC__
/*
 *	Build some pseudo variables giving our starting state.
 */
	sprintf(buf,"CUR_DISK=%c",_getdrive() + 'a' - 1);
	add_macro(ms.mac_env,buf);
	strcpy(buf,"CUR_DIR=");
	i = strlen(buf);
	(void) getcwd(&buf[i],MAXSYMSIZE - i);
	memmove(&buf[i],&buf[i + 2],strlen(&buf[i]));
	add_macro(ms.mac_env,buf);
#endif	/* __TURBOC__ */
/*
 *	Process user supplied makefiles.
 */
	if(ms.dispcount > 0)
		ms.display = 1;		/* Display makefile's on -D */

	ms.first_targ = NULL;		/* Get first target in `makefile' */

	/* Parse the makefiles given on command line */
	for(i = 0; ms.flist[i] != NULL; i++)
		{
		if(equal(ms.flist[i],"-"))
			parse(ms.mac_usr,stdin);
		else
			parse(ms.mac_usr,fopen(ms.flist[i],"r"));
		}

	tfree(ms.flist);			/* All done with makefile's */

	/* No makefiles specified, so use "makefile"/"Makefile" */
	if(i == 0)
		{
		if(! parse(ms.mac_usr,fopen("makefile","r")))
			parse(ms.mac_usr,fopen("Makefile","r"));
		}
/*
 *	All makefiles read.  Display macro contents and rules to user if
 *	desired.
 */
	if(ms.print)
		dump_data();
/*
 *	Process various targets.
 */
	if((targp = get_target(".INIT")) != NULL)
		{
		add_metas(".INIT","",".INIT");
		build(envp,targp,NULL);	/* Process the .INIT rule */
		}

	for(i = 0; ms.tlist[i] != NULL; i++)
		make(envp,ms.tlist[i],1);	/* Process cmd line targets */

	tfree(ms.tlist);			/* All done with targets */

	/* If no targets specified, make the first one */
	if((i == 0) && (ms.first_targ != NULL))
		make(envp,ms.first_targ->tfile->fname,1);

	if((targp = get_target(".DONE")) != NULL)
		{
		add_metas(".DONE","",".DONE");
		build(envp,targp,NULL);		/* Process the .DONE rule */
		}
	nexit(0);			/* Nothing to complain about */
	}

/*
 *	add_metas -	Add symbols for $*, $< and $@
 *			Make them all at max (command line), priority.
 */
void
add_metas(basename,preqname,targname)
	char			*basename;
	char			*preqname;
	char			*targname;
	{
	add_symbol(COMMAND_MACRO,"*",basename);
	add_symbol(COMMAND_MACRO,"<",preqname);
	add_symbol(COMMAND_MACRO,"@",targname);
	}

/*
 *	build -		Process the shell commands.
 */
void
build(envp,targp,targname)
	char			**envp;
	REGISTER targptr	targp;
	char			*targname;
	{
	char			**argv;	/* Argified version of scmd */
	int			runst;	/* Exec return status */
	char			*scmd;	/* Command with symbols broken out */
#if	defined(MSDOS) || defined(__TURBOC__)
	char			**shcp;	/* Pointer to shell command list */
#endif	/* MSDOS || __TURBOC__ */
	REGISTER shellptr	*shellp;/* Shell command list pointer */
	struct	statitem	*sp;	/* Cache variable		*/
	char			*tcmd;	/* Copy of scmd for `tokenize' */

	if((targp == NULL) || ((shellp = targp->tshell) == NULL))
		return;

	for(; *shellp != NULL; tfree(scmd), tfree(tcmd), tfree(argv), ++shellp)
		{
		/* Breakout runtime symbols (e.g. $* $@ $<) */
		scmd = breakout((*shellp)->scmd);

		if((! (*shellp)->s_silent) && (! ms.silent))
			puts(scmd);

		/* Make a copy because tokenize litters '\0's */
		tcmd = tstrcpy(scmd);
		argv = tokenize(tcmd);

		if(ms.noexec && (! (*shellp)->s_force))
			continue;

		/* Any indirection MUST be handled by the shell */
/*
 *	There was also a colon (:) in the list of reserved characters.  This is
 *	rough on MS-DOS filenames which happen to have drive designators.
 *		16 Oct 1990		DCR
 * If just changing drives, we need to use system.
 */
		if((! (*shellp)->s_shell) && (strpbrk(scmd,"<|>") != NULL))
			(*shellp)->s_shell = 1;
#if	defined(MSDOS) || defined(__TURBOC__)
		/* Likewise, check for COMMAND.COM builtin commands */
		shcp = shell_cmds;
		for(; ((! (*shellp)->s_shell) && (*shcp != NULL)); ++shcp)
			{
			if(equal(*shcp,argv[0]))
				{
				(*shellp)->s_shell = 1;
				break;
				}
			}
#endif	/* MSDOS || __TURBOC__ */
		/* Run without COMMAND.COM if possible, 'cause it uses RAM */

		if((*shellp)->s_shell)
			runst = system(scmd);
		else if(equal(argv[0],"make"))	/* Call MAKE recursively */
			runst = new_make(envp,argv);
/*
 *	The following clause was written by Giorgio Beltrame and allows the
 *	"set" command to affect the macro values of the current environment
 *	as well as the macros passed to a program spawned by MAKE.  If the
 *	semicolon operator is implemented this may have to be removed to avoid
 *	potential conflicts...
 */
#ifdef	NOSEMICOLON
		else if(equal(argv[0],"set"))	/* Modify environment */
			{
			char *buf = strstr(scmd,argv[1]);
			runst = putenv(tstrcpy(buf));
			add_macro(COMMAND_MACRO,buf);
			}
#endif	/* NOSEMICOLON */
		else
			{
			runst = spawnvp(P_WAIT,argv[0],(const char **)argv); 
			if((runst == -1) && (errno == ENOENT))
				runst = system(scmd);
			}

		if(targname != NULL)
			{
			sp = statlook(targname);
			if(sp != (struct statitem *) NULL)
/* Touch the time */		sp->filetime = curr_time();
			}

		if(runst == 0)
			continue;

		/* Uh-oh, an error */
		if(runst == -1)
			{
#if	defined(MSDOS) || defined(__TURBOC__) || defined(__WATCOMC__)
			fprintf(STDERR,"Exit code was %d.  [%s.]\n",
				runst,strerror(errno));
#else	/* MSDOS || __TURBOC__ || __WATCOMC__ */
			fprintf(STDERR,"Exit code was %d.  ",runst);
			perror("make");
#endif	/* MSDOS || __TURBOC__ || __WATCOMC__ */
			}
		else
			fprintf(STDERR,"Exit code was %d.\n",runst);

		if((! ms.ignore) && (! (*shellp)->s_ignore))
			{
			tfree(scmd);
			tfree(tcmd);
			tfree(argv);
/*
 *	Remove target on error.
 */
			if(targname != NULL)
				{
				(void) remove(targname);
				fprintf(STDERR,"Removing target: %s\n",
					targname);
				}
			nexit(1);
			}
		}
	}

/*
 *	curr_time -	Return current time in same format as "file_time".
 *			Used with the `-n' option.
 */
long
curr_time()
	{
	return (time(NULL));
	}

/*
 *	default_rule -	Try the .SUFFIXES when we don't have explicit target.
 *			if `worry' is set, it is an ERROR to NOT build this
 *			target.  `mustbuild' is set if make() has out-of-date
 *			prereq's but no explicit shell rules.
 */
int
default_rule(envp,targname,worry,mustbuild)
	char			**envp;
	char			*targname;
	int			worry;
	int			mustbuild;
	{
	char			*basename;
	int			built;
	char			*ext;
	fileptr			filep;
	char			*preqname;
	REGISTER fileptr	*preqp;
	long			preqtime;
	char			suffrule[MAXFNSIZE + 1];
	REGISTER targptr	targp;
	long			targtime;

	/* Find the extension */
	ext = strrchr(targname,'.');
	if(ext == NULL)
		ext = targname + strlen(targname);

	/* Find the base name */
	basename = tstrncpy(targname,ext - targname);

	targtime = file_time(targname);

	/* Suffix_targ is used to (slightly) speed up this function */
	preqp = (ms.suffix_targ != NULL) ? ms.suffix_targ->tpreq : NULL;
	built = 0;

	while((preqp != NULL) && (*preqp != NULL) && (! built))
		{
		/* Look for a default rule from SUFFIX to `ext' */
		strcat(strcpy(suffrule,(*preqp)->fname),ext);
		targp = get_target(suffrule);	/* e.g. `.c.obj' */

		if(targp != NULL)
			{
			/* Found a rule; see if file exists */
			preqname = tstrcat(basename,(*preqp)->fname);
			preqtime = file_time(preqname);
/*
 *	Don't bother recursive makes unless necessary, e.g. we have .c.obj and
 *	.l.c, but also .l.obj!  We want to use .l.obj if a .c file does not
 *	exist.
 */
			if(preqtime != MAXNEGTIME)
				built = make(envp,preqname,0);

			/* Check if pre-req file exists and is newer */
			preqtime = file_time(preqname);
			if((preqtime > targtime) || (mustbuild && built))
				{
				if(ms.depend)
					{
					display_prereq(targname,targtime,
						preqname,preqtime);
					}

				if(ms.touch)	/* Won't be set if `noexec' */
					touch_file(targname);
				else
					{
					add_metas(basename,preqname,targname);
					build(envp,targp,targname);
					}
				built = 1;
				}
			else if((ms.depend > 1) && (preqtime != MAXNEGTIME))
				{
				display_prereq(targname,targtime,
					preqname,preqtime);
				}

			tfree(preqname);
			}

		++preqp;		/* Try next .SUFFIXES rule */
		}

	if(! built)
		{
		/* Didn't find anything; try the default rule */
		targp = get_target(".DEFAULT");
		if(targp != NULL)
			{
			add_metas(basename,"",targname);
			build(envp,targp,NULL);
			built = 1;
			}
		else if((targtime == MAXNEGTIME) && worry)
			{
			tfree(basename);
			terror("Don't know how to make: ",targname);
			}
		}

	tfree(basename);

	/* Record the current file time */
	if((filep = get_file(targname)) != NULL)
		{
		filep->ftime = (built == 1 && ms.noexec) ? ms.now
					: file_time(targname);
		}

	return (built ? built : ((targtime == MAXNEGTIME) ? 0 : -1));
	}

/*
 *	display_prereq - Display a prerequiste relationship to user.
 */
void
display_prereq(targname,targtime,preqname,preqtime)
	char			*targname;
	long			targtime;
	char			*preqname;
	long			preqtime;
	{
	printf("%s (%08lx) %s than %s (%08lx)\n",
		targname,targtime,((targtime < preqtime) ? "older" : "newer"),
			preqname,preqtime);
	}

/*
 *	file_time -	Get modification time/date of file.
 */
long
file_time(fname)
	char			*fname;
	{
	struct	stat		sbuf;
	struct	statitem	*sp;

	sp = statlook(fname);
	if(sp != (struct statitem *) NULL)
		{
		return (sp->filetime);
		}

	if(stat(fname,&sbuf) || (! (sbuf.st_mode & (short) (S_IFDIR|S_IFREG))))
		return (MAXNEGTIME);

	sp = (struct statitem *) malloc(sizeof(struct statitem) +
		strlen(fname));
	if(sp != (struct statitem *) NULL)
		{
		sp->next = stat_list;
		stat_list = sp;
		sp->filetime = sbuf.st_mtime;
		strcpy(sp->fname,fname);
		}

	return (sbuf.st_mtime);
	}

/*
 *	See if modification time of file is already cached.
 */
struct	statitem	*
statlook(fname)
	char			*fname;
	{
	struct	statitem	*sp;

	for(sp = stat_list; sp != (struct statitem *) NULL; sp = sp->next)
		{
		if(equal(fname,sp->fname))
			return (sp);
		}
	return ((struct statitem *) NULL);
	}

/*
 *	fopenp -	Open file in current directory or along PATH.
 */
FILE	*
fopenp(fname,type)
	char			*fname;
	char			*type;
	{
	FILE			*fd;
	REGISTER char		*fpath;
	REGISTER int		len;
	char			*path;
	char			*tp;

	/* Try to open file relative to current directory */
	if((fd = fopen(fname,type)) != NULL)
		return (fd);

	/* Didn't work, search along path */

	if((path = getenv("PATH")) == NULL)
		return (NULL);

	path = tstrcpy(path);		/* Allocate string and copy */
	fpath = talloc((int) (strlen(path) + strlen(fname) + 2));

	/* Look for tokens separated by semi-colons (;) or colons (:) */

	tp = token(path,PATH_SEPARATOR);
	while(tp != NULL)
		{
		strcpy(fpath,tp);
		len = strlen(fpath) - 1;

		/* Make sure there is a separator between path and filename */

		if(strchr(FILE_SEPARATOR,fpath[len]) == NULL)
			fpath[++len] = '/';

		strcpy(&fpath[len + 1],fname);
		if((fd = fopen(fpath,type)) != NULL)
			break;

		tp = token(NULL,PATH_SEPARATOR);
		}

	tfree(path);
	tfree(fpath);

	return (fd);
	}

/*
 *	grow_list -	Expand the list of pointers by a factor of two.
 */
char	**
grow_list(list,len)
	char			**list;
	int			*len;
	{
	REGISTER int		l;

	l = *len;			/* Get current length */

	/* If list is NULL, start off with a default list */

	if(list == NULL)
		list = (char **) talloc(((l = 1) + 1) * (int) sizeof(char *));
	else
		{
		list = (char **) trealloc((char *)list,
			((l <<= 1) + 1) * (int) sizeof(char *));
		}

	/* If we are initially allocating it, set first pointer to NULL */

	if(l == 1)
		*list = NULL;

	*len = l;			/* Update current length */
	return (list);
	}

/*
 *	make -		Guts of the make command.  Make all pre-requisites,
 *			and if necessary, build target.
 *
 *	returns		-1 target was already up to date w.r.t. pre-requisites
 *			0 target has not been built
 *			1 target is now built (and up to date)
 */
int
make(envp,targname,worry)
	char			**envp;
	char			*targname;
	int			worry;	/* If set, its an error to NOT build */
	{
	char			*basename;
	char			*ext;
	fileptr			filep;
	REGISTER fileptr	*preqp;
	long			preqtime;
	REGISTER targptr 	targp;
	long			targtime;

	/* If recorded time of file is not default, we've already built it */
	filep = get_file(targname);
	if((filep != NULL) && (filep->ftime != MAXNEGTIME))
		return (-1);

	targp = get_target(targname);	/* Find the target node */
	if(targp == NULL)
		return (default_rule(envp,targname,worry,0));

	targtime = file_time(targname);	/* Keep actual time of cur. target */

	/* Must build non-existant files, even with no pre-requisites */
	preqtime = MAXNEGTIME + 1;

	/* Make all pre-requisites */
	preqp = targp->tpreq;
	while((preqp != NULL) && (*preqp != NULL))
		{
		make(envp,(*preqp)->fname,worry);

		/* Keep track of newest pre-requisite */
		if(preqtime < (*preqp)->ftime)
			preqtime = (*preqp)->ftime;

		/* Display as necessary */
		if((ms.depend > 1) ||
			(ms.depend && (*preqp)->ftime > targtime))
			{
			display_prereq(targname,targtime,(*preqp)->fname,
				(*preqp)->ftime);
			}

		++preqp;
		}

	if(targp->tshell == NULL)	/* Try default rules anyway */
		return (default_rule(envp,targname,0,(preqtime > targtime)));
	else if(preqtime > targtime)
		{
		if(ms.touch)		/* Won't be set when `noexec' */
			touch_file(targname);
		else
			{
			/* Compute basename */
			ext = strrchr(targname,'.');
			if(ext == NULL)
				ext = targname + strlen(targname);

			basename = tstrncpy(targname,ext - targname);
			add_metas(basename,"",targname);
			tfree(basename);

			build(envp,targp,targname);
			}

		targp->tfile->ftime =
			(ms.noexec) ? ms.now : file_time(targname);
		return (1);
		}

	targp->tfile->ftime = targtime;

	return (-1);
	}

/*
 *	make_args -	Crack command line arguments
 */
void
make_args(argc,argv)
	int			argc;
	char			**argv;
	{
	REGISTER int		flen;
	REGISTER int		mlen;
	int			show_vers;
	REGISTER int		tlen;

	/* Get current date & time */
	ms.now = curr_time();

	tlen = flen = mlen = 0;
	show_vers = 1;			/* Shown version unless told not to */
	while(++argv, --argc)
		{
		/* Look for macros and targets */

		if(**argv != '-')
			{
			if(strchr(*argv,'=') != NULL)
				{	/* Store as a macro */
				if(mlen == ms.mmax)
					ms.mlist = grow_list(ms.mlist,
							&ms.mmax);
				ms.mlist[mlen++] = *argv;
				}
			else
				{	/* Store as a target */
				if(tlen == ms.tmax)
					ms.tlist = grow_list(ms.tlist,
							&ms.tmax);
				ms.tlist[tlen++] = *argv;
				}
			continue;
			}

		while((*argv != NULL) && *++*argv)
			{
			switch((int) **argv)
				{
				case 'd':	/* Show dependencies */
					ms.depend++;
					break;

				case 'D':	/* Display makefiles */
					ms.dispcount++;
					break;

				case 'e':	/* Environ. macros override */
					ms.mac_env = HIGH_MACRO;
					ms.mac_usr = LOW_MACRO;
					break;

				case 'f':	/* New makefile name */
					if(argc < 2)
						usage();
					if(flen == ms.fmax)
						ms.flist = grow_list(ms.flist,
								&ms.fmax);
					++argv;
					--argc;
					ms.flist[flen++] = *argv;
					*argv = NULL;
					break;

				case 'i':	/* Ignore errors */
					ms.ignore = 1;
					break;

				case 'n':	/* Don't execute commands */
					ms.noexec = 1;
					break;

				case 'r':	/* Don't read default.mk */
					ms.readdef = 0;
					break;

				case 's':	/* Don't echo commands */
					ms.silent = 1;
					break;

				case 't':	/* Touch files, don't build */
					ms.touch = 1;
					break;

				case 'v':
					show_vers = 0;
					break;

				case 'p':
					ms.print = 1;
					break;

				default:
					usage();
				}
			}
		}

	if(show_vers && (nesting == 0))
		printf("MAKE Program Maintenance Utility, %s\n",VERSION);

	/* Terminate all lists with a NULL pointer */

	ms.tlist[tlen] = NULL;
	ms.flist[flen] = NULL;
	ms.mlist[mlen] = NULL;
	}

/*
 *	new_make -	Save current environment, call make() recursively
 *			(actually "main").  Clean up new environment and
 *			restore old environment when done.  Returns "simulated"
 *			program exit code.
 */
int
new_make(envp,argv)
	char			**envp;
	char			**argv;
	{
	int			i;
	fileptr			fnext;
	struct	msi		old_state;
	int			rval = 0;
	shellptr		shnext;
	symptr			snext;
	targptr			tnext;

	/* Save the old state */
	memcpy((char *) &old_state,(char *) &ms,sizeof(struct msi));

	/* Count the arguments */
	for(i = 0; argv[i] != NULL; ++i)
		;

	/* Call ourselves recursively; this inherits flags */
	nesting++;			/* Keep track of level */

	/* Setup "longjmp" target for next level down ("main" never returns) */
	if((rval = setjmp(ms.up_one)) == 0)
		main(i,argv,envp);
	else
		{
		if(rval == -1)
			rval = 0;
		}
	nesting--;			/* We are back up to this level */

	/* We're back, so gotta clean up and dispose of a few things */
	while(ms.target_list != NULL)
		{
		tnext = ms.target_list->tnext;
		if(ms.target_list->tpreq != NULL)
			tfree(ms.target_list->tpreq);
		if(ms.target_list->tshell != NULL)
			tfree(ms.target_list->tshell);
		tfree(ms.target_list);
		ms.target_list = tnext;
		}

	while(ms.file_list != NULL)
		{
		fnext = ms.file_list->fnext;
		tfree(ms.file_list->fname);
		tfree(ms.file_list);
		ms.file_list = fnext;
		}

	while(ms.symbol_list != NULL)
		{
		snext = ms.symbol_list->snext;
		tfree(ms.symbol_list->sname);
		tfree(ms.symbol_list->svalue);
		tfree(ms.symbol_list);
		ms.symbol_list = snext;
		}

	while(ms.shell_list != NULL)
		{
		shnext = ms.shell_list->slink;
		tfree(ms.shell_list->scmd);
		tfree(ms.shell_list);
		ms.shell_list = shnext;
		}

	/* Restore our original state */
	memcpy((char *) &ms,(char *) &old_state,sizeof(struct msi));
	return (rval);
	}

#if	! (defined(MSDOS) || defined(__TURBOC__) || defined(__WATCOMC__))
int
spawnvp(mode,path,args)
	int			mode;
	char			*path;
	char			**args;
	{
	int			pid = 0;
	int			waitword;

	if(mode != P_OVERLAY)
		pid = fork();

	if(pid == 0)
		execvp(path, args);

	wait(&waitword);
	return (waitword);
	}
#endif	/* ! (MSDOS || __TURBOC__ || __WATCOMC__) */

/*
 *	touch_file -	Set the MODIFICATION time of the file to NOW
 */
void
touch_file(targname)
	char			*targname;
	{
	REGISTER int		handle;
	struct	statitem	*sp;

#if	! (defined(MSDOS) || defined(__TURBOC__) || defined(__WATCOMC__))
	time_t			timep[2];

	time(&timep[0]);
	timep[1] = timep[0];
	handle = utime(targname,timep);
#else	/* ! (MSDOS || __TURBOC__ || __WATCOMC__) */
	handle = utime(targname,NULL);
#endif	/* ! (MSDOS || __TURBOC__ || __WATCOMC__) */
	sp = statlook(targname);
	if(sp != (struct statitem *) NULL)
		{
		sp->filetime = curr_time();	/* Touch the time */
		}

	fputs("touch ",stdout);
	puts(targname);

	if(handle == 0)
		return;

	/* Create the file, if it did not exist */
	if(errno == ENOENT)
		{
		handle = open(targname,(O_CREAT | O_TRUNC),S_IWRITE);
		if(handle != -1)
			{
			close(handle);
			return;
			}
		}

	terror("Unable to set modification time of file: ",targname);
	}

/*
 *	usage -		Print usage error message.
 */
void
usage()
	{
terror("Usage: 'make [-f fname] [-edDinprstv] [target...] [macro=value...]'",
		"");
	}

#ifdef __TURBOC__
#pragma	warn	-par

/*
 *	utime -		Set the file time
 */
int
utime(fname,ptr)
	char			*fname;
	void			*ptr;
	{
	FILE			*file;
	struct ftime		ft;
	int			retcode;
	time_t			timeval;
	struct tm		*tm;

	if((file = fopen(fname,"r")) == NULL)
		retcode = -1;
	else
		{
		(void) time(&timeval);
		tm = localtime(&timeval);
		ft.ft_tsec = (unsigned int) tm->tm_sec / 2;
		ft.ft_min = (unsigned int) tm->tm_min;
		ft.ft_hour = (unsigned int) tm->tm_hour;
		ft.ft_day = (unsigned int) tm->tm_mday;
		ft.ft_month = (unsigned int) tm->tm_mon + 1;
		ft.ft_year = (unsigned int) tm->tm_year - 80;
		retcode = setftime(fileno(file),&ft);
		fclose(file);
		}
	return (retcode);
	}

#pragma	warn	.par
#endif	/* __TURBOC__ */
