/************************************************************************/
/*									*/
/*	File:	make3.c							*/
/*									*/
/*	Token processing routines for MAKE.				*/
/*									*/
/*	Written by Greg Yachuk.  Placed in public domain 10/1/88.	*/
/*	Modeled after a MAKE written by Neil Russell which in turn was	*/
/*	modeled after UN*X MAKE...					*/
/*									*/
/*	Modified/maintained by Kirk Bailey of Willamette RF, Inc.	*/
/*									*/
/*	THIS FREE SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR	*/
/*	IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED	*/
/*	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR	*/
/*	PURPOSE ARE SPECIFICALLY DISCLAIMED. IN NO EVENT SHALL THE	*/
/*	AUTHOR OR OTHER CONTRIBUTORS BE LIABLE FOR ANY DIRECT,		*/
/*	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL	*/
/*	DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF		*/
/*	SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR	*/
/*	BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF	*/
/*	LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT	*/
/*	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF	*/
/*	THE USE OF THIS SOFTWARE OR ANY DERIVATIVE WORK, EVEN IF	*/
/*	ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.			*/
/*									*/
/*	This program is hereby placed in the public domain.  In		*/
/*	contrast to other claims of "public domain", this means no	*/
/*	copyright is claimed and you may do anything you like with	*/
/*	MAKE, including selling it!  As a gesture of courtesy, please	*/
/*	retain the authorship information in the source code and	*/
/*	documentation.							*/
/*									*/
/*	Contact:	Kirk Bailey (bailey@ieee.org)			*/
/*									*/
/************************************************************************/

#include	<stdio.h>
#include	<ctype.h>
/*lint -esym(628,__get_errno_ptr) Inhibit stdlib.h warning */
#include	<stdlib.h>
#include	<string.h>
#include	<setjmp.h>

#include	"make.h"

/*
 *	nexit -		Exit with supplied return code if at top level.  If
 *			in a nested MAKE use 'setjmp' to simulate a program
 *			exit to the next higher level of MAKE.
 */
void
nexit(n)
	int			n;
	{
	if(nesting == 0)
		exit(n);
	else
		{
		if(n == 0)
			n = -1;		/* OK flag to "setjmp" */
		longjmp(ms.up_one,n);	/* Simulate MAKE "exit" */
		}
	}

/*
 *	talloc -	Allocate space from heap and bitch if none available.
 */
char	*
talloc(n)
	int			n;
	{
	REGISTER char		*s;

	if((s = malloc((unsigned) n)) == NULL)
		terror("Insufficient free memory","");
	return (s);
	}

/*
 *	terror -	Print error message to user.
 */
void
terror(s1,s2)
	char			*s1;
	char			*s2;
	{
	fprintf(STDERR,"FATAL: %s%s\n",s1,s2);
	nexit(1);
	}

/*
 *	tgets -		Read input, swallowing escaped newlines as necessary.
 */
char	*
tgets(fd)
	FILE			*fd;
	{
	REGISTER char		*ep;
	int			len;

	if(ms.inlen == 0)
		ms.input = talloc(ms.inlen = MAXLNSIZE);

	ms.input[ms.inlen-2] = '\n';
	ep = ms.input - 1;
	while((fgets(ms.input,ms.inlen,fd)) != NULL)
		{
		for(;;)
			{
			while((ms.input[ms.inlen-2] != '\n') &&
				(ms.input[ms.inlen-2] != '\0'))
				{
				len = ms.inlen;
				ms.input = trealloc(ms.input,(ms.inlen <<= 1));
				ep = &ms.input[len - 2];
				ms.input[ms.inlen - 2] = '\n';
				(void) fgets(ep + 1,len + 1,fd);
				}

			while (*++ep)
				;
			*--ep = '\0';

			do
				{
				--ep;
				} while (isspace(*ep) && (ep >= ms.input));

			if((*ep == '\\') && (*--ep != '\\'))
				{
				(void) fgets(ep + 1,ms.inlen -
					((int) (ep-ms.input)) - 1,fd);
				}
			else
				break;
			}

		return (ms.input);
		}

	ms.inlen = 0;
	tfree(ms.input);
	ms.input = NULL;

	return (NULL);
	}

/*
 *	token -		Take an input string and return a token each call.  The
 *			default token delimiter characters are `isspace()'.
 *			Text between quotes (" and ') is a single token.
 *
 *	Called as	s = token(string,seps);
 *	or		s = token(string,NULL);
 *
 *	Followed by	s = token(NULL,seps);
 *	or		s = token(NULL,NULL);
 *
 *	Returns NULL when no more tokens are available.
 */
char	*
token(s,sep)
	char			*s;
	char			*sep;
	{
	int			had_sep = 0;
	char			quote = 0;	/* Processing a quoted token */

	if(s != NULL)
		ms.olds = s;		/* We are starting all over again */

	if(ms.olds == NULL)
		return (NULL);

/*lint -e514 Lint apparently doesn't like applying |= to a boolean */
	while((*ms.olds != '\0') &&
		(isspace(*ms.olds) || ((sep != NULL) &&
			(had_sep |= (strchr(sep,*ms.olds) != NULL)))))
		{
		++ms.olds;		/* Skip leading spaces and sep's */
		}
/*lint +e514 Lint apparently doesn't like applying |= to a boolean */

	if((*ms.olds == '\0') && (had_sep == 0))
		return (NULL);		/* No tokens to return */
	else
		s = ms.olds;

	while(*ms.olds != '\0')
		{
		if(isspace(*ms.olds) || ((sep != NULL) &&
			(strchr(sep,*ms.olds) != NULL)))
			{
			*ms.olds++ = '\0';	/* Delimit the token */
			return (s);
			}
		else if((*ms.olds == '"') || (*ms.olds == '\''))
			{
			quote = *ms.olds++;
			while((*ms.olds != quote) && (*ms.olds != '\0'))
				++ms.olds;	/* Search for quote char */
			if(*ms.olds != '\0')
				++ms.olds;	/* No EOS, skip quote */
			}
		else
			++ms.olds;		/* Otherwise, pass over char */
		}

	ms.olds = NULL;
	return (s);			/* Return last token */
	}

/*
 *	tokenize -	chop a string up into an array of (char *)'s.
 */
char	**
tokenize(input)
	char			*input;
	{
	REGISTER int		alen;
	REGISTER int		argc = 0;
	char			**argv;

	alen = 20;			/* Good initial guess */
	argv = (char **) talloc((alen + 1) * (int) sizeof (char *));

	input = token(input,NULL);	/* Use default separators */
	while(input != NULL)
		{
		if(alen == argc)
			argv = (char **) trealloc((char *) argv,
					(alen <<= 1) * (int) sizeof (char *));
/*
 *	Remove quotes from around tokens.  Note that quotes must be same sex.
 *	The trailing quote isn't required.
 */
/*lint -e{613} */
		if((*input == '\'') || (*input == '"'))
			{
			*strrchr(input, *input) = '\0';
			input++;
			}
		argv[argc++] = input;
		input = token(NULL,NULL);
		}

	argv[argc] = NULL;		/* Mark end of array */

	return (argv);
	}

/*
 *	trealloc -	(Re)allocate space from heap and bitch if unavailable.
 */
char   *
trealloc(s,n)
	REGISTER char		*s;
	int			n;
	{
	s = realloc(s,(unsigned) n);
	if(s == NULL)
		talloc(n);		/* Force an error */
	return (s);
	}

/*
 *	tstrncpy -	Get the storage for a string from the heap and copy
 *			it into it.
 */
char   *
tstrncpy(s,n)
	REGISTER char		*s;
	int			n;
	{
	s = strncpy(talloc(n + 1),s,(unsigned) n);
	s[n] = '\0';
	return (s);
	}
